@@ -1,5 +1,5 @@
 /*
- license: The MIT License, Copyright (c) 2018-2020 YUKI "Piro" Hiroshi
+ license: The MIT License, Copyright (c) 2018-2025 YUKI "Piro" Hiroshi
  original:
    https://github.com/piroor/webextensions-lib-menu-ui
 */
@@ -53,9 +53,32 @@
       return array;
     }
 
+    RTL_LANGUAGES = new Set([
+      'ar',
+      'he',
+      'fa',
+      'ur',
+      'ps',
+      'sd',
+      'ckb',
+      'prs',
+      'rhg',
+    ]);
+
+    get isRTL() {
+      const lang = (
+        navigator.language ||
+        navigator.userLanguage ||
+        //(new Intl.DateTimeFormat()).resolvedOptions().locale ||
+        ''
+      ).split('-')[0];
+      return this.RTL_LANGUAGES.has(lang);
+    }
+
     constructor(params = {}) {
       this.$lastHoverItem   = null;
       this.$lastFocusedItem = null;
+      this.$mouseDownFired  = false;
 
       this.root              = params.root;
       this.onCommand         = params.onCommand || (() => {});
@@ -65,6 +88,8 @@
       this.subMenuOpenDelay  = params.subMenuOpenDelay || 300;
       this.subMenuCloseDelay = params.subMenuCloseDelay || 300;
       this.appearance        = params.appearance || 'menu';
+      this.incrementalSearch = params.incrementalSearch || false;
+      this.incrementalSearchTimeout = params.incrementalSearchTimeout || 1000;
 
       this.$onBlur            = this.$onBlur.bind(this);
       this.$onMouseOver       = this.$onMouseOver.bind(this);
@@ -76,12 +101,15 @@
       this.$onTransitionEnd   = this.$onTransitionEnd.bind(this);
       this.$onContextMenu     = this.$onContextMenu.bind(this);
 
+      this.root.classList.toggle('rtl', this.isRTL);
+
       if (!this.root.id)
         this.root.id = `MenuUI-root-${this.$uniqueKey}-${parseInt(Math.random() * Math.pow(2, 16))}`;
 
       this.root.classList.add(this.$commonClass);
       this.root.classList.add('menu-ui');
       this.root.classList.add(this.appearance);
+      this.root.setAttribute('role', 'menu');
 
       this.$screen = document.createElement('div');
       this.$screen.classList.add(this.$commonClass);
@@ -93,6 +121,9 @@
       this.$marker.classList.add('menu-ui-marker');
       this.$marker.classList.add(this.appearance);
       this.root.parentNode.insertBefore(this.$marker, this.root.nextSibling);
+
+      this.$lastKeyInputAt = -1;
+      this.$incrementalSearchString = '';
     }
 
     get opened() {
@@ -101,10 +132,14 @@
 
     $updateAccessKey(item) {
       const ACCESS_KEY_MATCHER = /(&([^\s]))/i;
+
       const title = item.getAttribute('title');
       if (title)
         item.setAttribute('title', title.replace(ACCESS_KEY_MATCHER, '$2'));
+
       const label = MenuUI.$evaluateXPath('child::text()', item, XPathResult.STRING_TYPE).stringValue;
+      item.dataset.lowerCasedText = label.toLowerCase();
+
       const matchedKey = label.match(ACCESS_KEY_MATCHER);
       if (matchedKey) {
         const textNode = MenuUI.$evaluateXPath(
@@ -198,9 +233,30 @@
     }
 
     $updateItems(parent) {
+      parent.setAttribute('role', 'menu');
       for (const item of parent.querySelectorAll('li:not(.separator)')) {
         item.setAttribute('tabindex', -1);
         item.classList.remove('open');
+
+        if (item.classList.contains('checkbox'))
+          item.setAttribute('role', 'menuitemcheckbox');
+        else if (item.classList.contains('radio'))
+          item.setAttribute('role', 'menuitemradio');
+        else if (item.classList.contains('separator'))
+          item.setAttribute('role', 'separator');
+        else
+          item.setAttribute('role', 'menuitem');
+
+        if (item.matches('.checked, .radio')) {
+          if (item.classList.contains('checked'))
+            item.setAttribute('aria-checked', 'true');
+          else
+            item.setAttribute('aria-checked', 'false');
+        }
+        else {
+          item.removeAttribute('aria-checked');
+        }
+
         this.$updateAccessKey(item);
         const icon = item.querySelector('span.icon') || document.createElement('span');
         if (!icon.parentNode) {
@@ -271,13 +327,31 @@
           this.$marker.style.top = `calc(${top}px + ${menuRect.height}px - 0.6em)`;
         }
 
-        if (containerRect.right - anchorRect.left >= menuRect.width) {
-          left = anchorRect.left;
-          this.$marker.style.left = `calc(${left}px + 0.5em)`;
-        }
-        else if (anchorRect.left - containerRect.left >= menuRect.width) {
-          left = Math.max(0, anchorRect.right - menuRect.width);
-          this.$marker.style.left = `calc(${left}px + ${menuRect.width}px - 1.5em)`;
+        const canPlaceAtRight = containerRect.right - anchorRect.left >= menuRect.width;
+        const canPlaceAtLeft  = anchorRect.left - containerRect.left >= menuRect.width;
+
+        if (canPlaceAtRight || canPlaceAtLeft) {
+          if (this.isRTL) {
+            if (canPlaceAtLeft) {
+              left = Math.max(0, anchorRect.right - menuRect.width);
+              this.$marker.style.left = `calc(${left}px + ${menuRect.width}px - 1.5em)`;
+            }
+            else {
+              left = anchorRect.left;
+              this.$marker.style.left = `calc(${left}px + 0.5em)`;
+            }
+          }
+          else {
+            if (canPlaceAtRight) {
+              left = anchorRect.left;
+              this.$marker.style.left = `calc(${left}px + 0.5em)`;
+            }
+            else {
+              left = Math.max(0, anchorRect.right - menuRect.width);
+              this.$marker.style.left = `calc(${left}px + ${menuRect.width}px - 1.5em)`;
+            }
+          }
+
         }
         else {
           left = Math.max(0, containerRect.left - menuRect.width);
@@ -288,12 +362,15 @@
       let parentRect;
       if (menu.parentNode.localName == 'li') {
         parentRect = menu.parentNode.getBoundingClientRect();
-        left = parentRect.right;
+        left = this.isRTL ? parentRect.left - menuRect.width : parentRect.right;
         top  = parentRect.top;
       }
 
       if (left === undefined)
         left = Math.max(0, (containerRect.width - menuRect.width) / 2);
+      else if (this.isRTL)
+        left -= menuRect.width;
+
       if (top === undefined)
         top = Math.max(0, (containerRect.height - menuRect.height) / 2);
 
@@ -342,6 +419,7 @@
       this.$mouseDownAfterOpen = false;
       this.$lastFocusedItem = null;
       this.$lastHoverItem = null;
+      this.$mouseDownFired = false;
       this.anchor = null;
       this.canceller = null;
       return new Promise((resolve, _reject) => {
@@ -446,7 +524,7 @@
 
     $openSubmenuFor(item) {
       const items = MenuUI.$evaluateXPath(
-        `ancestor-or-self::li[${MenuUI.$hasClass('has-submenu')}]`,
+        `ancestor-or-self::li[${MenuUI.$hasClass('has-submenu')}][not(${MenuUI.$hasClass('disabled')})]`,
         item
       );
       for (const item of MenuUI.$getArrayFromXPathResult(items)) {
@@ -474,6 +552,7 @@
       event.stopPropagation();
       event.preventDefault();
       this.$mouseDownAfterOpen = true;
+      this.$mouseDownFired = true;
     }
 
     $getEffectiveItem(node) {
@@ -505,7 +584,8 @@
         target.classList.contains('separator') ||
         target.classList.contains('has-submenu') ||
         target.classList.contains('disabled')) {
-        if (!event.target.closest(`#${this.root.id}`))
+        if (this.$mouseDownFired && // ignore "click" event triggered by a mousedown fired before the menu is opened (like long-press)
+            !event.target.closest(`#${this.root.id}`))
           return this.close();
         return;
       }
@@ -524,6 +604,31 @@
       return null;
     }
 
+    $incrementalSearchNextFocusedItem(key) {
+      this.$incrementalSearchString += key.toLowerCase();
+      const current = this.$lastHoverItem || this.$lastFocusedItem || this.root.firstChild;
+      const condition = `starts-with(@data-lower-cased-text, "${this.$incrementalSearchString.replace(/"/g, '\\"')}")`;
+      if (this.$isItemMatches(current, condition))
+        return current;
+      const item = this.$getNextItem(current, condition);
+      return item;
+    }
+
+    $shouldSearchIncremental() {
+      if (!this.incrementalSearch)
+        return false;
+
+      const last = this.$lastKeyInputAt;
+      const now = this.$lastKeyInputAt = Date.now();
+      if (last < 0)
+        return true; // start
+
+      if (now - last > this.incrementalSearchTimeout)
+        this.$incrementalSearchString = '';
+
+      return true; // continue
+    }
+
     $onKeyDown(event) {
       switch (event.key) {
         case 'ArrowUp':
@@ -541,13 +646,19 @@
         case 'ArrowRight':
           event.stopPropagation();
           event.preventDefault();
-          this.$digIn();
+          if (this.isRTL)
+            this.$digOut();
+          else
+            this.$digIn();
           break;
 
         case 'ArrowLeft':
           event.stopPropagation();
           event.preventDefault();
-          this.$digOut();
+          if (this.isRTL)
+            this.$digIn();
+          else
+            this.$digOut();
           break;
 
         case 'Home':
@@ -593,14 +704,43 @@
             this.$digOut();
         }; break;
 
+        case 'BackSpace': {
+          if (this.$shouldSearchIncremental()) {
+            event.stopPropagation();
+            event.preventDefault();
+            this.$incrementalSearchString = this.$incrementalSearchString.slice(0, this.$incrementalSearchString.length - 1);
+            const item = this.$incrementalSearchNextFocusedItem('');
+            if (item) {
+              this.focusTo(item);
+              this.$setHover(null);
+            }
+          }
+        }; break;
+
         default:
           if (event.key.length == 1) {
+            if (this.$shouldSearchIncremental()) {
+              event.stopPropagation();
+              event.preventDefault();
+              const item = this.$incrementalSearchNextFocusedItem(event.key);
+              if (item) {
+                this.focusTo(item);
+                this.$setHover(null);
+              }
+              return;
+            }
+
             const item = this.$getNextFocusedItemByAccesskey(event.key);
             if (item) {
               this.focusTo(item);
               this.$setHover(null);
-              if (this.$getNextFocusedItemByAccesskey(event.key) == item)
-                this.onCommand(item, event);
+              if (this.$getNextFocusedItemByAccesskey(event.key) == item &&
+                  !item.classList.contains('disabled')) {
+                if (item.querySelector('ul'))
+                  this.$digIn();
+                else
+                  this.onCommand(item, event);
+              }
             }
           }
           return;
@@ -617,13 +757,15 @@
         case 'End':
         case 'Enter':
         case 'Escape':
+        case 'Bakcspace':
           event.stopPropagation();
           event.preventDefault();
           return;
 
         default:
           if (event.key.length == 1 &&
-            this.$getNextFocusedItemByAccesskey(event.key)) {
+              (this.$shouldSearchIncremental() ||
+               this.$getNextFocusedItemByAccesskey(event.key))) {
             event.stopPropagation();
             event.preventDefault();
           }
@@ -639,16 +781,16 @@
           base,
           XPathResult.FIRST_ORDERED_NODE_TYPE
         ).singleNodeValue ||
-      MenuUI.$evaluateXPath(
-        `following-sibling::li[not(${MenuUI.$hasClass('separator')})]${extrcondition}[last()]`,
-        base,
-        XPathResult.FIRST_ORDERED_NODE_TYPE
-      ).singleNodeValue ||
-      MenuUI.$evaluateXPath(
-        `self::li[not(${MenuUI.$hasClass('separator')})]${extrcondition}`,
-        base,
-        XPathResult.FIRST_ORDERED_NODE_TYPE
-      ).singleNodeValue
+        MenuUI.$evaluateXPath(
+          `following-sibling::li[not(${MenuUI.$hasClass('separator')})]${extrcondition}[last()]`,
+          base,
+          XPathResult.FIRST_ORDERED_NODE_TYPE
+        ).singleNodeValue ||
+        MenuUI.$evaluateXPath(
+          `self::li[not(${MenuUI.$hasClass('separator')})]${extrcondition}`,
+          base,
+          XPathResult.FIRST_ORDERED_NODE_TYPE
+        ).singleNodeValue
       );
       if (window.getComputedStyle(item, null).display == 'none')
         return this.$getPreviousItem(item, condition);
@@ -663,22 +805,31 @@
           base,
           XPathResult.FIRST_ORDERED_NODE_TYPE
         ).singleNodeValue ||
-      MenuUI.$evaluateXPath(
-        `preceding-sibling::li[not(${MenuUI.$hasClass('separator')})]${extrcondition}[last()]`,
-        base,
-        XPathResult.FIRST_ORDERED_NODE_TYPE
-      ).singleNodeValue ||
-      MenuUI.$evaluateXPath(
-        `self::li[not(${MenuUI.$hasClass('separator')})]${extrcondition}`,
-        base,
-        XPathResult.FIRST_ORDERED_NODE_TYPE
-      ).singleNodeValue
+        MenuUI.$evaluateXPath(
+          `preceding-sibling::li[not(${MenuUI.$hasClass('separator')})]${extrcondition}[last()]`,
+          base,
+          XPathResult.FIRST_ORDERED_NODE_TYPE
+        ).singleNodeValue ||
+        MenuUI.$evaluateXPath(
+          `self::li[not(${MenuUI.$hasClass('separator')})]${extrcondition}`,
+          base,
+          XPathResult.FIRST_ORDERED_NODE_TYPE
+        ).singleNodeValue
       );
       if (item && window.getComputedStyle(item, null).display == 'none')
         return this.$getNextItem(item, condition);
       return item;
     }
 
+    $isItemMatches(base, condition = '') {
+      const extrcondition = condition ? `[${condition}]` : '' ;
+      return !!MenuUI.$evaluateXPath(
+        `self::li[not(${MenuUI.$hasClass('separator')})]${extrcondition}`,
+        base,
+        XPathResult.FIRST_ORDERED_NODE_TYPE
+      ).singleNodeValue;
+    }
+
     $advanceFocus(direction, lastFocused = null) {
       lastFocused = lastFocused || this.$lastHoverItem || this.$lastFocusedItem;
       if (!lastFocused) {
@@ -758,6 +909,9 @@
           position: fixed;
           z-index: 999999;
         }
+        ${common}.menu-ui.rtl {
+          direction: rtl;
+        }
 
         ${common}.menu-ui.open,
         ${common}.menu-ui.open li.open > ul {
@@ -785,7 +939,7 @@
         ${common}.menu-ui li.checkbox.checked::before {
           content: "✔";
           position: absolute;
-          left: 0.25em;
+          inset-inline-start: 0.5em;
         }
 
         ${common}.menu-ui li.separator {
@@ -797,12 +951,12 @@
 
         ${common}.menu-ui li.has-submenu,
         ${common}.menu-ui.menu li.has-submenu {
-          padding-right: 1em;
+          padding-inline-end: 1em;
         }
         ${common}.menu-ui li.has-submenu::after {
           content: "❯";
+          inset-inline-end: 0.25em;
           position: absolute;
-          right: 0.25em;
           transform: scale(0.75);
         }
 
@@ -863,10 +1017,32 @@
           opacity: 0.5;
         }
 
-        ${common}.menu-ui li[data-icon] {
-          background-position: left center;
+        ${common}.menu-ui li[data-icon],
+        ${common}.menu-ui.menu li[data-icon],
+        ${common}.menu-ui.panel li[data-icon] {
+          --icon-size: 16px;
           background-repeat: no-repeat;
-          background-size: 16px;
+          background-size: var(--icon-size);
+          padding-inline-start: calc(var(--icon-size) + 0.7em);
+        }
+        ${common}.menu-ui:not(.rtl) li[data-icon],
+        ${common}.menu-ui:not(.rtl).menu li[data-icon],
+        ${common}.menu-ui:not(.rtl).panel li[data-icon] {
+          background-position: left center;
+        }
+        ${common}.menu-ui.rtl li[data-icon],
+        ${common}.menu-ui.rtl.menu li[data-icon],
+        ${common}.menu-ui.rtl.panel li[data-icon] {
+          background-position: right center;
+        }
+
+        ${common}.menu-ui li.checkbox,
+        ${common}.menu-ui li.radio,
+        ${common}.menu-ui.menu li.checkbox,
+        ${common}.menu-ui.panel li.checkbox,
+        ${common}.menu-ui.menu li.radio,
+        ${common}.menu-ui.panel li.radio {
+          padding-inline-start: 1.7em;
         }
 
         /* panel-like appearance */
@@ -887,7 +1063,7 @@
         }
 
         ${common}.menu-ui.panel li {
-          padding: 0.15em 1em;
+          padding: 0.15em 1em 0.15em 0.7em;
         }
 
 
@@ -910,7 +1086,7 @@
         }
 
         ${common}.menu-ui.menu li {
-          padding: 0.15em 0.5em 0.15em 1.5em;
+          padding: 0.15em 0.5em 0.15em 0.7em;
         }
 
         ${common}.menu-ui.menu li.separator {
@@ -923,28 +1099,28 @@
           visibility: visible;
         }
 
-        ${common}.menu-ui.menu li[data-icon]:not([data-icon-color]),
-        ${common}.menu-ui.panel li[data-icon]:not([data-icon-color]) {
-          background-position: calc(1em - 16px + 0.5em) center;
+        ${common}.menu-ui.menu:not(.rtl) li[data-icon]:not([data-icon-color]),
+        ${common}.menu-ui.panel:not(.rtl) li[data-icon]:not([data-icon-color]) {
+          background-position: 0.5em center;
         }
-
-        ${common}.menu-ui.panel li[data-icon]:not([data-icon-color]) {
-          padding-left: calc(16px + 0.2em);
+        ${common}.menu-ui.menu.rtl li[data-icon]:not([data-icon-color]),
+        ${common}.menu-ui.panel.rtl li[data-icon]:not([data-icon-color]) {
+          background-position: calc(100% - 0.5em) center;
         }
 
+        ${common}.menu-ui li:not([data-icon]) .icon,
         ${common}.menu-ui li[data-icon]:not([data-icon-color]) .icon {
           display: none;
         }
 
         ${common}.menu-ui li[data-icon][data-icon-color] .icon {
           display: inline-block;
-          height: 1em;
-          left: 0.5em;
-          max-height: 1em;
-          max-width: 1em;
+          height: var(--icon-size);
+          inset-inline-start: 0.5em;
+          max-height: var(--icon-size);
+          max-width: var(--icon-size);
           position: absolute;
-          margin-top: 0.15em;
-          width: 1em;
+          width: var(--icon-size);
         }
       `;
       document.head.appendChild(this.style);
