@@ -5,6 +5,8 @@
 */
 'use strict';
 
+import EventListenerManager from '/extlib/EventListenerManager.js';
+
 import {
   log as internalLogger,
   dumpTab,
@@ -13,14 +15,12 @@ import {
 
 import * as TabsStore from './tabs-store.js';
 
-import EventListenerManager from '/extlib/EventListenerManager.js';
-
 function log(...args) {
   internalLogger('common/Window', ...args);
 }
 
 export default class Window {
-  constructor(windowId) {
+  constructor(windowId, tabGroups) {
     const alreadyTracked = TabsStore.windows.get(windowId);
     if (alreadyTracked)
       return alreadyTracked;
@@ -29,41 +29,61 @@ export default class Window {
 
     this.id    = windowId;
     this.tabs  = new Map();
+    this.tabGroups = new Map();
+    if (tabGroups) {
+      this.initTabGroups(tabGroups);
+    }
     this.order = [];
 
-    this.element = null;
-    this.classList = null;
+    this.containerElement = null;
+    this.containerClassList = null;
+    this.pinnedContainerElement = null;
 
-    this.internalMovingTabs  = new Set();
-    this.alreadyMovedTabs    = new Set();
+    this.internalMovingTabs  = new Map();
+    this.alreadyMovedTabs    = new Map();
     this.internalClosingTabs = new Set();
+    this.keepDescendantsTabs = new Set();
+    this.highlightingTabs    = new Set();
     this.tabsToBeHighlightedAlone = new Set();
+    this.internallyMovingTabsForUpdatedNativeTabGroups = new Set();
 
     this.subTreeMovingCount =
       this.subTreeChildrenMovingCount =
       this.doingIntelligentlyCollapseExpandCount =
-      this.internalFocusCount =
-      this.internalSilentlyFocusCount =
       this.duplicatingTabsCount = 0;
 
-    this.preventAutoGroupNewTabsUntil = Date.now() + configs.autoGroupNewTabsDelayOnNewWindow;
+    this.internallyFocusingTabs = new Set();
+    this.internallyFocusingByMouseTabs = new Set();
+    this.internallyFocusingSilentlyTabs = new Set();
+
+    this.preventToDetectTabBunchesUntil = Date.now() + configs.tabBunchesDetectionDelayOnNewWindow;
 
     this.openingTabs   = new Set();
 
     this.openedNewTabs = new Map();
 
+    this.bypassTabControlCount       = 0;
+    this.toBeOpenedNewTabCommandTab  = 0;
     this.toBeOpenedTabsWithPositions = 0;
+    this.toBeOpenedTabsWithCookieStoreId = 0;
     this.toBeOpenedOrphanTabs        = 0;
 
     this.toBeAttachedTabs = new Set();
     this.toBeDetachedTabs = new Set();
 
+    this.lastRelatedTabs = new Map();
+    this.previousLastRelatedTabs = new Map();
+
     TabsStore.windows.set(windowId, this);
     TabsStore.prepareIndexesForWindow(windowId);
 
     // We should initialize private properties with blank value for better performance with a fixed shape.
     this.delayedDestroy = null;
   }
+  initTabGroups(tabGroups) {
+    log(`initializing tabGroups of window ${this.id}: `, tabGroups);
+    this.tabGroups = new Map((tabGroups || []).map(group => [group.id, group]));
+  }
 
   destroy() {
     for (const tab of this.tabs.values()) {
@@ -71,18 +91,24 @@ export default class Window {
         tab.$TST.destroy();
     }
     this.tabs.clear();
+    this.tabGroups.clear();
     TabsStore.windows.delete(this.id);
     TabsStore.unprepareIndexesForWindow(this.id);
 
-    if (this.element) {
-      const element = this.element;
-      if (element.parentNode && !element.hasChildNodes()) {
+    if (this.containerElement) {
+      const element = this.containerElement;
+      if (element.parentNode && !element.hasChildNodes())
         element.parentNode.removeChild(element);
-        this.unbindElement();
-      }
     }
+    if (this.pinnedContainerElement) {
+      const element = this.element;
+      if (element.parentNode && !element.hasChildNodes())
+        element.parentNode.removeChild(element);
+    }
+    this.unbindElements();
 
     this.tabs = null;
+    this.tabGroups = null;
     this.order = null;
     this.id = null;
   }
@@ -92,16 +118,35 @@ export default class Window {
     this.order = [];
     TabsStore.unprepareIndexesForWindow(this.id);
     TabsStore.prepareIndexesForWindow(this.id);
+    this.clearLastRelatedTabs();
   }
 
-  bindElement(element) {
-    this.element = element;
-    this.classList = element.classList;
+  clearLastRelatedTabs() {
+    for (const openerId of this.lastRelatedTabs.keys()) {
+      const opener = this.tabs.get(openerId);
+      if (!opener)
+        continue;
+      opener.$TST.newRelatedTabsCount = 0;
+    }
+    this.lastRelatedTabs.clear();
+    this.previousLastRelatedTabs.clear();
   }
 
-  unbindElement() {
-    this.element = null;
-    this.classList = null;
+  bindContainerElement(element) {
+    this.containerElement = element;
+    this.containerClassList = element.classList;
+  }
+
+  bindPinnedContainerElement(element) {
+    this.pinnedContainerElement = element;
+    this.pinnedContainerClassList = element.classList;
+  }
+
+  unbindElements() {
+    this.containerElement = null;
+    this.containerClassList = null;
+    this.pinnedContainerElement = null;
+    this.pinnedContainerClassList = null;
   }
 
   getOrderedTabs(startId, endId, tabs) {
@@ -156,22 +201,24 @@ export default class Window {
 
     const order = this.order;
     if (this.tabs.has(tab.id)) { // already tracked: update
+      const prevState = tab.reindexedBy;
       const index = order.indexOf(tab.id);
       order.splice(index, 1);
       order.splice(tab.index, 0, tab.id);
-      for (let i = Math.min(index, tab.index), maxi = Math.min(Math.max(index + 1, tab.index + 1), order.length); i < maxi; i++) {
+      for (let i = Math.min(index, tab.index), maxi = Math.min(Math.max(index, tab.index) + 1, order.length); i < maxi; i++) {
         const tab = this.tabs.get(order[i]);
         if (!tab)
           throw new Error(`Unknown tab: ${i}/${order[i]} (${order.join(', ')})`);
         tab.index = i;
         tab.reindexedBy = `Window.property.trackTab/update (${tab.index})`;
+        tab.$TST.invalidateCache();
       }
       const parent = tab.$TST.parent;
       if (parent) {
-        parent.$TST.sortChildren();
+        parent.$TST.sortAndInvalidateChildren();
         parent.$TST.invalidateCachedAncestors();
       }
-      log(`tab ${dumpTab(tab)} is re-tracked under the window ${this.id}: `, order);
+      log(`tab ${dumpTab(tab)} is re-tracked under the window ${this.id}: `, prevState, index, '=>', tab.reindexedBy, order.join(', '));
     }
     else { // not tracked yet: add
       this.tabs.set(tab.id, tab);
@@ -182,10 +229,12 @@ export default class Window {
           throw new Error(`Unknown tab: ${i}/${order[i]} (${order.join(', ')})`);
         tab.index = i;
         tab.reindexedBy = `Window.property.trackTab/new (${tab.index})`;
+        tab.$TST.invalidateCache();
       }
       log(`tab ${dumpTab(tab)} is newly tracked under the window ${this.id}: `, order);
     }
     TabsStore.updateIndexesForTab(tab);
+    tab.$TST.invalidateCache();
     return tab;
   }
 
@@ -235,14 +284,20 @@ export default class Window {
     for (const tab of this.getOrderedTabs()) {
       tabs.push(tab.$TST.export(full));
     }
-    return tabs;
+    return {
+      tabs,
+      tabGroups: [...this.tabGroups.values()].map(group => group.$TST.sanitized),
+    };
   }
 }
 
 Window.onInitialized = new EventListenerManager();
 
-Window.init = windowId => {
-  const window = TabsStore.windows.get(windowId) || new Window(windowId);
-  Window.onInitialized.dispatch(window);
-  return window;
+Window.init = (windowId, tabGroups) => {
+  const win = TabsStore.windows.get(windowId) || new Window(windowId, tabGroups);
+  if (tabGroups && tabGroups.size != win.tabGroups.size) {
+    win.initTabGroups(tabGroups);
+  }
+  Window.onInitialized.dispatch(win);
+  return win;
 }
