@@ -29,56 +29,42 @@ export async function getIndexes(...queriedTabIds) {
   return indexes.map(tab => tab ? tab.index : -1);
 }
 
-// workaround for https://bugzilla.mozilla.org/show_bug.cgi?id=1394477 + fix pinned/unpinned status
-export async function safeMoveAcrossWindows(tabIds, moveOptions) {
-  log('safeMoveAcrossWindows ', tabIds, moveOptions);
-  if (!Array.isArray(tabIds))
-    tabIds = [tabIds];
-  const tabs = await Promise.all(tabIds.map(id => browser.tabs.get(id).catch(handleMissingTabError)));
-  const window = await browser.windows.get(moveOptions.windowId || tabs[0].windowId, { populate: true });
-  return (await Promise.all(tabs.map(async (tab, index) => {
-    try {
-      const destIndex = moveOptions.index + index;
-      if (tab.pinned) {
-        if (window.tabs[destIndex - 1] &&
-            window.tabs[destIndex - 1].pinned != tab.pinned)
-          await browser.tabs.update(tab.id, { pinned: false });
-      }
-      else {
-        if (window.tabs[destIndex] &&
-            window.tabs[destIndex].pinned != tab.pinned)
-          await browser.tabs.update(tab.id, { pinned: true });
-      }
-      let movedTab = await browser.tabs.move(tab.id, {
-        ...moveOptions,
-        index: destIndex
-      });
-      log(`safeMoveAcrossWindows: movedTab[${index}] = `, movedTab);
-      if (Array.isArray(movedTab))
-        movedTab = movedTab[0];
-      return movedTab;
-    }
-    catch(e) {
-      handleMissingTabError(e);
-      return null;
-    }
-  }))).filter(tab => !!tab);
+export function isMissingTabError(error) {
+  return (
+    error &&
+    error.message &&
+    error.message.includes('Invalid tab ID:')
+  );
 }
-
 export function handleMissingTabError(error) {
-  if (!error ||
-      !error.message ||
-      error.message.indexOf('Invalid tab ID:') != 0)
+  if (!isMissingTabError(error))
     throw error;
   // otherwise, this error is caused from a tab already closed.
   // we just ignore it.
   //console.log('Invalid Tab ID error on: ' + error.stack);
 }
 
+export function isUnloadedError(error) {
+  return (
+    error &&
+    error.message &&
+    error.message.includes('can\'t access dead object')
+  );
+}
 export function handleUnloadedError(error) {
-  if (!error ||
-      !error.message ||
-      error.message.indexOf('can\'t access dead object') != 0)
+  if (!isUnloadedError(error))
+    throw error;
+}
+
+export function isMissingHostPermissionError(error) {
+  return (
+    error &&
+    error.message &&
+    error.message.includes('Missing host permission for the tab')
+  );
+}
+export function handleMissingHostPermissionError(error) {
+  if (!isMissingHostPermissionError(error))
     throw error;
 }
 
