@@ -5,6 +5,8 @@
 */
 'use strict';
 
+import { SequenceMatcher } from '/extlib/diff.js';
+
 import {
   log as internalLogger,
   dumpTab,
@@ -12,14 +14,16 @@ import {
   mapAndFilter,
   configs
 } from '/common/common.js';
-import * as Constants from '/common/constants.js';
 import * as ApiTabs from '/common/api-tabs.js';
+import * as CacheStorage from '/common/cache-storage.js';
+import * as Constants from '/common/constants.js';
+import * as TabsInternalOperation from '/common/tabs-internal-operation.js';
 import * as TabsStore from '/common/tabs-store.js';
 import * as TabsUpdate from '/common/tabs-update.js';
-import * as TabsInternalOperation from '/common/tabs-internal-operation.js';
-import * as MetricsData from '/common/metrics-data.js';
+import * as UniqueId from '/common/unique-id.js';
 
-import Tab from '/common/Tab.js';
+import MetricsData from '/common/MetricsData.js';
+import { Tab } from '/common/TreeItem.js';
 
 import * as Tree from './tree.js';
 
@@ -30,10 +34,21 @@ function log(...args) {
 const kCONTENTS_VERSION = 5;
 
 let mActivated = false;
+const mCaches = {};
 
 export function activate() {
   mActivated = true;
   configs.$addObserver(onConfigChange);
+
+  if (!configs.persistCachedTree) {
+    // clear obsolete cache
+    browser.windows.getAll().then(windows => {
+      for (const win of windows) {
+        browser.sessions.removeWindowValue(win.id, Constants.kWINDOW_STATE_CACHED_TABS).catch(ApiTabs.createErrorSuppressor());
+        browser.sessions.removeWindowValue(win.id, Constants.kWINDOW_STATE_CACHED_SIDEBAR_TABS_DIRTY).catch(ApiTabs.createErrorSuppressor());
+      }
+    });
+  }
 }
 
 
@@ -54,18 +69,18 @@ export async function restoreWindowFromEffectiveWindowCache(windowId, options =
   if (configs.debug)
     log(`restoreWindowFromEffectiveWindowCache for ${windowId} tabs: `, () => tabs.map(dumpTab));
   const actualSignature = getWindowSignature(tabs);
-  let cache = options.caches && options.caches.get(owner.id) || await MetricsData.addAsync('restoreWindowFromEffectiveWindowCache: window cache', getWindowCache(owner, Constants.kWINDOW_STATE_CACHED_TABS));
+  let cache = options.caches?.get(`window-${owner.windowId}`) || await MetricsData.addAsync('restoreWindowFromEffectiveWindowCache: window cache', getWindowCache(owner, Constants.kWINDOW_STATE_CACHED_TABS));
   if (!cache) {
     log(`restoreWindowFromEffectiveWindowCache for ${windowId} fail: no cache`);
     return false;
   }
   const promisedPermanentStates = Promise.all(tabs.map(tab => Tab.get(tab.id).$TST.getPermanentStates())); // don't await at here for better performance
   MetricsData.add('restoreWindowFromEffectiveWindowCache: validity check: start');
-  let cachedSignature = cache && cache.signature;
+  let cachedSignature = cache?.signature;
   log(`restoreWindowFromEffectiveWindowCache for ${windowId}: got from the owner `, {
     owner, cachedSignature, cache
   });
-  const signatureGeneratedFromCache = signatureFromTabsCache(cache.tabs).join('\n');
+  const signatureGeneratedFromCache = getWindowSignature(cache.tabs).join('\n');
   if (cache &&
       cache.tabs &&
       cachedSignature &&
@@ -89,40 +104,49 @@ export async function restoreWindowFromEffectiveWindowCache(windowId, options =
     cachedSignature = trimSignature(cachedSignature, cache.pinnedTabsCount);
   }
   MetricsData.add('restoreWindowFromEffectiveWindowCache: validity check: matching actual signature of got cache');
-  const signatureMatched = matcheSignatures({
+  const signatureMatchResult = matcheSignatures({
     actual: actualSignature,
     cached: cachedSignature
   });
   log(`restoreWindowFromEffectiveWindowCache for ${windowId}: verify cache`, {
-    cache, actualSignature, cachedSignature, signatureMatched
+    cache, actualSignature, cachedSignature,
+    ...signatureMatchResult,
   });
   if (!cache ||
       cache.version != kCONTENTS_VERSION ||
-      !signatureMatched) {
+      !signatureMatchResult.matched) {
     log(`restoreWindowFromEffectiveWindowCache for ${windowId}: no effective cache`);
     TabsInternalOperation.clearCache(owner);
     MetricsData.add('restoreWindowFromEffectiveWindowCache: validity check: actual signature failed.');
     return false;
   }
   MetricsData.add('restoreWindowFromEffectiveWindowCache: validity check: actual signature passed.');
-  cache.offset = actualSignature.length - cachedSignature.length;
+  cache.offset = signatureMatchResult.offset;
 
   log(`restoreWindowFromEffectiveWindowCache for ${windowId}: restore from cache`);
 
   const permanentStates = await MetricsData.addAsync('restoreWindowFromEffectiveWindowCache: permanentStatus', promisedPermanentStates); // await at here for better performance
   const restored = await MetricsData.addAsync('restoreWindowFromEffectiveWindowCache: restoreTabsFromCache', restoreTabsFromCache(windowId, { cache, tabs, permanentStates }));
-  if (restored)
+  if (restored) {
     MetricsData.add(`restoreWindowFromEffectiveWindowCache: window ${windowId} succeeded`);
-  else
+    // Now we reload the sidebar if it is opened, because it is the easiest way
+    // to synchronize state of tabs completely.
+    log('reload sidebar for a tree restored from cache');
+    browser.runtime.sendMessage({
+      type: Constants.kCOMMAND_RELOAD,
+      windowId,
+    }).catch(ApiTabs.createErrorSuppressor());
+  }
+  else {
     MetricsData.add(`restoreWindowFromEffectiveWindowCache: window ${windowId} failed`);
+  }
 
   log(`restoreWindowFromEffectiveWindowCache for ${windowId}: restored = ${restored}`);
   return restored;
 }
 
 function getWindowSignature(tabs) {
-  const tabIds = tabs.map(tab => tab.id);
-  return tabs.map(tab => `${tab.openerTabId ? tabIds.indexOf(tab.$TST && tab.$TST.parentId || tab.openerTabId) : -1 },${tab.cookieStoreId},${tab.incognito},${tab.pinned}`);
+  return tabs.map(tab => `${tab.cookieStoreId},${tab.incognito},${tab.pinned},${tab.url}`);
 }
 
 function trimSignature(signature, ignoreCount) {
@@ -138,16 +162,21 @@ function trimTabsCache(cache, ignoreCount) {
 }
 
 function matcheSignatures(signatures) {
-  return (
-    signatures.actual &&
-    signatures.cached &&
-    signatures.actual.slice(-signatures.cached.length).join('\n') == signatures.cached.join('\n')
-  );
-}
-
-function signatureFromTabsCache(cachedTabs) {
-  const cachedTabIds = cachedTabs.map(tab => tab.id);
-  return cachedTabs.map(tab => `${tab.$TST.parentId ? cachedTabIds.indexOf(tab.$TST.parentId) : -1 },${tab.cookieStoreId},${tab.incognito},${tab.pinned}`);
+  const operations = (new SequenceMatcher(signatures.cached, signatures.actual)).operations();
+  log('matcheSignatures: operations ', operations);
+  let matched = false;
+  let offset  = 0;
+  for (const operation of operations) {
+    const [tag, fromStart, fromEnd, toStart, toEnd] = operation;
+    if (tag == 'equal' &&
+        fromEnd - fromStart == signatures.cached.length) {
+      matched = true;
+      break;
+    }
+    offset += toEnd - toStart;
+  }
+  log('matcheSignatures: ', { matched, offset });
+  return { matched, offset };
 }
 
 
@@ -157,7 +186,7 @@ async function restoreTabsFromCache(windowId, params = {}) {
     return false;
 
   return (await restoreTabsFromCacheInternal({
-    windowId:     windowId,
+    windowId,
     tabs:         params.tabs,
     permanentStates: params.permanentStates,
     offset:       params.cache.offset || 0,
@@ -169,14 +198,14 @@ async function restoreTabsFromCacheInternal(params) {
   MetricsData.add('restoreTabsFromCacheInternal: start');
   log(`restoreTabsFromCacheInternal: restore tabs for ${params.windowId} from cache`);
   const offset = params.offset || 0;
-  const window = TabsStore.windows.get(params.windowId);
+  const win    = TabsStore.windows.get(params.windowId);
   const tabs   = params.tabs.slice(offset).map(tab => Tab.get(tab.id));
   if (offset > 0 &&
       tabs.length <= offset) {
     log('restoreTabsFromCacheInternal: missing window');
     return [];
   }
-  log(`restoreTabsFromCacheInternal: there is ${window.tabs.size} tabs`);
+  log(`restoreTabsFromCacheInternal: there is ${win.tabs.size} tabs`);
   if (params.cache.length != tabs.length) {
     log('restoreTabsFromCacheInternal: Mismatched number of restored tabs?');
     return [];
@@ -200,6 +229,7 @@ async function fixupTabsRestoredFromCache(tabs, permanentStates, cachedTabs) {
     throw new Error(`fixupTabsRestoredFromCache: Mismatched number of tabs restored from cache, tabs=${tabs.length}, cachedTabs=${cachedTabs.length}`);
   log('fixupTabsRestoredFromCache start ', () => ({ tabs: tabs.map(dumpTab), cachedTabs }));
   const idMap = new Map();
+  let remappedCount = 0;
   // step 1: build a map from old id to new id
   tabs = tabs.map((tab, index) => {
     const cachedTab = cachedTabs[index];
@@ -207,8 +237,12 @@ async function fixupTabsRestoredFromCache(tabs, permanentStates, cachedTabs) {
     tab = Tab.get(tab.id);
     log(`fixupTabsRestoredFromCache: remap ${oldId} => ${tab.id}`);
     idMap.set(oldId, tab);
+    if (oldId != tab.id)
+      remappedCount++;
     return tab;
   });
+  if (remappedCount && remappedCount < tabs.length)
+    throw new Error(`fixupTabsRestoredFromCache: not a window restoration, only ${remappedCount} tab(s) are restored (maybe restoration of closed tabs)`);
   MetricsData.add('fixupTabsRestoredFromCache: step 1 done.');
   // step 2: restore information of tabs
   // Do this from bottom to top, to reduce post operations for modified trees.
@@ -241,7 +275,7 @@ function fixupTabRestoredFromCache(tab, permanentStates, cachedTab, idMap) {
   log('fixupTabRestoredFromCache children: ', cachedTab.$TST.childIds);
   const childIds = mapAndFilter(cachedTab.$TST.childIds, oldId => {
     const tab = idMap.get(oldId);
-    return tab && tab.id || undefined;
+    return tab?.id || undefined;
   });
   tab.$TST.children = childIds;
   if (childIds.length > 0)
@@ -258,6 +292,12 @@ function fixupTabRestoredFromCache(tab, permanentStates, cachedTab, idMap) {
   else
     tab.$TST.removeAttribute(Constants.kPARENT);
   log('fixupTabRestoredFromCache parent: => ', tab.$TST.parentId);
+
+  if (tab.discarded) {
+    tab.$TST.addState(Constants.kTAB_STATE_PENDING);
+  }
+
+  tab.$TST.temporaryMetadata.set('treeStructureAlreadyRestoredFromSessionData', true);
 }
 
 function fixupTabRestoredFromCachePostProcess(tab) {
@@ -285,48 +325,69 @@ function fixupTabRestoredFromCachePostProcess(tab) {
 async function updateWindowCache(owner, key, value) {
   if (!owner)
     return;
-  if (value === undefined) {
+
+  if (configs.persistCachedTree) {
     try {
-      if (configs.storeCacheAsWindowValue)
-        return browser.sessions.removeWindowValue(owner.windowId, key).catch(ApiTabs.createErrorSuppressor());
-      return browser.sessions.removeTabValue(owner.id || owner, key).catch(ApiTabs.createErrorSuppressor(ApiTabs.handleMissingTabError));
+      if (value)
+        await CacheStorage.setValue({
+          windowId: owner.windowId,
+          key,
+          value,
+          store: CacheStorage.BACKGROUND,
+        });
+      else
+        await CacheStorage.deleteValue({
+          windowId: owner.windowId,
+          key,
+          store: CacheStorage.BACKGROUND,
+        });
+      return;
     }
-    catch(e) {
-      console.log(new Error('fatal error: failed to delete window cache'), e, owner, key, value);
-    }
-  }
-  else {
-    try {
-      if (configs.storeCacheAsWindowValue)
-        return browser.sessions.setWindowValue(owner.windowId, key, value).catch(ApiTabs.createErrorSuppressor());
-      return browser.sessions.setTabValue(owner.id || owner, key, value).catch(ApiTabs.createErrorSuppressor(ApiTabs.handleMissingTabError));
-    }
-    catch(e) {
-      console.log(new Error('fatal error: failed to update window cache'), e, owner, key, value);
+    catch(error) {
+      console.log(`BackgroundCache.updateWindowCache for ${owner.windowId}/${key} failed: `, error.message, error.stack, error);
     }
   }
+
+  const storageKey = `backgroundCache-${await UniqueId.ensureWindowId(owner.windowId)}-${key}`;
+  if (value)
+    mCaches[storageKey] = value;
+  else
+    delete mCaches[storageKey];
 }
 
 export function markWindowCacheDirtyFromTab(tab, akey) {
-  const window = TabsStore.windows.get(tab.windowId);
-  if (!window) // the window may be closed
+  const win = TabsStore.windows.get(tab.windowId);
+  if (!win) // the window may be closed
     return;
-  if (window.markWindowCacheDirtyFromTabTimeout)
-    clearTimeout(window.markWindowCacheDirtyFromTabTimeout);
-  window.markWindowCacheDirtyFromTabTimeout = setTimeout(() => {
-    window.markWindowCacheDirtyFromTabTimeout = null;
-    updateWindowCache(window.lastWindowCacheOwner, akey, true);
+  if (win.markWindowCacheDirtyFromTabTimeout)
+    clearTimeout(win.markWindowCacheDirtyFromTabTimeout);
+  win.markWindowCacheDirtyFromTabTimeout = setTimeout(() => {
+    win.markWindowCacheDirtyFromTabTimeout = null;
+    updateWindowCache(win.lastWindowCacheOwner, akey, true);
   }, 100);
 }
 
 async function getWindowCache(owner, key) {
-  if (configs.storeCacheAsWindowValue)
-    return browser.sessions.getWindowValue(owner.windowId, key).catch(ApiTabs.createErrorHandler());
-  return browser.sessions.getTabValue(owner.id, key).catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
+  if (configs.persistCachedTree) {
+    try {
+      const value = await CacheStorage.getValue({
+        windowId: owner.windowId,
+        key,
+        store: CacheStorage.BACKGROUND,
+      });
+      return value;
+    }
+    catch(error) {
+      console.log(`BackgroundCache.getWindowCache for ${owner.windowId}/${key} failed: `, error.message, error.stack, error);
+    }
+  }
+
+  const storageKey = `backgroundCache-${await UniqueId.ensureWindowId(owner.windowId)}-${key}`;
+  return mCaches[storageKey];
 }
 
 function getWindowCacheOwner(windowId) {
-  const tab = Tab.getFirstNormalTab(windowId) || Tab.getFirstTab(windowId);
+  const tab = Tab.getFirstTab(windowId);
   if (!tab)
     return null;
   return {
@@ -335,13 +396,13 @@ function getWindowCacheOwner(windowId) {
   };
 }
 
-export async function reserveToCacheTree(windowId) {
+export async function reserveToCacheTree(windowId, trigger) {
   if (!mActivated ||
       !configs.useCachedTree)
     return;
 
-  const window = TabsStore.windows.get(windowId);
-  if (!window)
+  const win = TabsStore.windows.get(windowId);
+  if (!win)
     return;
 
   // If there is any opening (but not resolved its unique id yet) tab,
@@ -351,45 +412,59 @@ export async function reserveToCacheTree(windowId) {
   if (Tab.needToWaitTracked(windowId))
     await Tab.waitUntilTrackedAll(windowId);
 
-  if (window.allTabsRestored)
+  if (win.promisedAllTabsRestored) // not restored yet
     return;
 
-  log('reserveToCacheTree for window ', windowId, { stack: configs.debug && new Error().stack });
-  TabsInternalOperation.clearCache(windowId.lastWindowCacheOwner);
+  if (!trigger && configs.debug)
+    trigger = new Error().stack;
 
-  if (window.waitingToCacheTree)
-    clearTimeout(window.waitingToCacheTree);
-  window.waitingToCacheTree = setTimeout(() => {
-    cacheTree(windowId);
+  log('reserveToCacheTree for window ', windowId, trigger);
+  TabsInternalOperation.clearCache(win.lastWindowCacheOwner);
+
+  if (trigger)
+    reserveToCacheTree.triggers.add(trigger);
+
+  if (win.waitingToCacheTree)
+    clearTimeout(win.waitingToCacheTree);
+  win.waitingToCacheTree = setTimeout(() => {
+    const triggers = [...reserveToCacheTree.triggers];
+    reserveToCacheTree.triggers.clear();
+    cacheTree(windowId, triggers);
   }, 500);
 }
+reserveToCacheTree.triggers = new Set();
 
 function cancelReservedCacheTree(windowId) {
-  const window = TabsStore.windows.get(windowId);
-  if (window && window.waitingToCacheTree) {
-    clearTimeout(window.waitingToCacheTree);
-    delete window.waitingToCacheTree;
+  const win = TabsStore.windows.get(windowId);
+  if (win?.waitingToCacheTree) {
+    clearTimeout(win.waitingToCacheTree);
+    delete win.waitingToCacheTree;
   }
 }
 
-async function cacheTree(windowId) {
+async function cacheTree(windowId, triggers) {
   if (Tab.needToWaitTracked(windowId))
     await Tab.waitUntilTrackedAll(windowId);
-  const window = TabsStore.windows.get(windowId);
-  if (!window ||
+  const win = TabsStore.windows.get(windowId);
+  if (!win ||
       !configs.useCachedTree)
     return;
   const signature = getWindowSignature(Tab.getAllTabs(windowId));
-  if (window.allTabsRestored)
+  if (win.promisedAllTabsRestored) // not restored yet
     return;
   //log('save cache for ', windowId);
-  window.lastWindowCacheOwner = getWindowCacheOwner(windowId);
-  if (!window.lastWindowCacheOwner)
+  win.lastWindowCacheOwner = getWindowCacheOwner(windowId);
+  if (!win.lastWindowCacheOwner)
     return;
-  log('cacheTree for window ', windowId, { stack: configs.debug && new Error().stack });
-  updateWindowCache(window.lastWindowCacheOwner, Constants.kWINDOW_STATE_CACHED_TABS, {
+  const firstTab = Tab.getFirstTab(windowId);
+  if (firstTab.incognito) { // never save cache for incognito windows
+    updateWindowCache(win.lastWindowCacheOwner, Constants.kWINDOW_STATE_CACHED_TABS, null);
+    return;
+  }
+  log('cacheTree for window ', windowId, triggers/*{ stack: configs.debug && new Error().stack }*/);
+  updateWindowCache(win.lastWindowCacheOwner, Constants.kWINDOW_STATE_CACHED_TABS, {
     version:         kCONTENTS_VERSION,
-    tabs:            TabsStore.windows.get(windowId).export(true),
+    tabs:            TabsStore.windows.get(windowId).export(true).tabs,
     pinnedTabsCount: Tab.getPinnedTabs(windowId).length,
     signature
   });
@@ -399,22 +474,20 @@ async function cacheTree(windowId) {
 // update cache on events
 
 Tab.onCreated.addListener((tab, _info = {}) => {
-  if (!tab.$TST.nextTab) { // new last tab is added, then the last cache owner is not the owner anymore
-    const window = TabsStore.windows.get(tab.windowId);
-    if (window.lastWindowCacheOwner)
-      TabsInternalOperation.clearCache(window.lastWindowCacheOwner);
+  if (!tab.$TST.previousTab) { // it is a new cache owner
+    const win = TabsStore.windows.get(tab.windowId);
+    if (win.lastWindowCacheOwner)
+      TabsInternalOperation.clearCache(win.lastWindowCacheOwner);
   }
-  reserveToCacheTree(tab.windowId);
+  reserveToCacheTree(tab.windowId, 'tab created');
 });
 
 // Tree restoration for "Restore Previous Session"
-Tab.onWindowRestoring.addListener(async windowId => {
+Tab.onWindowRestoring.addListener(async ({ windowId, restoredCount }) => {
   if (!configs.useCachedTree)
     return;
 
-  log('Tabs.onWindowRestoring ', windowId);
-  const window = TabsStore.windows.get(windowId);
-  const restoredCount = await window.allTabsRestored;
+  log('Tabs.onWindowRestoring ', { windowId, restoredCount });
   if (restoredCount == 1) {
     log('Tabs.onWindowRestoring: single tab restored');
     return;
@@ -424,90 +497,124 @@ Tab.onWindowRestoring.addListener(async windowId => {
   MetricsData.add('Tabs.onWindowRestoring restore start');
 
   const tabs = await browser.tabs.query({ windowId }).catch(ApiTabs.createErrorHandler());
+  let restoredFromCache = false;
   try {
-    await restoreWindowFromEffectiveWindowCache(windowId, {
+    restoredFromCache = await restoreWindowFromEffectiveWindowCache(windowId, {
       ignorePinnedTabs: true,
       owner: tabs[tabs.length - 1],
       tabs
     });
+    log('Tabs.onWindowRestoring: restoredFromCache = ', restoredFromCache);
     MetricsData.add('Tabs.onWindowRestoring restore end');
   }
   catch(e) {
     log('Tabs.onWindowRestoring: FATAL ERROR while restoring tree from cache', String(e), e.stack);
   }
+  if (!restoredFromCache) {
+    // We need to treat tabs for a restored window as pending even if we failed to restore tabs from cache.
+    // See https://github.com/piroor/treestyletab/issues/3794
+    log('Tabs.onWindowRestoring: failsafe: setting discarded tabs as pending...');
+    for (const tab of tabs) {
+      const trackedTab = Tab.get(tab.id);
+      if (trackedTab?.discarded) {
+        trackedTab.$TST.addState(Constants.kTAB_STATE_PENDING);
+      }
+    }
+  }
 });
 
 Tab.onRemoved.addListener((tab, info) => {
-  if (!tab.$TST.nextTab) // the tab was the cache owner
+  if (!tab.$TST.previousTab) // the tab was the cache owner
     TabsInternalOperation.clearCache(tab);
   wait(0).then(() => {
   // "Restore Previous Session" closes some tabs at first, so we should not clear the old cache yet.
   // See also: https://dxr.mozilla.org/mozilla-central/rev/5be384bcf00191f97d32b4ac3ecd1b85ec7b18e1/browser/components/sessionstore/SessionStore.jsm#3053
-    reserveToCacheTree(info.windowId);
+    reserveToCacheTree(info.windowId, 'tab removed');
   });
 });
 
 Tab.onMoved.addListener((tab, info) => {
-  if (tab.$TST.nextTab) // the tab is not the cache owner anymore
+  if (info.fromIndex == 0) // the tab is not the cache owner anymore
     TabsInternalOperation.clearCache(tab);
-  reserveToCacheTree(info.windowId);
+  reserveToCacheTree(info.windowId, 'tab moved');
 });
 
-Tab.onUpdated.addListener((tab, _info) => {
+Tab.onUpdated.addListener((tab, info) => {
   markWindowCacheDirtyFromTab(tab, Constants.kWINDOW_STATE_CACHED_SIDEBAR_TABS_DIRTY);
+  if ('url' in info)
+    reserveToCacheTree(tab.windowId, 'tab updated');
+});
+
+Tab.onStateChanged.addListener((tab, state, _has) => {
+  if (state == Constants.kTAB_STATE_STICKY)
+    markWindowCacheDirtyFromTab(tab, Constants.kWINDOW_STATE_CACHED_SIDEBAR_TABS_DIRTY);
 });
 
 Tree.onSubtreeCollapsedStateChanging.addListener(tab => {
-  reserveToCacheTree(tab.windowId);
+  reserveToCacheTree(tab.windowId, 'subtree collapsed/expanded');
 });
 
 Tree.onAttached.addListener((tab, _info) => {
   wait(0).then(() => {
     // "Restore Previous Session" closes some tabs at first and it causes tree changes, so we should not clear the old cache yet.
     // See also: https://dxr.mozilla.org/mozilla-central/rev/5be384bcf00191f97d32b4ac3ecd1b85ec7b18e1/browser/components/sessionstore/SessionStore.jsm#3053
-    reserveToCacheTree(tab.windowId);
+    reserveToCacheTree(tab.windowId, 'tab attached to tree');
   });
 });
 
 Tree.onDetached.addListener((tab, _info) => {
-  if (!tab.$TST.nextTab) // the tab was the cache owner
-    TabsInternalOperation.clearCache(tab);
+  TabsInternalOperation.clearCache(tab);
   wait(0).then(() => {
     // "Restore Previous Session" closes some tabs at first and it causes tree changes, so we should not clear the old cache yet.
     // See also: https://dxr.mozilla.org/mozilla-central/rev/5be384bcf00191f97d32b4ac3ecd1b85ec7b18e1/browser/components/sessionstore/SessionStore.jsm#3053
-    reserveToCacheTree(tab.windowId);
+    reserveToCacheTree(tab.windowId, 'tab detached from tree');
   });
 });
 
 Tab.onPinned.addListener(tab => {
-  if (!tab.$TST.nextTab) // the tab was the cache owner
-    TabsInternalOperation.clearCache(tab);
-  reserveToCacheTree(tab.windowId);
+  reserveToCacheTree(tab.windowId, 'tab pinned');
 });
 
 Tab.onUnpinned.addListener(tab => {
-  reserveToCacheTree(tab.windowId);
+  if (tab.$TST.previousTab) // the tab was the cache owner
+    TabsInternalOperation.clearCache(tab);
+  reserveToCacheTree(tab.windowId, 'tab unpinned');
 });
 
 Tab.onShown.addListener(tab => {
-  reserveToCacheTree(tab.windowId);
+  reserveToCacheTree(tab.windowId, 'tab shown');
 });
 
 Tab.onHidden.addListener(tab => {
-  reserveToCacheTree(tab.windowId);
+  reserveToCacheTree(tab.windowId, 'tab hidden');
+});
+
+browser.windows.onRemoved.addListener(async windowId => {
+  try {
+    CacheStorage.clearForWindow(windowId);
+  }
+  catch(_error) {
+  }
+
+  const storageKeyPart = `Cache-${await UniqueId.ensureWindowId(windowId)}-`;
+  for (const key in mCaches) {
+    if (key.includes(storageKeyPart))
+      delete mCaches[key];
+  }
 });
 
 function onConfigChange(key) {
   switch (key) {
     case 'useCachedTree':
+    case 'persistCachedTree':
       browser.windows.getAll({
         populate:    true,
         windowTypes: ['normal']
       }).then(windows => {
-        for (const window of windows) {
-          const owner = window.tabs[window.tabs.length - 1];
+        for (const win of windows) {
+          const owner = win.tabs[win.tabs.length - 1];
           if (configs[key]) {
-            reserveToCacheTree(window.id);
+            reserveToCacheTree(win.id, 'config change');
           }
           else {
             TabsInternalOperation.clearCache(owner);
