@@ -7,26 +7,28 @@
 
 import {
   log as internalLogger,
-  configs
+  configs,
 } from '/common/common.js';
 
 import * as ApiTabs from '/common/api-tabs.js';
-import * as TSTAPI from '/common/tst-api.js';
 import * as Bookmark from '/common/bookmark.js';
-import * as TabContextMenu from './tab-context-menu.js';
+import * as Sync from '/common/sync.js';
+import * as TSTAPI from '/common/tst-api.js';
 
-import Tab from '/common/Tab.js';
+import { Tab } from '/common/TreeItem.js';
 
-import * as TabsGroup from './tabs-group.js';
 import * as Commands from './commands.js';
+import * as TabContextMenu from './tab-context-menu.js';
 
 function log(...args) {
   internalLogger('background/context-menu', ...args);
 }
 
-export function init() {
-  addTabItems();
-  addBookmarkItems();
+export async function init() {
+  return Promise.all([
+    addTabItems(),
+    addBookmarkItems(),
+  ]);
 }
 
 const SAFE_CREATE_PROPERTIES = [
@@ -66,6 +68,30 @@ const mTabItemsById = {
     title:              browser.i18n.getMessage('context_reloadDescendants_label'),
     titleMultiselected: browser.i18n.getMessage('context_reloadDescendants_label_multiselected')
   },
+  // This item won't be handled by the onClicked handler, so you may need to handle it with something experiments API.
+  'unblockAutoplayTree': {
+    title:                browser.i18n.getMessage('context_unblockAutoplayTree_label'),
+    titleMultiselected:   browser.i18n.getMessage('context_unblockAutoplayTree_label_multiselected'),
+    requireAutoplayBlockedTab: true,
+  },
+  // This item won't be handled by the onClicked handler, so you may need to handle it with something experiments API.
+  'unblockAutoplayDescendants': {
+    title:                browser.i18n.getMessage('context_unblockAutoplayDescendants_label'),
+    titleMultiselected:   browser.i18n.getMessage('context_unblockAutoplayDescendants_label_multiselected'),
+    requireAutoplayBlockedDescendant: true,
+  },
+  'toggleMuteTree': {
+    titleMuteTree:                browser.i18n.getMessage('context_toggleMuteTree_label_mute'),
+    titleMultiselectedMuteTree:   browser.i18n.getMessage('context_toggleMuteTree_label_multiselected_mute'),
+    titleUnmuteTree:              browser.i18n.getMessage('context_toggleMuteTree_label_unmute'),
+    titleMultiselectedUnmuteTree: browser.i18n.getMessage('context_toggleMuteTree_label_multiselected_unmute')
+  },
+  'toggleMuteDescendants': {
+    titleMuteDescendant:                browser.i18n.getMessage('context_toggleMuteDescendants_label_mute'),
+    titleMultiselectedMuteDescendant:   browser.i18n.getMessage('context_toggleMuteDescendants_label_multiselected_mute'),
+    titleUnmuteDescendant:              browser.i18n.getMessage('context_toggleMuteDescendants_label_unmute'),
+    titleMultiselectedUnmuteDescendant: browser.i18n.getMessage('context_toggleMuteDescendants_label_multiselected_unmute')
+  },
   'separatorAfterReload': {
     type: 'separator'
   },
@@ -85,6 +111,16 @@ const mTabItemsById = {
   'separatorAfterClose': {
     type: 'separator'
   },
+  'toggleSticky': {
+    titleStick:                browser.i18n.getMessage('context_toggleSticky_label_stick'),
+    titleMultiselectedStick:   browser.i18n.getMessage('context_toggleSticky_label_multiselected_stick'),
+    titleUnstick:              browser.i18n.getMessage('context_toggleSticky_label_unstick'),
+    titleMultiselectedUnstick: browser.i18n.getMessage('context_toggleSticky_label_multiselected_unstick'),
+    requireNormal: true,
+  },
+  'separatorAfterToggleSticky': {
+    type: 'separator'
+  },
   'collapseTree': {
     title:              browser.i18n.getMessage('context_collapseTree_label'),
     titleMultiselected: browser.i18n.getMessage('context_collapseTree_label_multiselected'),
@@ -120,9 +156,9 @@ const mTabItemsById = {
     title:              browser.i18n.getMessage('context_bookmarkTree_label'),
     titleMultiselected: browser.i18n.getMessage('context_bookmarkTree_label_multiselected')
   },
-  'groupTabs': {
-    title:                browser.i18n.getMessage('context_groupTabs_label'),
-    requireMultiselected: true
+  'sendTreeToDevice': {
+    title:              browser.i18n.getMessage('context_sendTreeToDevice_label'),
+    titleMultiselected: browser.i18n.getMessage('context_sendTreeToDevice_label_multiselected')
   },
   'separatorAfterBookmark': {
     type: 'separator'
@@ -143,6 +179,7 @@ const mTabItemsById = {
 };
 const mTabItems = [];
 const mGroupedTabItems = [];
+const mGroupedTabItemsById = {};
 for (const id of Object.keys(mTabItemsById)) {
   const item = mTabItemsById[id];
   item.id = id;
@@ -151,18 +188,21 @@ for (const id of Object.keys(mTabItemsById)) {
   item.enabled = true;
   // Access key is not supported by WE API.
   // See also: https://bugzilla.mozilla.org/show_bug.cgi?id=1320462
-  item.titleWithoutAccesskey = item.title && item.title.replace(/\(&[a-z]\)|&([a-z])/i, '$1');
-  item.titleMultiselectedWithoutAccesskey = item.titleMultiselected && item.titleMultiselected.replace(/\(&[a-z]\)|&([a-z])/i, '$1');
+  item.titleWithoutAccesskey = item.title?.replace(/\(&[a-z]\)|&([a-z])/i, '$1');
+  item.titleMultiselectedWithoutAccesskey = item.titleMultiselected?.replace(/\(&[a-z]\)|&([a-z])/i, '$1');
   item.type = item.type || 'normal';
   item.contexts = ['tab'];
   item.lastVisible = item.visible = false;
   item.lastTitle = item.title;
   mTabItems.push(item);
-  mGroupedTabItems.push({
+
+  const groupedItem = {
     ...item,
     id:       `grouped:${id}`,
     parentId: kROOT_TAB_ITEM
-  });
+  };
+  mGroupedTabItems.push(groupedItem);
+  mGroupedTabItemsById[groupedItem.id] = groupedItem;
 }
 
 const mTabSeparator = {
@@ -192,25 +232,27 @@ const mAllTabItems = [
 ];
 
 function addTabItems() {
+  const promises = [];
   if (addTabItems.done) {
     for (const item of mAllTabItems) {
-      browser.menus.remove(item.id);
+      promises.push(browser.menus.remove(item.id));
     }
   }
 
   for (const item of mAllTabItems) {
     const params = getSafeCreateParams(item);
-    browser.menus.create(params);
+    promises.push(browser.menus.create(params));
     if (item.id == mTabSeparator.id ||
         addTabItems.done)
       continue;
-    TabContextMenu.onMessageExternal({
+    promises.push(TabContextMenu.onMessageExternal({
       type: TSTAPI.kCONTEXT_MENU_CREATE,
       params
-    }, browser.runtime);
+    }, browser.runtime));
   }
 
   addTabItems.done = true;
+  return Promise.all(promises);
 }
 addTabItems.done = false;
 
@@ -271,15 +313,17 @@ const mAllBookmarkItems = [
 ];
 
 function addBookmarkItems() {
+  const promises = [];
   if (addBookmarkItems.done) {
     for (const item of mAllBookmarkItems) {
-      browser.menus.remove(item.id);
+      promises.push(browser.menus.remove(item.id));
     }
   }
   for (const item of mAllBookmarkItems) {
-    browser.menus.create(getSafeCreateParams(item));
+    promises.push(browser.menus.create(getSafeCreateParams(item)));
   }
   addBookmarkItems.done = true;
+  return Promise.all(promises);
 }
 addBookmarkItems.done = false;
 
@@ -289,17 +333,18 @@ addBookmarkItems.done = false;
 TabContextMenu.onTopLevelItemAdded.addListener(reserveToRefreshItems);
 
 function reserveToRefreshItems() {
-  if (reserveToRefreshItems.reserved)
-    clearTimeout(reserveToRefreshItems.reserved);
-  reserveToRefreshItems.reserved = setTimeout(() => {
-    reserveToRefreshItems.reserved = null;
+  if (reserveToRefreshItems.invoked)
+    return;
+  reserveToRefreshItems.invoked = true;
+  setTimeout(() => { // because window.requestAnimationFrame is decelerate for an invisible document.
+    reserveToRefreshItems.invoked = false;
     addTabItems();
     addBookmarkItems();
-  }, 100);
+  }, 0);
 }
-reserveToRefreshItems.reserved = null;
 
 function updateItem(id, params) {
+  log('updateItem ', id, params);
   browser.menus.update(id, params).catch(ApiTabs.createErrorSuppressor());
   TabContextMenu.onMessageExternal({
     type:   TSTAPI.kCONTEXT_MENU_UPDATE,
@@ -307,7 +352,8 @@ function updateItem(id, params) {
   }, browser.runtime);
 }
 
-function updateItemsVisibility(items, { forceVisible = null, multiselected = false } = {}) {
+function updateItemsVisibility(items, { forceVisible = null, multiselected = false, hasUnmutedTab = false, hasUnmutedDescendant = false, hasAutoplayBlockedTab = false, hasAutoplayBlockedDescendant = false, sticky = false, hidden = false } = {}) {
+  log('updateItemsVisibility ', items, { forceVisible, multiselected, hasUnmutedTab, hasUnmutedDescendant, hasAutoplayBlockedTab, hasAutoplayBlockedDescendant, sticky });
   let updated = false;
   let visibleItemsCount = 0;
   let visibleNormalItemsCount = 0;
@@ -324,11 +370,19 @@ function updateItemsVisibility(items, { forceVisible = null, multiselected = fal
       lastSeparator = item;
     }
     else {
-      const title = multiselected && item.titleMultiselected || item.title;
-      let visible = !(item.configKey in configs) || configs[item.configKey];
-      if (forceVisible !== null)
+      const title = Commands.getMenuItemTitle(item, { multiselected, hasUnmutedTab, hasUnmutedDescendant, sticky });
+      let visible = !hidden && (!(item.configKey in configs) || configs[item.configKey] || false);
+      log('checking ', item.id, {
+        config: visible,
+        multiselected: item.hideOnMultiselected && multiselected,
+        lastVisible: item.lastVisible,
+        forceVisible,
+      });
+      if (forceVisible !== null && forceVisible !== undefined)
         visible = forceVisible;
-      if (item.hideOnMultiselected && multiselected)
+      if ((item.hideOnMultiselected && multiselected) ||
+          (item.requireAutoplayBlockedTab && !hasAutoplayBlockedTab) ||
+          (item.requireAutoplayBlockedDescendant && !hasAutoplayBlockedDescendant))
         visible = false;
       if (visible) {
         if (lastSeparator) {
@@ -356,7 +410,7 @@ function updateItemsVisibility(items, { forceVisible = null, multiselected = fal
       updated = true;
     }
   }
-  if (lastSeparator && lastSeparator.lastVisible) {
+  if (lastSeparator?.lastVisible) {
     updateItem(lastSeparator.id, { visible: false });
     lastSeparator.lastVisible = false;
     updated = true;
@@ -364,31 +418,37 @@ function updateItemsVisibility(items, { forceVisible = null, multiselected = fal
   return { updated, visibleItemsCount };
 }
 
-function updateItems({ multiselected } = {}) {
+async function updateItems({ multiselected, hasUnmutedTab, hasUnmutedDescendant, hasAutoplayBlockedTab, hasAutoplayBlockedDescendant, sticky, hidden } = {}) {
   let updated = false;
 
-  const groupedItems = updateItemsVisibility(mGroupedTabItems, { multiselected });
+  const groupedItems = updateItemsVisibility(mGroupedTabItems, { multiselected, hasUnmutedTab, hasUnmutedDescendant, hasAutoplayBlockedTab, hasAutoplayBlockedDescendant, sticky, hidden });
   if (groupedItems.updated)
     updated = true;
 
-  const separatorVisible = configs.emulateDefaultContextMenu && groupedItems.visibleItemsCount > 0;
+  const separatorVisible = !hidden && configs.emulateDefaultContextMenu && groupedItems.visibleItemsCount > 0;
   if (separatorVisible != mTabSeparator.lastVisible) {
     updateItem(mTabSeparator.id, { visible: separatorVisible });
     mTabSeparator.lastVisible = separatorVisible;
     updated = true;
   }
 
-  const grouped = configs.emulateDefaultContextMenu && groupedItems.visibleItemsCount > 1;
+  const grouped = !hidden && configs.emulateDefaultContextMenu && groupedItems.visibleItemsCount > 1;
   if (grouped != mTabRootItem.lastVisible) {
     updateItem(mTabRootItem.id, { visible: grouped });
     mTabRootItem.lastVisible = grouped;
     updated = true;
   }
 
-  const topLevelItems = updateItemsVisibility(mTabItems, { forceVisible: grouped ? false : null, multiselected });
+  const topLevelItems = updateItemsVisibility(mTabItems, { forceVisible: grouped ? false : null, multiselected, hasUnmutedTab, hasUnmutedDescendant, hasAutoplayBlockedTab, hasAutoplayBlockedDescendant, sticky, hidden });
   if (topLevelItems.updated)
     updated = true;
 
+  if (mGroupedTabItemsById['grouped:sendTreeToDevice'].lastVisible &&
+      await TabContextMenu.updateSendToDeviceItems('grouped:sendTreeToDevice', {
+        manage: navigator.userAgent.includes('Fennec'), // see also https://github.com/piroor/treestyletab/issues/3174
+      }))
+    updated = true;
+
   return updated;
 }
 
@@ -430,6 +490,19 @@ function onTabItemClick(info, tab) {
         Commands.reloadDescendants(contextTabs);
       break;
 
+    case 'toggleMuteTree':
+      if (inverted)
+        Commands.toggleMuteDescendants(contextTabs);
+      else
+        Commands.toggleMuteTree(contextTabs);
+      break;
+    case 'toggleMuteDescendants':
+      if (inverted)
+        Commands.toggleMuteTree(contextTabs);
+      else
+        Commands.toggleMuteDescendants(contextTabs);
+      break;
+
     case 'closeTree':
       if (inverted)
         Commands.closeDescendants(contextTabs);
@@ -446,6 +519,10 @@ function onTabItemClick(info, tab) {
       Commands.closeOthers(contextTabs);
       break;
 
+    case 'toggleSticky':
+      Commands.toggleSticky(contextTabs, !contextTab.$TST.sticky);
+      break;
+
     case 'collapseTree':
       Commands.collapseTree(contextTabs, { recursively: inverted });
       break;
@@ -469,9 +546,8 @@ function onTabItemClick(info, tab) {
       Commands.bookmarkTree(contextTabs);
       break;
 
-    case 'groupTabs':
-      if (contextTabs.length > 1)
-        TabsGroup.groupTabs(contextTabs, { broadcast: true });
+    case 'sendTreeToDevice:all':
+      Sync.sendTabsToAllDevices(contextTabs, { recursively: true });
       break;
 
     case 'collapsed':
@@ -493,8 +569,16 @@ function onTabItemClick(info, tab) {
           .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
     }; break;
 
-    default:
-      break;
+    default: {
+      const sendToDeviceMatch = info.menuItemId.match(/^sendTreeToDevice:device:(.+)$/);
+      if (contextTab &&
+          sendToDeviceMatch)
+        Sync.sendTabsToDevice(
+          contextTabs,
+          { to: sendToDeviceMatch[1],
+            recursively: true }
+        );
+    }; break;
   }
 }
 TabContextMenu.onTSTItemClick.addListener(onTabItemClick);
@@ -511,26 +595,52 @@ async function onBookmarkItemClick(info) {
   }
 }
 
-function onShown(info, tab) {
+async function onShown(info, tab) {
   if (info.contexts.includes('tab'))
-    onTabContextMenuShown(info, tab);
+    await onTabContextMenuShown(info, tab);
   else if (info.contexts.includes('bookmark'))
     onBookmarkContextMenuShown(info);
 }
 browser.menus.onShown.addListener(onShown);
 
-function onTabContextMenuShown(info, tab) {
-  tab = tab && Tab.get(tab.id);
-  const multiselected = tab && tab.$TST.multiselected;
-  const contextTabs      = multiselected ? Tab.getSelectedTabs(tab.windowId) : tab ? [tab] : [];
-  const hasChild         = contextTabs.some(tab => tab.$TST.hasChild);
-  const subtreeCollapsed = contextTabs.some(tab => tab.$TST.subtreeCollapsed);
+let mLastContextTabId = null;
+async function onTabContextMenuShown(info, tab) {
+  const contextTabId = tab?.id;
+  mLastContextTabId = contextTabId;
 
-  let updated = updateItems({ multiselected });
+  tab = tab && Tab.get(contextTabId);
+  const multiselected = tab?.$TST.multiselected;
+  const contextTabs      = multiselected ? Tab.getSelectedTabs(tab.windowId) : tab ? [tab] : [];
+  const hasChild         = contextTabs.length > 0 && contextTabs.some(tab => tab.$TST.hasChild);
+  const subtreeCollapsed = contextTabs.length > 0 && contextTabs.some(tab => tab.$TST.subtreeCollapsed);
+  const grouped          = contextTabs.length > 0 && contextTabs.some(tab => tab.$TST.isGroupTab);
+  const { hasUnmutedTab, hasUnmutedDescendant } = Commands.getUnmutedState(contextTabs);
+  const { hasAutoplayBlockedTab, hasAutoplayBlockedDescendant } = Commands.getAutoplayBlockedState(contextTabs);
+
+  let updated = await updateItems({
+    multiselected,
+    hasUnmutedTab,
+    hasUnmutedDescendant,
+    hasAutoplayBlockedTab,
+    hasAutoplayBlockedDescendant,
+    sticky: tab?.$TST.sticky,
+    hidden: !configs.showTreeCommandsInTabsContextMenuGlobally && info.viewType != 'sidebar',
+  });
+  if (mLastContextTabId != contextTabId)
+    return; // Skip further operations if the menu was already reopened on a different context tab.
 
   for (const item of mTabItems) {
+    let newVisible;
     let newEnabled;
-    if (item.requireTree) {
+    if (item.id == 'sendTreeToDevice' &&
+        item.visible) {
+      newVisible = contextTabs.filter(Sync.isSendableTab).length > 0;
+      newEnabled = (
+        hasChild &&
+        Sync.getOtherDevices().length > 0
+      );
+    }
+    else if (item.requireTree) {
       newEnabled = hasChild;
       switch (item.id) {
         case 'collapseTree':
@@ -546,16 +656,29 @@ function onTabContextMenuShown(info, tab) {
     else if (item.requireMultiselected) {
       newEnabled = multiselected;
     }
+    else if (item.requireGrouped) {
+      newEnabled = grouped;
+    }
+    else if (item.requireNormal) {
+      newEnabled = tab?.pinned;
+    }
     else {
       continue;
     }
 
-    if (newEnabled == !!item.enabled)
+    if ((newVisible === undefined ||
+         newVisible == !!item.visible) &&
+        (newEnabled === undefined ||
+         newEnabled == !!item.enabled))
       continue;
 
-    const params = {
-      enabled: item.enabled = newEnabled
-    };
+    const params = {};
+    if (newVisible !== undefined &&
+        newVisible != !!item.visible)
+      params.visible = item.visible = !!newVisible;
+    if (newEnabled !== undefined &&
+        newEnabled != !!item.enabled)
+      params.enabled = item.enabled = !!newEnabled;
 
     updateItem(item.id, params);
     updateItem(`grouped:${item.id}`, params);
@@ -578,10 +701,16 @@ function onTabContextMenuShown(info, tab) {
 }
 TabContextMenu.onTSTTabContextMenuShown.addListener(onTabContextMenuShown);
 
+let mLastContextItemId = null;
 async function onBookmarkContextMenuShown(info) {
+  const contextItemId = info.bookmarkId;
+  mLastContextItemId = contextItemId;
+
   let isFolder = true;
   if (info.bookmarkId) {
     const item = await Bookmark.getItemById(info.bookmarkId);
+    if (mLastContextItemId != contextItemId)
+      return; // Skip further operations if the menu was already reopened on a different context item.
     isFolder = (
       item.type == 'folder' ||
       (item.type == 'bookmark' &&
@@ -612,7 +741,7 @@ async function onBookmarkContextMenuShown(info) {
   }
   for (const item of [...mBookmarkItems, ...mGroupedBookmarkItems]) {
     browser.menus.update(item.id, {
-      visible: item.visible
+      visible: !!item.visible,
     });
   }
 
@@ -624,3 +753,12 @@ async function onBookmarkContextMenuShown(info) {
   });
   browser.menus.refresh().catch(ApiTabs.createErrorSuppressor());
 }
+
+
+export function getItemIdsWithIcon() {
+  return [
+    kROOT_TAB_ITEM,
+    kROOT_BOOKMARK_ITEM,
+    ...Object.keys(mBookmarkItemsById),
+  ];
+}
