@@ -9,36 +9,36 @@ import {
   log as internalLogger,
   wait,
   mapAndFilterUniq,
-  configs
+  configs,
+  loadUserStyleRules,
+  doProgressively,
 } from '/common/common.js';
-
-import * as Constants from '/common/constants.js';
 import * as ApiTabs from '/common/api-tabs.js';
-import * as TabsStore from '/common/tabs-store.js';
-import * as TabsUpdate from '/common/tabs-update.js';
-import * as TabsInternalOperation from '/common/tabs-internal-operation.js';
-import * as TSTAPI from '/common/tst-api.js';
-import * as SidebarConnection from '/common/sidebar-connection.js';
-import * as Permissions from '/common/permissions.js';
-import * as TreeBehavior from '/common/tree-behavior.js';
 import * as Bookmark from '/common/bookmark.js';
 import * as BrowserTheme from '/common/browser-theme.js';
+import * as Constants from '/common/constants.js';
 import * as ContextualIdentities from '/common/contextual-identities.js';
+import * as Permissions from '/common/permissions.js';
+import * as SidebarConnection from '/common/sidebar-connection.js';
+import * as TabsInternalOperation from '/common/tabs-internal-operation.js';
+import * as TabsStore from '/common/tabs-store.js';
+import * as TabsUpdate from '/common/tabs-update.js';
+import * as TreeBehavior from '/common/tree-behavior.js';
+import * as TSTAPI from '/common/tst-api.js';
 
-import Tab from '/common/Tab.js';
+import { Tab, TreeItem } from '/common/TreeItem.js';
 
 import * as Background from './background.js';
-import * as TabsGroup from './tabs-group.js';
-import * as Tree from './tree.js';
 import * as Commands from './commands.js';
 import * as Migration from './migration.js';
+import * as TabsGroup from './tabs-group.js';
+import * as TabsOpen from './tabs-open.js';
+import * as Tree from './tree.js';
 
 function log(...args) {
   internalLogger('background/handle-misc', ...args);
 }
 
-let mInitialized = false;
-
 
 /* message observer */
 
@@ -47,46 +47,78 @@ let mInitialized = false;
 // and BackgroundConnection, because making listeners asynchornous (async
 // functions) will break things - those listeners must not return Promise for
 // unneeded cases.
+// So we simply ignore messages delivered before completely initialized, for now.
 // See also: https://github.com/piroor/treestyletab/issues/2200
 
+const PHASE_LOADING                = 0;
+const PHASE_BACKGROUND_INITIALIZED = 1;
+const PHASE_BACKGROUND_BUILT       = 2;
+const PHASE_BACKGROUND_READY       = 3;
+let mInitializationPhase = PHASE_LOADING;
+
 Background.onInit.addListener(() => {
-  browser.browserAction.onClicked.addListener(onToolbarButtonClick);
-  browser.commands.onCommand.addListener(onShortcutCommand);
-  TSTAPI.onMessageExternal.addListener(onMessageExternal);
+  mInitializationPhase = PHASE_BACKGROUND_INITIALIZED;
 });
 
 Background.onBuilt.addListener(() => {
-  log('Start listening of messages');
-  browser.runtime.onMessage.addListener(onMessage);
+  mInitializationPhase = PHASE_BACKGROUND_BUILT;
 });
 
 Background.onReady.addListener(() => {
-  mInitialized = true;
+  mInitializationPhase = PHASE_BACKGROUND_READY;
+});
+
+if (browser.sidebarAction)
+  (browser.action || browser.browserAction)?.onClicked.addListener(onToolbarButtonClick);
+browser.commands.onCommand.addListener(onShortcutCommand);
+browser.runtime.onMessage.addListener(onMessage);
+TSTAPI.onMessageExternal.addListener(onMessageExternal);
+
+
+Background.onReady.addListener(() => {
   Bookmark.startTracking();
 });
 
 Background.onDestroy.addListener(() => {
   browser.runtime.onMessage.removeListener(onMessage);
   TSTAPI.onMessageExternal.removeListener(onMessageExternal);
-  browser.browserAction.onClicked.removeListener(onToolbarButtonClick);
+  if (browser.sidebarAction)
+    (browser.action || browser.browserAction)?.onClicked.removeListener(onToolbarButtonClick);
 });
 
 
 function onToolbarButtonClick(tab) {
-  if (Permissions.requestPostProcess())
+  if (mInitializationPhase < PHASE_BACKGROUND_INITIALIZED ||
+      Permissions.requestPostProcess()) {
     return;
+  }
 
-  if (SidebarConnection.isSidebarOpen(tab.windowId))
+  if (Migration.isInitialStartup()) {
+    Migration.openInitialStartupPage();
+    return;
+  }
+
+  if (typeof browser.sidebarAction.toggle == 'function')
+    browser.sidebarAction.toggle();
+  else if (SidebarConnection.isSidebarOpen(tab.windowId))
     browser.sidebarAction.close();
   else
     browser.sidebarAction.open();
 }
 
 async function onShortcutCommand(command) {
-  const activeTab = Tab.get((await browser.tabs.query({
+  if (mInitializationPhase < PHASE_BACKGROUND_INITIALIZED)
+    return;
+
+  let activeTabs = command.tab ? [command.tab] : await browser.tabs.query({
     active:        true,
-    currentWindow: true
-  }).catch(ApiTabs.createErrorHandler()))[0].id);
+    currentWindow: true,
+  }).catch(ApiTabs.createErrorHandler());
+  if (activeTabs.length == 0)
+    activeTabs = await browser.tabs.query({
+      currentWindow: true,
+    }).catch(ApiTabs.createErrorHandler());
+  const activeTab = Tab.get(activeTabs[0].id);
   const selectedTabs = activeTab.$TST.multiselected ? Tab.getSelectedTabs(activeTab.windowId) : [activeTab];
   log('onShortcutCommand ', { command, activeTab, selectedTabs });
 
@@ -100,6 +132,12 @@ async function onShortcutCommand(command) {
     case 'reloadDescendants':
       Commands.reloadDescendants(selectedTabs);
       return;
+    case 'toggleMuteTree':
+      Commands.toggleMuteTree(selectedTabs);
+      return;
+    case 'toggleMuteDescendants':
+      Commands.toggleMuteDescendants(selectedTabs);
+      return;
     case 'closeTree':
       Commands.closeTree(selectedTabs);
       return;
@@ -109,6 +147,9 @@ async function onShortcutCommand(command) {
     case 'closeOthers':
       Commands.closeOthers(selectedTabs);
       return;
+    case 'toggleSticky':
+      Commands.toggleSticky(selectedTabs);
+      return;
     case 'collapseTree':
       Commands.collapseTree(selectedTabs);
       return;
@@ -143,6 +184,18 @@ async function onShortcutCommand(command) {
         as:      Constants.kNEWTAB_OPEN_AS_CHILD
       });
       return;
+    case 'newChildTabTop':
+      Commands.openNewTabAs({
+        baseTab: activeTab,
+        as:      Constants.kNEWTAB_OPEN_AS_CHILD_TOP
+      });
+      return;
+    case 'newChildTabEnd':
+      Commands.openNewTabAs({
+        baseTab: activeTab,
+        as:      Constants.kNEWTAB_OPEN_AS_CHILD_END
+      });
+      return;
     case 'newSiblingTab':
       Commands.openNewTabAs({
         baseTab: activeTab,
@@ -163,18 +216,6 @@ async function onShortcutCommand(command) {
       });
       return;
 
-    case 'groupSelectedTabs':
-      if (selectedTabs.length > 1)
-        TabsGroup.groupTabs(selectedTabs, { broadcast: true });
-      return;
-
-    case 'indent':
-      Commands.indent(activeTab, { followChildren: true });
-      return;
-    case 'outdent':
-      Commands.outdent(activeTab, { followChildren: true });
-      return;
-
     case 'tabMoveUp':
       Commands.moveUp(activeTab, { followChildren: false });
       return;
@@ -189,35 +230,151 @@ async function onShortcutCommand(command) {
       return;
 
     case 'focusPrevious':
-    case 'focusPreviousSilently': {
-      const nextActive = activeTab.$TST.nearestVisiblePrecedingTab ||
-        Tab.getLastVisibleTab(activeTab.windowId);
-      TabsInternalOperation.activateTab(nextActive, {
-        silently: /Silently/.test(command)
-      });
-    }; return;
+      focusPrevious(activeTab);
+      return;
+    case 'focusPreviousSilently':
+      focusPreviousSilently(activeTab);
+      return;
     case 'focusNext':
-    case 'focusNextSilently': {
-      const nextActive = activeTab.$TST.nearestVisibleFollowingTab ||
-        Tab.getFirstVisibleTab(activeTab.windowId);
-      TabsInternalOperation.activateTab(nextActive, {
-        silently: /Silently/.test(command)
-      });
-    }; return;
+      focusNext(activeTab);
+      return;
+    case 'focusNextSilently':
+      focusNextSilently(activeTab);
+      return;
     case 'focusParent':
       TabsInternalOperation.activateTab(activeTab.$TST.parent);
       return;
+    case 'focusParentOrCollapse':
+      collapseOrFocusToParent(activeTab);
+      return;
     case 'focusFirstChild':
       TabsInternalOperation.activateTab(activeTab.$TST.firstChild);
       return;
+    case 'focusFirstChildOrExpand':
+      expandOrFocusToFirstChild(activeTab);
+      return;
     case 'focusLastChild':
       TabsInternalOperation.activateTab(activeTab.$TST.lastChild);
       return;
     case 'focusPreviousSibling':
-      TabsInternalOperation.activateTab(activeTab.$TST.previousSiblingTab);
+      TabsInternalOperation.activateTab(
+        activeTab.$TST.previousSiblingTab ||
+          (activeTab.$TST.parent ?
+            activeTab.$TST.parent.$TST.lastChild :
+            Tab.getLastRootTab(activeTab.windowId))
+      );
       return;
     case 'focusNextSibling':
-      TabsInternalOperation.activateTab(activeTab.$TST.nextSiblingTab);
+      TabsInternalOperation.activateTab(
+        activeTab.$TST.nextSiblingTab ||
+          (activeTab.$TST.parent ?
+            activeTab.$TST.parent.$TST.firstChild :
+            Tab.getFirstVisibleTab(activeTab.windowId))
+      );
+      return;
+
+    case 'simulateUpOnTree':
+      if (SidebarConnection.isOpen(activeTab.windowId)) {
+        if (configs.faviconizePinnedTabs &&
+            (activeTab.pinned ||
+             activeTab == Tab.getFirstNormalTab(activeTab.windowId))) {
+          const nextActiveId = await browser.runtime.sendMessage({
+            type:     Constants.kCOMMAND_GET_ABOVE_TAB,
+            windowId: activeTab.windowId,
+            tabId:    activeTab.id,
+          });
+          log(`simulateUpOnTree: nextActiveId = ${nextActiveId}`);
+          const nextActive = (
+            Tab.get(nextActiveId) ||
+            Tab.getLastVisibleTab(activeTab.windowId)
+          );
+          TabsInternalOperation.activateTab(nextActive, {
+            silently: true,
+          });
+        }
+        else {
+          focusPreviousSilently(activeTab);
+        }
+      }
+      else {
+        focusPrevious(activeTab);
+      }
+      return;
+    case 'simulateDownOnTree':
+      if (SidebarConnection.isOpen(activeTab.windowId)) {
+        if (configs.faviconizePinnedTabs &&
+            activeTab.pinned) {
+          const nextActiveId = await browser.runtime.sendMessage({
+            type:     Constants.kCOMMAND_GET_BELOW_TAB,
+            windowId: activeTab.windowId,
+            tabId:    activeTab.id,
+          });
+          log(`simulateDownOnTree: nextActiveId = ${nextActiveId}`);
+          const nextActive = (
+            Tab.get(nextActiveId) ||
+            Tab.getFirstNormalTab(activeTab.windowId)
+          );
+          TabsInternalOperation.activateTab(nextActive, {
+            silently: true,
+          });
+        }
+        else {
+          focusNextSilently(activeTab);
+        }
+      }
+      else {
+        focusNext(activeTab);
+      }
+      return;
+    case 'simulateLeftOnTree':
+      if (SidebarConnection.isOpen(activeTab.windowId)) {
+        if (configs.faviconizePinnedTabs &&
+            activeTab.pinned) {
+          const nextActiveId = await browser.runtime.sendMessage({
+            type:     Constants.kCOMMAND_GET_LEFT_TAB,
+            windowId: activeTab.windowId,
+            tabId:    activeTab.id,
+          });
+          log(`simulateLeftOnTree: nextActiveId = ${nextActiveId}`);
+          TabsInternalOperation.activateTab(Tab.get(nextActiveId), {
+            silently: true,
+          });
+        }
+        else if (await isSidebarRightSide(activeTab.windowId)) {
+          expandOrFocusToFirstChild(activeTab);
+        }
+        else {
+          collapseOrFocusToParent(activeTab);
+        }
+      }
+      else {
+        focusPrevious(activeTab);
+      }
+      return;
+    case 'simulateRightOnTree':
+      if (SidebarConnection.isOpen(activeTab.windowId)) {
+        if (configs.faviconizePinnedTabs &&
+            activeTab.pinned) {
+          const nextActiveId = await browser.runtime.sendMessage({
+            type:     Constants.kCOMMAND_GET_RIGHT_TAB,
+            windowId: activeTab.windowId,
+            tabId:    activeTab.id,
+          });
+          log(`simulateRightOnTree: nextActiveId = ${nextActiveId}`);
+          TabsInternalOperation.activateTab(Tab.get(nextActiveId), {
+            silently: true,
+          });
+        }
+        else if (await isSidebarRightSide(activeTab.windowId)) {
+          collapseOrFocusToParent(activeTab);
+        }
+        else {
+          expandOrFocusToFirstChild(activeTab);
+        }
+      }
+      else {
+        focusNext(activeTab);
+      }
       return;
 
     case 'tabbarUp':
@@ -264,6 +421,19 @@ async function onShortcutCommand(command) {
       });
       return;
 
+    case 'toggleTreeCollapsed':
+      if (activeTab.$TST.subtreeCollapsed)
+        Commands.expandTree(selectedTabs);
+      else
+        Commands.collapseTree(selectedTabs);
+      return;
+    case 'toggleTreeCollapsedRecursively':
+      if (activeTab.$TST.subtreeCollapsed)
+        Commands.expandTree(selectedTabs, { recursively: true });
+      else
+        Commands.collapseTree(selectedTabs, { recursively: true });
+      return;
+
     case 'toggleSubPanel':
       SidebarConnection.sendMessage({
         type:     Constants.kCOMMAND_TOGGLE_SUBPANEL,
@@ -291,10 +461,67 @@ async function onShortcutCommand(command) {
   }
 }
 
+function focusPrevious(activeTab) {
+  const nextActive = activeTab.$TST.nearestVisiblePrecedingTab ||
+    (!SidebarConnection.isOpen(activeTab.windowId) && activeTab.$TST.previousTab) ||
+    Tab.getLastVisibleTab(activeTab.windowId);
+  TabsInternalOperation.activateTab(nextActive);
+}
+
+function focusPreviousSilently(activeTab) {
+  const nextActive = activeTab.$TST.nearestVisiblePrecedingTab ||
+    (!SidebarConnection.isOpen(activeTab.windowId) && activeTab.$TST.previousTab) ||
+    Tab.getLastVisibleTab(activeTab.windowId);
+  TabsInternalOperation.activateTab(nextActive, {
+    silently: true,
+  });
+}
+
+function focusNext(activeTab) {
+  const nextActive = activeTab.$TST.nearestVisibleFollowingTab ||
+    (!SidebarConnection.isOpen(activeTab.windowId) && activeTab.$TST.nextTab) ||
+    Tab.getFirstVisibleTab(activeTab.windowId);
+  TabsInternalOperation.activateTab(nextActive);
+}
+
+function focusNextSilently(activeTab) {
+  const nextActive = activeTab.$TST.nearestVisibleFollowingTab ||
+    (!SidebarConnection.isOpen(activeTab.windowId) && activeTab.$TST.nextTab) ||
+    Tab.getFirstVisibleTab(activeTab.windowId);
+  TabsInternalOperation.activateTab(nextActive, {
+    silently: true,
+  });
+}
+
+async function isSidebarRightSide(windowId) {
+  const position = await browser.runtime.sendMessage({
+    type: Constants.kCOMMAND_GET_SIDEBAR_POSITION,
+    windowId,
+  });
+  return position == Constants.kTABBAR_POSITION_RIGHT;
+}
+
+function collapseOrFocusToParent(activeTab) {
+  if (!activeTab.$TST.subtreeCollapsed && activeTab.$TST.hasChild)
+    Commands.collapseTree(activeTab);
+  else
+    TabsInternalOperation.activateTab(activeTab.$TST.parent);
+}
+
+function expandOrFocusToFirstChild(activeTab) {
+  if (activeTab.$TST.subtreeCollapsed && activeTab.$TST.hasChild)
+    Commands.expandTree(activeTab);
+  else
+    TabsInternalOperation.activateTab(activeTab.$TST.firstChild, {
+      silently: true,
+    });
+}
+
 // This must be synchronous and return Promise on demando, to avoid
 // blocking to other listeners.
 function onMessage(message, sender) {
-  if (!message ||
+  if (mInitializationPhase < PHASE_BACKGROUND_BUILT ||
+      !message ||
       typeof message.type != 'string' ||
       message.type.indexOf('treestyletab:') != 0)
     return;
@@ -311,7 +538,8 @@ function onMessage(message, sender) {
       return (async () => {
         if (!Tab.get(message.tabId))
           await Tab.waitUntilTracked(message.tabId);
-        return Tab.get(message.tabId).$TST.promisedUniqueId;
+        const tab = Tab.get(message.tabId);
+        return tab ? tab.$TST.promisedUniqueId : {};
       })();
 
     case Constants.kCOMMAND_GET_THEME_DECLARATIONS:
@@ -333,15 +561,21 @@ function onMessage(message, sender) {
     case Constants.kCOMMAND_SET_CONFIG_VALUE:
       return Promise.resolve(configs[message.key] = message.value);
 
+    case Constants.kCOMMAND_GET_USER_STYLE_RULES:
+      return Promise.resolve(loadUserStyleRules());
+
     case Constants.kCOMMAND_PING_TO_BACKGROUND: // return tabs as the pong, to optimizie further initialization tasks in the sidebar
+      TabsUpdate.completeLoadingTabs(message.windowId); // don't wait here for better perfomance
+      return Promise.resolve(TabsStore.windows.get(message.windowId).export(true));
+
     case Constants.kCOMMAND_PULL_TABS:
       if (message.windowId) {
         TabsUpdate.completeLoadingTabs(message.windowId); // don't wait here for better perfomance
-        return Promise.resolve(TabsStore.windows.get(message.windowId).export(true));
+        return Promise.resolve(TabsStore.windows.get(message.windowId).export(true).tabs);
       }
       return Promise.resolve(message.tabIds.map(id => {
         const tab = Tab.get(id);
-        return tab && tab.$TST.export(true);
+        return tab?.$TST.export(true);
       }));
 
     case Constants.kCOMMAND_PULL_TABS_ORDER:
@@ -349,10 +583,14 @@ function onMessage(message, sender) {
 
     case Constants.kCOMMAND_PULL_TREE_STRUCTURE:
       return (async () => {
-        while (!mInitialized) {
+        while (mInitializationPhase < PHASE_BACKGROUND_READY) {
           await wait(10);
         }
-        const structure = TreeBehavior.getTreeStructureFromTabs(Tab.getAllTabs(message.windowId));
+        const structure = TreeBehavior.getTreeStructureFromTabs(
+          message.windowId ?
+            Tab.getAllTabs(message.windowId) :
+            message.tabIds.map(id => Tab.get(id))
+        );
         return { structure };
       })();
 
@@ -381,10 +619,11 @@ function onMessage(message, sender) {
       return Background.confirmToCloseTabs(message.tabs, message);
 
     default:
-      const API_PREFIX_MATCHER = /^treestyletab:api:/;
-      if (API_PREFIX_MATCHER.test(message.type)) {
-        message.type = message.type.replace(API_PREFIX_MATCHER, '');
-        return onMessageExternal(message, sender);
+      if (TSTAPI.INTERNAL_CALL_PREFIX_MATCHER.test(message.type)) {
+        return onMessageExternal({
+          ...message,
+          type: message.type.replace(TSTAPI.INTERNAL_CALL_PREFIX_MATCHER, ''),
+        }, sender);
       }
       break;
   }
@@ -393,26 +632,104 @@ function onMessage(message, sender) {
 // This must be synchronous and return Promise on demando, to avoid
 // blocking to other listeners.
 function onMessageExternal(message, sender) {
-  //log('onMessageExternal: ', message, sender);
+  if (mInitializationPhase < PHASE_BACKGROUND_INITIALIZED)
+    return;
+
   switch (message.type) {
+    case TSTAPI.kGET_VERSION:
+      return Promise.resolve(browser.runtime.getManifest().version);
+
     case TSTAPI.kGET_TREE:
       return (async () => {
-        const tabs = await TSTAPI.getTargetTabs(message, sender);
+        const tabs = await (message.rendered ?
+          TSTAPI.getTargetRenderedTabs(message, sender) :
+          TSTAPI.getTargetTabs(message, sender));
         const cache = {};
-        return TSTAPI.formatTabResult(
-          Array.from(tabs, tab => new TSTAPI.TreeItem(tab, {
-            interval: message.interval,
-            cache
-          })),
-          message,
+        const treeItems = Array.from(tabs, tab => TSTAPI.exportTab(tab, {
+          interval: message.interval,
+          cache,
+        }));
+        const result = TSTAPI.formatTabResult(
+          treeItems,
+          {
+            ...message,
+            // This must return an array of root tabs if just the window id is specified.
+            // See also: https://github.com/piroor/treestyletab/issues/2763
+            ...((message.window || message.windowId) && !message.tab && !message.tabs ? { tab: '*' } : {})
+          },
           sender.id
         );
+        TSTAPI.clearCache(cache);
+        return result;
+      })();
+
+    case TSTAPI.kGET_LIGHT_TREE:
+      return (async () => {
+        const tabs = await (message.rendered ?
+          TSTAPI.getTargetRenderedTabs(message, sender) :
+          TSTAPI.getTargetTabs(message, sender));
+        const cache = {};
+        const treeItems = Array.from(tabs, tab => TSTAPI.exportTab(tab, {
+          light:    true,
+          interval: message.interval,
+          cache,
+        }));
+        const result = TSTAPI.formatTabResult(
+          treeItems,
+          {
+            ...message,
+            // This must return an array of root tabs if just the window id is specified.
+            // See also: https://github.com/piroor/treestyletab/issues/2763
+            ...((message.window || message.windowId) && !message.tab && !message.tabs ? { tab: '*' } : {})
+          },
+          sender.id
+        );
+        TSTAPI.clearCache(cache);
+        return result;
+      })();
+
+    case TSTAPI.kSTICK_TAB:
+      return (async () => {
+        const tabs = await TSTAPI.getTargetTabs(message, sender);
+        await doProgressively(
+          tabs,
+          tab => Commands.toggleSticky(tab, true),
+          message.interval
+        );
+        return true;
+      })();
+
+    case TSTAPI.kUNSTICK_TAB:
+      return (async () => {
+        const tabs = await TSTAPI.getTargetTabs(message, sender);
+        await doProgressively(
+          tabs,
+          tab => Commands.toggleSticky(tab, false),
+          message.interval
+        );
+        return true;
+      })();
+
+    case TSTAPI.kTOGGLE_STICKY_STATE:
+      return (async () => {
+        const tabs = await TSTAPI.getTargetTabs(message, sender);
+        let firstTabIsSticky = undefined;
+        await doProgressively(
+          tabs,
+          tab => {
+            if (firstTabIsSticky === undefined)
+              firstTabIsSticky = tab.$TST.sticky;
+            Commands.toggleSticky(tab, !firstTabIsSticky);
+          },
+          message.interval
+        );
+        return true;
       })();
 
     case TSTAPI.kCOLLAPSE_TREE:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
-        await TSTAPI.doProgressively(
+        await doProgressively(
           tabs,
           tab => Commands.collapseTree(tab, {
             recursively: !!message.recursively
@@ -425,7 +742,7 @@ function onMessageExternal(message, sender) {
     case TSTAPI.kEXPAND_TREE:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
-        await TSTAPI.doProgressively(
+        await doProgressively(
           tabs,
           tab => Commands.expandTree(tab, {
             recursively: !!message.recursively
@@ -438,7 +755,7 @@ function onMessageExternal(message, sender) {
     case TSTAPI.kTOGGLE_TREE_COLLAPSED:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
-        await TSTAPI.doProgressively(
+        await doProgressively(
           tabs,
           tab => Tree.collapseExpandSubtree(tab, {
             collapsed: !tab.$TST.subtreeCollapsed,
@@ -501,7 +818,7 @@ function onMessageExternal(message, sender) {
     case TSTAPI.kDEMOTE:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
-        const results = await TSTAPI.doProgressively(
+        const results = await doProgressively(
           tabs,
           tab => Commands.indent(tab, message),
           message.interval
@@ -513,7 +830,7 @@ function onMessageExternal(message, sender) {
     case TSTAPI.kPROMOTE:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
-        const results = await TSTAPI.doProgressively(
+        const results = await doProgressively(
           tabs,
           tab => Commands.outdent(tab, message),
           message.interval
@@ -524,7 +841,7 @@ function onMessageExternal(message, sender) {
     case TSTAPI.kMOVE_UP:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
-        const results = await TSTAPI.doProgressively(
+        const results = await doProgressively(
           tabs,
           tab => Commands.moveUp(tab, message),
           message.interval
@@ -542,7 +859,7 @@ function onMessageExternal(message, sender) {
     case TSTAPI.kMOVE_DOWN:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
-        const results = await TSTAPI.doProgressively(
+        const results = await doProgressively(
           tabs,
           tab => Commands.moveDown(tab, message),
           message.interval
@@ -560,7 +877,7 @@ function onMessageExternal(message, sender) {
     case TSTAPI.kMOVE_BEFORE:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
-        const results = await TSTAPI.doProgressively(
+        const results = await doProgressively(
           tabs,
           tab => Commands.moveBefore(tab, message),
           message.interval
@@ -571,7 +888,7 @@ function onMessageExternal(message, sender) {
     case TSTAPI.kMOVE_AFTER:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
-        const results = await TSTAPI.doProgressively(
+        const results = await doProgressively(
           tabs,
           tab => Commands.moveAfter(tab, message),
           message.interval
@@ -582,7 +899,7 @@ function onMessageExternal(message, sender) {
     case TSTAPI.kFOCUS:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
-        const tabsArray = await TSTAPI.doProgressively(
+        const tabsArray = await doProgressively(
           tabs,
           tab => {
             TabsInternalOperation.activateTab(tab, {
@@ -595,13 +912,34 @@ function onMessageExternal(message, sender) {
         return TSTAPI.formatResult(tabsArray.map(() => true), message);
       })();
 
+    case TSTAPI.kCREATE:
+      return (async () => {
+        const windowId = message.params.windowId;
+        const win = TabsStore.windows.get(windowId);
+        if (!win)
+          throw new Error(`invalid windowId ${windowId}: it must be valid window id`);
+        win.bypassTabControlCount++;
+        const tab = await TabsOpen.openURIInTab(message.params, { windowId });
+        return TSTAPI.exportTab(tab, { addonId: sender.id });
+      })();
+
     case TSTAPI.kDUPLICATE:
       return (async () => {
         const tabs   = await TSTAPI.getTargetTabs(message, sender);
         let behavior = configs.autoAttachOnDuplicated;
         switch (String(message.as || 'sibling').toLowerCase()) {
           case 'child':
-            behavior = Constants.kNEWTAB_OPEN_AS_CHILD;
+            behavior = behavior == Constants.kNEWTAB_OPEN_AS_CHILD_TOP ?
+              Constants.kNEWTAB_OPEN_AS_CHILD_TOP :
+              behavior == Constants.kNEWTAB_OPEN_AS_CHILD_END ?
+                Constants.kNEWTAB_OPEN_AS_CHILD_END :
+                Constants.kNEWTAB_OPEN_AS_CHILD;
+            break;
+          case 'first-child':
+            behavior = Constants.kNEWTAB_OPEN_AS_CHILD_TOP;
+            break;
+          case 'last-child':
+            behavior = Constants.kNEWTAB_OPEN_AS_CHILD_END;
             break;
           case 'sibling':
             behavior = Constants.kNEWTAB_OPEN_AS_SIBLING;
@@ -615,7 +953,7 @@ function onMessageExternal(message, sender) {
           default:
             break;
         }
-        const tabsArray = await TSTAPI.doProgressively(
+        const tabsArray = await doProgressively(
           tabs,
           async tab => {
             return Commands.duplicateTab(tab, {
@@ -632,13 +970,31 @@ function onMessageExternal(message, sender) {
     case TSTAPI.kGROUP_TABS:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
+        const temporaryStateParams = (message.temporary && !message.temporaryAggressive) ?
+          {
+            temporary:           true,
+            temporaryAggressive: false,
+          } :
+          (!message.temporary && message.temporaryAggressive) ?
+            {
+              temporary:           false,
+              temporaryAggressive: true,
+            } :
+            (message.temporaryAggressive === false && message.temporary === false) ?
+              {
+                temporary:           false,
+                temporaryAggressive: false,
+              } :
+              {};
         const tab = await TabsGroup.groupTabs(Array.from(tabs), {
-          broadcast: true
+          title:     message.title,
+          broadcast: true,
+          ...TabsGroup.temporaryStateParams(configs.groupTabTemporaryStateForAPI),
+          ...temporaryStateParams,
         });
         if (!tab)
           return null;
-        const treeItem = new TSTAPI.TreeItem(tab);
-        return treeItem.exportFor(sender.id);
+        return TSTAPI.exportTab(tab, { addonId: sender.id });
       })();
 
     case TSTAPI.kOPEN_IN_NEW_WINDOW:
@@ -658,14 +1014,16 @@ function onMessageExternal(message, sender) {
           message.containerId || 'firefox-default'
         );
         const cache = {};
-        return TSTAPI.formatTabResult(
-          reopenedTabs.map(tab => new TSTAPI.TreeItem(tab, {
+        const result = await TSTAPI.formatTabResult(
+          reopenedTabs.map(tab => TSTAPI.exportTab(tab, {
             interval: message.interval,
-            cache
+            addonId:  sender.id,
+            cache,
           })),
-          message,
-          sender.id
+          message
         );
+        TSTAPI.clearCache(cache);
+        return result;
       })();
 
     case TSTAPI.kGET_TREE_STRUCTURE:
@@ -691,7 +1049,10 @@ function onMessageExternal(message, sender) {
         let states = message.state || message.states;
         if (!Array.isArray(states))
           states = [states];
-        const tabsArray = await TSTAPI.doProgressively(
+        states = mapAndFilterUniq(states, state => state && String(state) || undefined);
+        if (states.length == 0)
+          return true;
+        const tabsArray = await doProgressively(
           tabs,
           tab => {
             for (const state of states) {
@@ -713,7 +1074,10 @@ function onMessageExternal(message, sender) {
         let states = message.state || message.states;
         if (!Array.isArray(states))
           states = [states];
-        const tabsArray = await TSTAPI.doProgressively(
+        states = mapAndFilterUniq(states, state => state && String(state) || undefined);
+        if (states.length == 0)
+          return true;
+        const tabsArray = await doProgressively(
           tabs,
           tab => {
             for (const state of states) {
@@ -733,8 +1097,8 @@ function onMessageExternal(message, sender) {
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
         configs.grantedRemovingTabIds = mapAndFilterUniq(configs.grantedRemovingTabIds.concat(tabs), tab => {
-          tab = TabsStore.ensureLivingTab(tab);
-          return tab && tab.id || undefined;
+          tab = TabsStore.ensureLivingItem(tab);
+          return tab?.id || undefined;
         });
         return true;
       })();
@@ -753,5 +1117,45 @@ function onMessageExternal(message, sender) {
         message.id || message.bookmarkId,
         { discarded: message.discarded }
       );
+
+    case TSTAPI.kSET_TOOLTIP_TEXT:
+      return (async () => {
+        const tabs = await TSTAPI.getTargetTabs(message, sender);
+        for (const tab of tabs) {
+          tab.$TST.registerTooltipText(sender.id, message.text || '', !!message.force);
+        }
+        return true;
+      })();
+
+    case TSTAPI.kCLEAR_TOOLTIP_TEXT:
+      return (async () => {
+        const tabs = await TSTAPI.getTargetTabs(message, sender);
+        for (const tab of tabs) {
+          tab.$TST.unregisterTooltipText(sender.id);
+        }
+        return true;
+      })();
+
+    case TSTAPI.kREGISTER_AUTO_STICKY_STATES:
+      TreeItem.registerAutoStickyState(sender.id, message.state || message.states);
+      break;
+
+    case TSTAPI.kUNREGISTER_AUTO_STICKY_STATES:
+      TreeItem.unregisterAutoStickyState(sender.id, message.state || message.states);
+      break;
   }
 }
+
+Tab.onStateChanged.addListener((tab, state, added) => {
+  switch (state) {
+    case Constants.kTAB_STATE_STICKY:
+      if (TSTAPI.hasListenerForMessageType(TSTAPI.kNOTIFY_TAB_STICKY_STATE_CHANGED)) {
+        TSTAPI.broadcastMessage({
+          type:   TSTAPI.kNOTIFY_TAB_STICKY_STATE_CHANGED,
+          tab,
+          sticky: !!added,
+        }, { tabProperties: ['tab'] }).catch(_error => {});
+      }
+      break;
+  }
+});
