@@ -13,18 +13,15 @@ import {
   countMatched,
   configs
 } from '/common/common.js';
-
 import * as Constants from '/common/constants.js';
 import * as TabsStore from '/common/tabs-store.js';
 import * as TabsInternalOperation from '/common/tabs-internal-operation.js';
 import * as TreeBehavior from '/common/tree-behavior.js';
-import * as SidebarConnection from '/common/sidebar-connection.js';
 
-import Tab from '/common/Tab.js';
+import { Tab } from '/common/TreeItem.js';
 
 import * as Background from './background.js';
 import * as TabsGroup from './tabs-group.js';
-import * as TabsOpen from './tabs-open.js';
 import * as Tree from './tree.js';
 import * as Commands from './commands.js';
 
@@ -42,7 +39,8 @@ Tab.onRemoving.addListener(async (tab, removeInfo = {}) => {
   let newParent;
   const successor = tab.$TST.possibleSuccessorWithDifferentContainer;
   if (successor) {
-    closeParentBehavior = Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD;
+    log('override closeParentBehaior with kPARENT_TAB_OPERATION_BEHAVIOR_PROMOTE_FIRST_CHILD for different container successor ', successor);
+    closeParentBehavior = Constants.kPARENT_TAB_OPERATION_BEHAVIOR_PROMOTE_FIRST_CHILD;
     // When a new tab is created with a different container and this tab
     // is removed immediately before the new tab is completely handled,
     // TST fails to detect the new tab as the successor of this tab. Thus,
@@ -52,11 +50,14 @@ Tab.onRemoving.addListener(async (tab, removeInfo = {}) => {
       newParent = successor;
   }
   else {
-    closeParentBehavior = TreeBehavior.getCloseParentBehaviorForTabWithSidebarOpenState(tab, removeInfo);
+    closeParentBehavior = TreeBehavior.getParentTabOperationBehavior(tab, {
+      context: Constants.kPARENT_TAB_OPERATION_CONTEXT_CLOSE,
+      ...removeInfo,
+    });
+    log('detected closeParentBehaior: ', closeParentBehavior);
   }
 
-  if (!SidebarConnection.isOpen(tab.windowId) &&
-      closeParentBehavior != Constants.kCLOSE_PARENT_BEHAVIOR_CLOSE_ALL_CHILDREN &&
+  if (closeParentBehavior != Constants.kPARENT_TAB_OPERATION_BEHAVIOR_ENTIRE_TREE &&
       tab.$TST.subtreeCollapsed)
     Tree.collapseExpandSubtree(tab, {
       collapsed: false,
@@ -66,6 +67,11 @@ Tab.onRemoving.addListener(async (tab, removeInfo = {}) => {
 
   const postProcessParams = {
     windowId:     tab.windowId,
+    removedTab:   tab.$TST.export(true),
+    structure:    TreeBehavior.getTreeStructureFromTabs([tab, ...tab.$TST.descendants], {
+      full:                 true,
+      keepParentOfRootTabs: true
+    }),
     insertBefore: tab, // not firstChild, because the "tab" is disappeared from tree.
     parent:       tab.$TST.parent,
     newParent,
@@ -86,7 +92,7 @@ Tab.onRemoving.addListener(async (tab, removeInfo = {}) => {
     // They will be processed again after confirmation.
     Tree.detachAllChildren(tab, {
       newParent,
-      behavior:         Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_ALL_CHILDREN,
+      behavior:         Constants.kPARENT_TAB_OPERATION_BEHAVIOR_PROMOTE_ALL_CHILDREN,
       dontExpand:       true,
       dontUpdateIndent: true,
       broadcast:        true
@@ -96,58 +102,33 @@ Tab.onRemoving.addListener(async (tab, removeInfo = {}) => {
     await handleRemovingPostProcess(postProcessParams)
   }
 
-  const window = TabsStore.windows.get(tab.windowId);
-  if (!window.internalClosingTabs.has(tab.$TST.parentId))
+  const win = TabsStore.windows.get(tab.windowId);
+  if (!win.internalClosingTabs.has(tab.$TST.parentId))
     Tree.detachTab(tab, {
       dontUpdateIndent: true,
+      dontSyncParentToOpenerTab: true,
       broadcast:        true
     });
 });
-async function handleRemovingPostProcess({ closeParentBehavior, windowId, parent, newParent, insertBefore, nearestFollowingRootTab, children, descendants } = {}) {
-  if (closeParentBehavior == Constants.kCLOSE_PARENT_BEHAVIOR_CLOSE_ALL_CHILDREN)
-    await closeChildTabs(descendants);
+async function handleRemovingPostProcess({ closeParentBehavior, windowId, parent, newParent, insertBefore, nearestFollowingRootTab, children, descendants, removedTab, structure } = {}) {
+  log('handleRemovingPostProcess ', { closeParentBehavior, windowId, parent, newParent, insertBefore, nearestFollowingRootTab, children, descendants, removedTab, structure });
+  if (closeParentBehavior == Constants.kPARENT_TAB_OPERATION_BEHAVIOR_ENTIRE_TREE)
+    await closeChildTabs(descendants, {
+      triggerTab:        removedTab,
+      originalStructure: structure
+    });
 
-  if (closeParentBehavior == Constants.kCLOSE_PARENT_BEHAVIOR_REPLACE_WITH_GROUP_TAB &&
-      children.length > 1 &&
-      countMatched(children,
-                   tab => !tab.$TST.states.has(Constants.kTAB_STATE_TO_BE_REMOVED)) > 1) {
-    log('trying to replace the closing tab with a new group tab');
-    const firstChild = children[0];
-    const uri = TabsGroup.makeGroupTabURI({
-      title:     browser.i18n.getMessage('groupTab_label', firstChild.title),
-      temporaryAggressive: true
-    });
-    const window = TabsStore.windows.get(windowId);
-    window.toBeOpenedTabsWithPositions++;
-    const groupTab = await TabsOpen.openURIInTab(uri, {
+  const replacedGroupTab = (closeParentBehavior == Constants.kPARENT_TAB_OPERATION_BEHAVIOR_REPLACE_WITH_GROUP_TAB) ?
+    await TabsGroup.tryReplaceTabWithGroup(null, { windowId, parent, children, insertBefore, newParent }) :
+    null;
+
+  if (!replacedGroupTab) {
+    await Tree.detachAllChildren(null, {
       windowId,
-      insertBefore,
-      inBackground: true
-    });
-    log('group tab: ', dumpTab(groupTab));
-    if (!groupTab) // the window is closed!
-      return;
-    if (newParent || parent)
-      await Tree.attachTabTo(groupTab, newParent || parent, {
-        dontMove:  true,
-        broadcast: true
-      });
-    for (const child of children) {
-      await Tree.attachTabTo(child, groupTab, {
-        dontMove:  true,
-        broadcast: true
-      });
-    }
-    // This can be triggered on closing of multiple tabs,
-    // so we should cleanup it on such cases for safety.
-    // https://github.com/piroor/treestyletab/issues/2317
-    wait(1000).then(() => TabsGroup.reserveToCleanupNeedlessGroupTab(groupTab));
-  }
-  else {
-    Tree.detachAllChildren(null, {
       parent,
       newParent,
       children,
+      descendants,
       nearestFollowingRootTab,
       behavior:  closeParentBehavior,
       broadcast: true
@@ -159,25 +140,30 @@ async function tryGrantCloseTab(tab, closeParentBehavior) {
   log('tryGrantClose: ', { alreadyGranted: configs.grantedRemovingTabIds, closing: dumpTab(tab) });
   const alreadyGranted = configs.grantedRemovingTabIds.includes(tab.id);
   configs.grantedRemovingTabIds = configs.grantedRemovingTabIds.filter(id => id != tab.id);
-  if (alreadyGranted)
+  if (!tab || alreadyGranted) {
+    log(' => no need to confirm');
     return true;
+  }
 
   const self = tryGrantCloseTab;
 
   self.closingTabIds.push(tab.id);
-  if (closeParentBehavior == Constants.kCLOSE_PARENT_BEHAVIOR_CLOSE_ALL_CHILDREN) {
+  if (closeParentBehavior == Constants.kPARENT_TAB_OPERATION_BEHAVIOR_ENTIRE_TREE) {
     self.closingDescendantTabIds = self.closingDescendantTabIds
       .concat(TreeBehavior.getClosingTabsFromParent(tab).map(tab => tab.id));
     self.closingDescendantTabIds = Array.from(new Set(self.closingDescendantTabIds));
   }
 
-  if (self.promisedGrantedToCloseTabs)
+  if (self.promisedGrantedToCloseTabs) {
+    log(' => have promisedGrantedToCloseTabs');
     return self.promisedGrantedToCloseTabs;
+  }
 
   self.closingTabWasActive = self.closingTabWasActive || tab.active;
 
   let shouldRestoreCount;
   self.promisedGrantedToCloseTabs = wait(250).then(async () => {
+    log(' => confirmation with delay');
     const closingTabIds = new Set(self.closingTabIds);
     let allClosingTabs = new Set();
     allClosingTabs.add(tab);
@@ -185,7 +171,9 @@ async function tryGrantCloseTab(tab, closeParentBehavior) {
     self.closingDescendantTabIds = mapAndFilterUniq(self.closingDescendantTabIds, id => {
       if (closingTabIds.has(id))
         return undefined;
-      allClosingTabs.add(Tab.get(id));
+      const tab = Tab.get(id);
+      if (tab) // ignore already closed tabs
+        allClosingTabs.add(tab);
       return id;
     });
     allClosingTabs = Array.from(allClosingTabs);
@@ -193,19 +181,22 @@ async function tryGrantCloseTab(tab, closeParentBehavior) {
     const restorableClosingTabsCount = countMatched(
       allClosingTabs,
       tab => tab.url != 'about:blank' &&
-             tab.url != configs.guessNewOrphanTabAsOpenedByNewTabCommandUrl
+             !tab.$TST.isNewTabCommandTab
     );
+    log(' => restorableClosingTabsCount: ', restorableClosingTabsCount);
     if (restorableClosingTabsCount > 0) {
       log('tryGrantClose: show confirmation for ', allClosingTabs);
       return Background.confirmToCloseTabs(allClosingTabs.slice(1).map(tab => tab.$TST.sanitized), {
         windowId:   tab.windowId,
         messageKey: 'warnOnCloseTabs_fromOutside_message',
-        titleKey:   'warnOnCloseTabs_fromOutside_title'
+        titleKey:   'warnOnCloseTabs_fromOutside_title',
+        minConfirmCount: 0
       });
     }
     return true;
   })
     .then(async (granted) => {
+      log(' => granted: ', granted);
       // remove the closed tab itself because it is already closed!
       configs.grantedRemovingTabIds = configs.grantedRemovingTabIds.filter(id => id != tab.id);
       if (granted)
@@ -231,13 +222,13 @@ tryGrantCloseTab.closingDescendantTabIds    = [];
 tryGrantCloseTab.closingTabWasActive        = false;
 tryGrantCloseTab.promisedGrantedToCloseTabs = null;
 
-async function closeChildTabs(tabs) {
+async function closeChildTabs(tabs, { triggerTab, originalStructure } = {}) {
   //if (!fireTabSubtreeClosingEvent(parent, tabs))
   //  return;
 
   //markAsClosedSet([parent].concat(tabs));
   // close bottom to top!
-  await TabsInternalOperation.removeTabs(tabs.reverse());
+  await TabsInternalOperation.removeTabs(tabs.slice(0).reverse(), { triggerTab, originalStructure });
   //fireTabSubtreeClosedEvent(parent, tabs);
 }
 
@@ -255,12 +246,13 @@ Tab.onRemoved.addListener((tab, info) => {
     Tree.detachAllChildren(tab, {
       children:  info.oldChildren,
       parent:    info.oldParent,
-      behavior:  Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD,
+      behavior:  Constants.kPARENT_TAB_OPERATION_BEHAVIOR_PROMOTE_FIRST_CHILD,
       broadcast: true
     });
   }
   if (info.oldParent) {
     Tree.detachTab(tab, {
+      dontSyncParentToOpenerTab: true,
       parent:    info.oldParent,
       broadcast: true
     });
@@ -268,25 +260,31 @@ Tab.onRemoved.addListener((tab, info) => {
 });
 
 browser.windows.onRemoved.addListener(windowId  => {
-  const window = TabsStore.windows.get(windowId);
-  if (!window)
+  const win = TabsStore.windows.get(windowId);
+  if (!win)
     return;
-  configs.grantedRemovingTabIds = configs.grantedRemovingTabIds.filter(id => !window.tabs.has(id));
+  configs.grantedRemovingTabIds = configs.grantedRemovingTabIds.filter(id => !win.tabs.has(id));
 });
 
 
 Tab.onDetached.addListener((tab, info = {}) => {
   log('Tabs.onDetached ', dumpTab(tab));
-  let closeParentBehavior = TreeBehavior.getCloseParentBehaviorForTabWithSidebarOpenState(tab, info);
-  if (closeParentBehavior == Constants.kCLOSE_PARENT_BEHAVIOR_CLOSE_ALL_CHILDREN)
-    closeParentBehavior = Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD;
+  let closeParentBehavior = TreeBehavior.getParentTabOperationBehavior(tab, {
+    context: Constants.kPARENT_TAB_OPERATION_CONTEXT_MOVE,
+    ...info
+  });
+  if (closeParentBehavior == Constants.kPARENT_TAB_OPERATION_BEHAVIOR_ENTIRE_TREE)
+    closeParentBehavior = Constants.kPARENT_TAB_OPERATION_BEHAVIOR_PROMOTE_FIRST_CHILD;
 
+  const dontSyncParentToOpenerTab = info.trigger == 'tabs.onDetached';
   Tree.detachAllChildren(tab, {
+    dontSyncParentToOpenerTab,
     behavior:  closeParentBehavior,
     broadcast: true
   });
   //reserveCloseRelatedTabs(toBeClosedTabs);
   Tree.detachTab(tab, {
+    dontSyncParentToOpenerTab,
     dontUpdateIndent: true,
     broadcast:        true
   });
