@@ -9,17 +9,17 @@ import {
   log as internalLogger,
   dumpTab,
   wait,
-  configs
+  configs,
+  isMacOS,
 } from '/common/common.js';
-
 import * as ApiTabs from '/common/api-tabs.js';
 import * as Constants from '/common/constants.js';
+import * as Permissions from '/common/permissions.js';
 import * as TabsStore from '/common/tabs-store.js';
 import * as TabsInternalOperation from '/common/tabs-internal-operation.js';
-import * as SidebarConnection from '/common/sidebar-connection.js';
 import * as TSTAPI from '/common/tst-api.js';
 
-import Tab from '/common/Tab.js';
+import { Tab } from '/common/TreeItem.js';
 import Window from '/common/Window.js';
 
 import * as Background from './background.js';
@@ -30,32 +30,69 @@ function log(...args) {
 }
 
 
+const PHASE_LOADING                = 0;
+const PHASE_BACKGROUND_INITIALIZED = 1;
+const PHASE_BACKGROUND_BUILT       = 2;
+const PHASE_BACKGROUND_READY       = 3;
+let mInitializationPhase = PHASE_LOADING;
+
+Background.onInit.addListener(() => {
+  mInitializationPhase = PHASE_BACKGROUND_INITIALIZED;
+});
+Background.onBuilt.addListener(() => {
+  mInitializationPhase = PHASE_BACKGROUND_BUILT;
+});
+Background.onReady.addListener(() => {
+  mInitializationPhase = PHASE_BACKGROUND_READY;
+});
+
+
 let mTabSwitchedByShortcut       = false;
 let mMaybeTabSwitchingByShortcut = false;
 
-Window.onInitialized.addListener(window => {
+const mLastTabsCountInWindow = new Map();
+
+Window.onInitialized.addListener(win => {
   browser.tabs.query({
-    windowId: window.id,
+    windowId: win.id,
     active:   true
   })
     .then(activeTabs => {
-      if (!window.lastActiveTab)
-        window.lastActiveTab = activeTabs[0].id;
+      // There may be no active tab on a startup...
+      if (activeTabs.length > 0 &&
+          !win.lastActiveTab)
+        win.lastActiveTab = activeTabs[0].id;
     });
 });
 
+browser.windows.onRemoved.addListener(windowId => {
+  mLastTabsCountInWindow.delete(windowId);
+});
+
+
 Tab.onActivating.addListener(async (tab, info = {}) => { // return false if the activation should be canceled
-  log('Tabs.onActivating ', { tab: dumpTab(tab), info });
-  if (tab.$TST.shouldReloadOnSelect) {
+  log('Tabs.onActivating ', { tab: dumpTab(tab), info, mMaybeTabSwitchingByShortcut });
+
+  if (mMaybeTabSwitchingByShortcut) {
+    const lastCount = mLastTabsCountInWindow.get(tab.windowId);
+    const count = Tab.getAllTabs(tab.windowId).length;
+    if (lastCount != count) {
+      log('tabs are created or removed: cancel tab switching');
+      mMaybeTabSwitchingByShortcut = false;
+    }
+    mLastTabsCountInWindow.set(tab.windowId, count);
+  }
+
+  if (tab.$TST.temporaryMetadata.has('shouldReloadOnSelect')) {
     browser.tabs.reload(tab.id)
       .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
-    delete tab.$TST.shouldReloadOnSelect;
+    tab.$TST.temporaryMetadata.delete('shouldReloadOnSelect');
   }
-  const window = TabsStore.windows.get(tab.windowId);
-  log('  lastActiveTab: ', window.lastActiveTab);
-  const lastActiveTab = Tab.get(window.lastActiveTab);
+  const win = TabsStore.windows.get(tab.windowId);
+  log('  lastActiveTab: ', win.lastActiveTab); // it may be blank on a startup
+  const lastActiveTab = Tab.get(win.lastActiveTab || info.previousTabId);
   cancelDelayedExpand(lastActiveTab);
-  let shouldSkipCollapsed = (
+  const shouldSkipCollapsed = (
     !info.byInternalOperation &&
     mMaybeTabSwitchingByShortcut &&
     configs.skipCollapsedTabsForTabSwitchingShortcuts
@@ -92,10 +129,11 @@ Tab.onActivating.addListener(async (tab, info = {}) => { // return false if the
         log('  => apply unfocusableCollapsedTab');
         allowed = await TSTAPI.tryOperationAllowed(
           TSTAPI.kNOTIFY_TRY_EXPAND_TREE_FROM_FOCUSED_COLLAPSED_TAB,
-          { tab: new TSTAPI.TreeItem(tab, { cache }),
+          { tab,
             focusDirection },
-          { tabProperties: ['tab'] }
+          { tabProperties: ['tab'], cache }
         );
+        TSTAPI.clearCache(cache);
         if (allowed) {
           const toBeExpandedAncestors = [tab].concat(tab.$TST.ancestors) ;
           for (const ancestor of toBeExpandedAncestors) {
@@ -106,13 +144,11 @@ Tab.onActivating.addListener(async (tab, info = {}) => { // return false if the
           }
         }
         else {
-          shouldSkipCollapsed = true;
           log('  => canceled by someone.');
         }
       }
       info.allowed = allowed;
-      if (!shouldSkipCollapsed)
-        await handleNewActiveTab(tab, info);
+      await handleNewActiveTab(tab, info);
     }
     if (shouldSkipCollapsed) {
       log('=> reaction for focusing collapsed descendant while Ctrl-Tab/Ctrl-Shift-Tab');
@@ -121,8 +157,8 @@ Tab.onActivating.addListener(async (tab, info = {}) => { // return false if the
         return false;
       log('successor = ', successor.id);
       if (shouldSkipCollapsed &&
-          (window.lastActiveTab == successor.id ||
-           successor.$TST.descendants.some(tab => tab.id == window.lastActiveTab)) &&
+          (win.lastActiveTab == successor.id ||
+           successor.$TST.descendants.some(tab => tab.id == win.lastActiveTab)) &&
           focusDirection > 0) {
         log('=> redirect successor (focus moved from the successor itself or its descendants)');
         successor = successor.$TST.nearestVisibleFollowingTab;
@@ -147,18 +183,19 @@ Tab.onActivating.addListener(async (tab, info = {}) => { // return false if the
       }
       const allowed = await TSTAPI.tryOperationAllowed(
         TSTAPI.kNOTIFY_TRY_REDIRECT_FOCUS_FROM_COLLAPSED_TAB,
-        { tab: new TSTAPI.TreeItem(tab, { cache }),
+        { tab,
           focusDirection },
-        { tabProperties: ['tab'] }
+        { tabProperties: ['tab'], cache }
       );
+      TSTAPI.clearCache(cache);
       if (allowed) {
-        window.lastActiveTab = successor.id;
+        win.lastActiveTab = successor.id;
         if (mMaybeTabSwitchingByShortcut)
           setupDelayedExpand(successor);
         TabsInternalOperation.activateTab(successor, { silently: true });
-        log('Tabs.onActivating: discarded? ', dumpTab(tab), tab && tab.discarded);
+        log('Tabs.onActivating: discarded? ', dumpTab(tab), tab?.discarded);
         if (tab.discarded)
-          tab.$TST.discardURLAfterCompletelyLoaded = tab.url;
+          tab.$TST.temporaryMetadata.set('discardURLAfterCompletelyLoaded', tab.url);
         return false;
       }
       else {
@@ -170,7 +207,11 @@ Tab.onActivating.addListener(async (tab, info = {}) => { // return false if the
            (!configs.autoCollapseExpandSubtreeOnSelect ||
             configs.autoCollapseExpandSubtreeOnSelectExceptActiveTabRemove)) {
     log('=> reaction for removing current tab');
-    window.lastActiveTab = tab.id;
+    win.lastActiveTab = tab.id;
+    tryHighlightBundledTab(tab, {
+      ...info,
+      shouldSkipCollapsed
+    });
     return true;
   }
   else if (tab.$TST.hasChild &&
@@ -179,8 +220,8 @@ Tab.onActivating.addListener(async (tab, info = {}) => { // return false if the
     log('=> reaction for newly active parent tab');
     await handleNewActiveTab(tab, info);
   }
-  delete tab.$TST.discardOnCompletelyLoaded;
-  window.lastActiveTab = tab.id;
+  tab.$TST.temporaryMetadata.delete('discardOnCompletelyLoaded');
+  win.lastActiveTab = tab.id;
 
   if (mMaybeTabSwitchingByShortcut)
     setupDelayedExpand(tab);
@@ -193,25 +234,26 @@ Tab.onActivating.addListener(async (tab, info = {}) => { // return false if the
   return true;
 });
 
-async function handleNewActiveTab(tab, info = {}) {
-  log('handleNewActiveTab: ', dumpTab(tab), info);
+async function handleNewActiveTab(tab, { allowed, silently } = {}) {
+  log('handleNewActiveTab: ', dumpTab(tab), { allowed, silently });
   const shouldCollapseExpandNow = configs.autoCollapseExpandSubtreeOnSelect;
   const canCollapseTree         = shouldCollapseExpandNow;
-  const canExpandTree           = shouldCollapseExpandNow && !info.silently;
+  const canExpandTree           = shouldCollapseExpandNow && !silently;
   if (canExpandTree &&
-      info.allowed !== false) {
+      allowed !== false) {
+    const cache = {};
     const allowed = await TSTAPI.tryOperationAllowed(
       tab.active ?
         TSTAPI.kNOTIFY_TRY_EXPAND_TREE_FROM_FOCUSED_PARENT :
         TSTAPI.kNOTIFY_TRY_EXPAND_TREE_FROM_FOCUSED_BUNDLED_PARENT,
-      { tab: new TSTAPI.TreeItem(tab) },
-      { tabProperties: ['tab'] }
+      { tab },
+      { tabProperties: ['tab'], cache }
     );
     if (!allowed)
       return;
     if (canCollapseTree &&
         configs.autoExpandIntelligently)
-      Tree.collapseExpandTreesIntelligentlyFor(tab, {
+      await Tree.collapseExpandTreesIntelligentlyFor(tab, {
         broadcast: true
       });
     else
@@ -222,67 +264,77 @@ async function handleNewActiveTab(tab, info = {}) {
   }
 }
 
-async function tryHighlightBundledTab(tab, info) {
+async function tryHighlightBundledTab(tab, { shouldSkipCollapsed, allowed, silently } = {}) {
   const bundledTab = tab.$TST.bundledTab;
   const oldBundledTabs = TabsStore.bundledActiveTabsInWindow.get(tab.windowId);
+  log('tryHighlightBundledTab ', {
+    tab: tab.id,
+    bundledTab: bundledTab?.id,
+    oldBundledTabs,
+    shouldSkipCollapsed,
+    allowed,
+    silently,
+  });
   for (const tab of oldBundledTabs.values()) {
     if (tab == bundledTab)
       continue;
-    tab.$TST.removeState(Constants.kTAB_STATE_BUNDLED_ACTIVE, { broadcast: true });
+    tab.$TST.removeState(Constants.kTAB_STATE_BUNDLED_ACTIVE);
   }
 
   if (!bundledTab)
     return;
 
-  bundledTab.$TST.addState(Constants.kTAB_STATE_BUNDLED_ACTIVE, { broadcast: true });
+  bundledTab.$TST.addState(Constants.kTAB_STATE_BUNDLED_ACTIVE);
 
   await wait(100);
   if (!tab.active || // ignore tab already inactivated while waiting
       tab.$TST.hasOtherHighlighted || // ignore manual highlighting
-      bundledTab.pinned)
+      bundledTab.pinned ||
+      !configs.syncActiveStateToBundledTabs)
     return;
 
   if (bundledTab.$TST.hasChild &&
       bundledTab.$TST.subtreeCollapsed &&
-      !info.shouldSkipCollapsed)
-    await handleNewActiveTab(bundledTab, info);
+      !shouldSkipCollapsed)
+    await handleNewActiveTab(bundledTab, { allowed, silently });
 }
 
 Tab.onUpdated.addListener((tab, changeInfo = {}) => {
   if ('url' in changeInfo) {
-    if (tab.$TST.discardURLAfterCompletelyLoaded &&
-        tab.$TST.discardURLAfterCompletelyLoaded != changeInfo.url)
-      delete tab.$TST.discardURLAfterCompletelyLoaded;
+    if (tab.$TST.temporaryMetadata.has('discardURLAfterCompletelyLoaded') &&
+        tab.$TST.temporaryMetadata.get('discardURLAfterCompletelyLoaded') != changeInfo.url)
+      tab.$TST.temporaryMetadata.delete('discardURLAfterCompletelyLoaded');
   }
 });
 
 Tab.onStateChanged.addListener(tab => {
-  if (tab.status != 'complete')
+  if (!tab ||
+      tab.status != 'complete')
     return;
 
   if (typeof browser.tabs.discard == 'function') {
-    if (tab.url == tab.$TST.discardURLAfterCompletelyLoaded &&
+    if (tab.url == tab.$TST.temporaryMetadata.get('discardURLAfterCompletelyLoaded') &&
         configs.autoDiscardTabForUnexpectedFocus) {
       log('Try to discard accidentally restored tab (on restored) ', dumpTab(tab));
       wait(configs.autoDiscardTabForUnexpectedFocusDelay).then(() => {
-        if (!TabsStore.ensureLivingTab(tab) ||
+        if (!TabsStore.ensureLivingItem(tab) ||
             tab.active)
           return;
         if (tab.status == 'complete')
           browser.tabs.discard(tab.id)
             .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
         else
-          tab.$TST.discardOnCompletelyLoaded = true;
+          tab.$TST.temporaryMetadata.set('discardOnCompletelyLoaded', true);
       });
     }
-    else if (tab.$TST.discardOnCompletelyLoaded && !tab.active) {
+    else if (tab.$TST.temporaryMetadata.has('discardOnCompletelyLoaded') && !tab.active) {
       log('Discard accidentally restored tab (on complete) ', dumpTab(tab));
       browser.tabs.discard(tab.id)
         .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
     }
   }
-  delete tab.$TST.discardURLAfterCompletelyLoaded;
-  delete tab.$TST.discardOnCompletelyLoaded;
+  tab.$TST.temporaryMetadata.delete('discardURLAfterCompletelyLoaded');
+  tab.$TST.temporaryMetadata.delete('discardOnCompletelyLoaded');
 });
 
 async function setupDelayedExpand(tab) {
@@ -290,30 +342,41 @@ async function setupDelayedExpand(tab) {
     return;
   cancelDelayedExpand(tab);
   TabsStore.removeToBeExpandedTab(tab);
+  const cache = {};
+  const [ctrlTabHandlingEnabled, allowedToExpandViaAPI] = await Promise.all([
+    Permissions.isGranted(Permissions.ALL_URLS),
+    TSTAPI.tryOperationAllowed(
+      TSTAPI.kNOTIFY_TRY_EXPAND_TREE_FROM_LONG_PRESS_CTRL_KEY,
+      { tab },
+      { tabProperties: ['tab'], cache }
+    ),
+  ]);
   if (!configs.autoExpandOnTabSwitchingShortcuts ||
       !tab.$TST.hasChild ||
       !tab.$TST.subtreeCollapsed ||
-      !(await TSTAPI.tryOperationAllowed(
-        TSTAPI.kNOTIFY_TRY_EXPAND_TREE_FROM_LONG_PRESS_CTRL_KEY,
-        { tab: new TSTAPI.TreeItem(tab) },
-        { tabProperties: ['tab'] }
-      )))
+      !ctrlTabHandlingEnabled ||
+      !allowedToExpandViaAPI)
     return;
   TabsStore.addToBeExpandedTab(tab);
-  tab.$TST.delayedExpand = setTimeout(() => {
+  tab.$TST.temporaryMetadata.set('delayedExpand', setTimeout(async () => {
+    if (!tab.$TST.temporaryMetadata.has('delayedExpand')) { // already canceled
+      log('delayed expand is already canceled ', tab.id);
+      return;
+    }
+    log('delayed expand by long-press of ctrl key on ', tab.id);
     TabsStore.removeToBeExpandedTab(tab);
-    Tree.collapseExpandTreesIntelligentlyFor(tab, {
+    await Tree.collapseExpandTreesIntelligentlyFor(tab, {
       broadcast: true
     });
-  }, configs.autoExpandOnTabSwitchingShortcutsDelay);
+  }, configs.autoExpandOnTabSwitchingShortcutsDelay));
 }
 
 function cancelDelayedExpand(tab) {
   if (!tab ||
-      !tab.$TST.delayedExpand)
+      !tab.$TST.temporaryMetadata.has('delayedExpand'))
     return;
-  clearTimeout(tab.$TST.delayedExpand);
-  delete tab.$TST.delayedExpand;
+  clearTimeout(tab.$TST.temporaryMetadata.get('delayedExpand'));
+  tab.$TST.temporaryMetadata.delete('delayedExpand');
   TabsStore.removeToBeExpandedTab(tab);
 }
 
@@ -324,26 +387,28 @@ function cancelAllDelayedExpand(windowId) {
 }
 
 Tab.onCollapsedStateChanged.addListener((tab, info = {}) => {
-  if (info.collapsed)
-    tab.$TST.addState(Constants.kTAB_STATE_COLLAPSED_DONE);
-  else
-    tab.$TST.removeState(Constants.kTAB_STATE_COLLAPSED_DONE);
+  tab.$TST.toggleState(Constants.kTAB_STATE_COLLAPSED_DONE, info.collapsed, { broadcast: false });
 });
 
 
-Background.onInit.addListener(() => {
-  browser.windows.onFocusChanged.addListener(() => {
-    mMaybeTabSwitchingByShortcut = false;
-  });
+Background.onReady.addListener(() => {
+  for (const tab of Tab.getAllTabs(null, { iterator: true })) {
+    tab.$TST.removeState(Constants.kTAB_STATE_BUNDLED_ACTIVE);
+  }
+  for (const tab of Tab.getActiveTabs({ iterator: true })) {
+    tryHighlightBundledTab(tab);
+  }
 });
 
-Background.onBuilt.addListener(() => {
-  browser.runtime.onMessage.addListener(onMessage);
+browser.windows.onFocusChanged.addListener(() => {
+  mMaybeTabSwitchingByShortcut = false;
 });
 
+browser.runtime.onMessage.addListener(onMessage);
 
 function onMessage(message, sender) {
-  if (!message ||
+  if (mInitializationPhase < PHASE_BACKGROUND_BUILT ||
+      !message ||
       typeof message.type != 'string')
     return;
 
@@ -354,35 +419,44 @@ function onMessage(message, sender) {
         mTabSwitchedByShortcut = false;
       break;
 
-    case Constants.kCOMMAND_NOTIFY_START_TAB_SWITCH: {
-      log('Constants.kCOMMAND_NOTIFY_START_TAB_SWITCH');
+    case Constants.kCOMMAND_NOTIFY_MAY_START_TAB_SWITCH: {
+      if (message.modifier != (configs.accelKey || (isMacOS() ? 'meta' : 'control')))
+        return;
+      log('kCOMMAND_NOTIFY_MAY_START_TAB_SWITCH ', message.modifier);
       mMaybeTabSwitchingByShortcut = true;
-      if (sender.tab && sender.tab.active) {
-        const window = TabsStore.windows.get(sender.tab.windowId);
-        window.lastActiveTab = sender.tab.id;
+      if (sender.tab?.active) {
+        const win = TabsStore.windows.get(sender.tab.windowId);
+        win.lastActiveTab = sender.tab.id;
       }
+      if (sender.tab)
+        mLastTabsCountInWindow.set(sender.tab.windowId, Tab.getAllTabs(sender.tab.windowId).length);
     }; break;
-    case Constants.kCOMMAND_NOTIFY_END_TAB_SWITCH:
-      log('Constants.kCOMMAND_NOTIFY_END_TAB_SWITCH');
+    case Constants.kCOMMAND_NOTIFY_MAY_END_TAB_SWITCH:
+      if (message.modifier != (configs.accelKey || (isMacOS() ? 'meta' : 'control')))
+        return;
+      log('kCOMMAND_NOTIFY_MAY_END_TAB_SWITCH ', message.modifier);
       return (async () => {
         if (mTabSwitchedByShortcut &&
             configs.skipCollapsedTabsForTabSwitchingShortcuts &&
             sender.tab) {
           await Tab.waitUntilTracked(sender.tab.id);
-          let tab = sender.tab && Tab.get(sender.tab.id);
+          let tab = Tab.get(sender.tab.id);
           if (!tab) {
-            const tabs = await browser.tabs.query({ currentWindow: true, active: true }).catch(ApiTabs.createErrorHandler());
+            let tabs = await browser.tabs.query({ currentWindow: true, active: true }).catch(ApiTabs.createErrorHandler());
+            if (tabs.length == 0)
+              tabs = await browser.tabs.query({ currentWindow: true }).catch(ApiTabs.createErrorHandler());
             await Tab.waitUntilTracked(tabs[0].id);
             tab = Tab.get(tabs[0].id);
           }
           cancelAllDelayedExpand(tab.windowId);
+          const cache = {};
           if (configs.autoCollapseExpandSubtreeOnSelect &&
               tab &&
               TabsStore.windows.get(tab.windowId).lastActiveTab == tab.id &&
               (await TSTAPI.tryOperationAllowed(
                 TSTAPI.kNOTIFY_TRY_EXPAND_TREE_FROM_END_TAB_SWITCH,
-                { tab: new TSTAPI.TreeItem(tab) },
-                { tabProperties: ['tab'] }
+                { tab },
+                { tabProperties: ['tab'], cache }
               ))) {
             Tree.collapseExpandSubtree(tab, {
               collapsed: false,
@@ -395,16 +469,3 @@ function onMessage(message, sender) {
       })();
   }
 }
-
-SidebarConnection.onMessage.addListener(async (windowId, message) => {
-  switch (message.type) {
-    case Constants.kCOMMAND_SELECT_TAB: {
-      await Tab.waitUntilTracked(message.tabId);
-      const tab = Tab.get(message.tabId);
-      if (!tab)
-        return;
-      browser.tabs.update(tab.id, { active: true })
-        .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
-    }; break;
-  }
-});
