@@ -8,10 +8,9 @@
 import {
   log as internalLogger,
 } from '/common/common.js';
-
 import * as ApiTabs from '/common/api-tabs.js';
 
-import Tab from '/common/Tab.js';
+import { Tab } from '/common/TreeItem.js';
 
 function log(...args) {
   internalLogger('background/handle-tab-multiselect', ...args);
@@ -22,7 +21,7 @@ Tab.onUpdated.addListener((tab, info, options = {}) => {
       !tab.$TST.subtreeCollapsed ||
       tab.$TST.collapsed ||
       !tab.$TST.multiselected ||
-      options.inheritHighlighted === false)
+      !options.inheritHighlighted)
     return;
 
   const collapsedDescendants = tab.$TST.descendants;
