@@ -7,13 +7,14 @@
 
 import {
   log as internalLogger,
-  configs
+  configs,
+  shouldApplyAnimation
 } from '/common/common.js';
 
 import * as Constants from '/common/constants.js';
 import * as TabsStore from '/common/tabs-store.js';
 
-import Tab from '/common/Tab.js';
+import { Tab } from '/common/TreeItem.js';
 
 import * as BackgroundConnection from './background-connection.js';
 import * as CollapseExpand from './collapse-expand.js';
@@ -42,12 +43,6 @@ export function init() {
   mPromisedInitialized = mPromisedInitializedResolver = null;
 }
 
-const mLessAnimationMedia = window.matchMedia('(prefers-reduced-motion: reduce)');
-
-function canApplyAnimation() {
-  return configs.animation  && !mLessAnimationMedia.matches;
-}
-
 export function updateRestoredTree(cachedIndent) {
   updateVisualMaxTreeLevel();
   update({
@@ -59,7 +54,7 @@ export function updateRestoredTree(cachedIndent) {
 export function update(options = {}) {
   if (!options.cache) {
     const maxLevel  = getMaxTreeLevel(mTargetWindow);
-    const maxIndent = mTabBar.getBoundingClientRect().width * (0.33);
+    const maxIndent = mTabBar.offsetWidth * (0.33);
     if (maxLevel <= mLastMaxLevel &&
         maxIndent == mLastMaxIndent &&
         !options.force)
@@ -79,10 +74,8 @@ export function update(options = {}) {
     document.head.appendChild(mIndentDefinition);
   }
 
-  const indentProp = (configs.sidebarPosition == Constants.kTABBAR_POSITION_RIGHT) ? 'margin-right' : 'margin-left';
   if (options.cache &&
-      options.cache.definition &&
-      options.cache.definition.includes(indentProp)) {
+      options.cache.definition) {
     mIndentDefinition.textContent = options.cache.definition;
   }
   else {
@@ -98,7 +91,7 @@ export function update(options = {}) {
       const indents = Object.keys(indentSet);
       indents.sort((aA, aB) => parseInt(aA) - parseInt(aB));
       for (const indent of indents) {
-        definitions.push(`${indentSet[indent].join(',\n')} { ${indentProp}: ${indent}; }`);
+        definitions.push(`${indentSet[indent].join(',\n')} { --tab-indent: ${indent}; }`);
       }
     }
     mIndentDefinition.textContent = indentUnitDefinitions.concat(definitions).join('\n');
@@ -111,7 +104,7 @@ function generateIndentAndSelectorsForMaxLevel(maxLevel, indentToSelectors, defa
   if (configuredMaxLevel < 0)
     configuredMaxLevel = Number.MAX_SAFE_INTEGER;
 
-  const base = `:root[${Constants.kMAX_TREE_LEVEL}="${maxLevel}"]:not(.initializing) .tab:not(.${Constants.kTAB_STATE_COLLAPSED_DONE})[${Constants.kLEVEL}]`;
+  const base = `:root[${Constants.kMAX_TREE_LEVEL}="${maxLevel}"]:not(.initializing) .tab:not(.${Constants.kTAB_STATE_PINNED}):not(.${Constants.kTAB_STATE_COLLAPSED_DONE})[${Constants.kLEVEL}]`;
 
   // default indent for unhandled (deep) level tabs
   const defaultIndent = `${Math.min(maxLevel + 1, configuredMaxLevel) * indentUnit}px`;
@@ -144,27 +137,50 @@ export function getCacheInfo() {
 }
 
 
-export async function reserveToUpdateVisualMaxTreeLevel() {
-  if (mPromisedInitialized)
-    await mPromisedInitialized;
-  log('reserveToUpdateVisualMaxTreeLevel');
+export function tryUpdateVisualMaxTreeLevel() {
+  log('tryUpdateVisualMaxTreeLevel');
   if (updateVisualMaxTreeLevel.waiting) {
     clearTimeout(updateVisualMaxTreeLevel.waiting);
     delete updateVisualMaxTreeLevel.waiting;
   }
 
-  if (!canApplyAnimation()) {
+  tryUpdateVisualMaxTreeLevel.calledCount++;
+
+  const animation = shouldApplyAnimation();
+
+  // On no-animation mode, we should update max indent level immediately
+  // as possible as we can without delay, to reduce visual flicking which
+  // can trigger an epileptic seizure.
+  // But we also have to reduce needless function calls for better performance.
+  // This threshold is a safe guard for uncared cases with too many call
+  // of updateVisualMaxTreeLevel().
+  // See also: https://github.com/piroor/treestyletab/issues/3383
+  if (tryUpdateVisualMaxTreeLevel.calledCount <= configs.maxAllowedImmediateRefreshCount &&
+      !animation) {
     updateVisualMaxTreeLevel();
+    if (tryUpdateVisualMaxTreeLevel.waitingToResetCalledCount)
+      clearTimeout(tryUpdateVisualMaxTreeLevel.waitingToResetCalledCount);
+    tryUpdateVisualMaxTreeLevel.waitingToResetCalledCount = setTimeout(() => {
+      delete tryUpdateVisualMaxTreeLevel.waitingToResetCalledCount;
+      tryUpdateVisualMaxTreeLevel.calledCount = 0;
+    }, 0);
     return;
   }
 
+  const delay = animation ? Math.max(0, configs.collapseDuration) * 1.5 : 0;
+
   updateVisualMaxTreeLevel.waiting = setTimeout(() => {
     delete updateVisualMaxTreeLevel.waiting;
+    tryUpdateVisualMaxTreeLevel.calledCount = 0;
     updateVisualMaxTreeLevel();
-  }, configs.collapseDuration * 1.5);
+  }, delay);
 }
+tryUpdateVisualMaxTreeLevel.calledCount = 0;
+
+async function updateVisualMaxTreeLevel() {
+  if (mPromisedInitialized)
+    await mPromisedInitialized;
 
-function updateVisualMaxTreeLevel() {
   const maxLevel = getMaxTreeLevel(mTargetWindow, {
     onlyVisible: configs.indentAutoShrinkOnlyForVisible
   });
@@ -190,17 +206,33 @@ async function reserveToUpdateIndent() {
   log('reserveToUpdateIndent');
   if (reserveToUpdateIndent.waiting)
     clearTimeout(reserveToUpdateIndent.waiting);
+  const delay = shouldApplyAnimation() ? Math.max(configs.indentDuration, configs.collapseDuration) * 1.5 : 100;
   reserveToUpdateIndent.waiting = setTimeout(() => {
     delete reserveToUpdateIndent.waiting;
     update();
-  }, Math.max(configs.indentDuration, configs.collapseDuration) * 1.5);
+  }, delay);
 }
 
 
-CollapseExpand.onUpdated.addListener((_tab, _options) => {
-  if (configs.indentAutoShrink &&
-      configs.indentAutoShrinkOnlyForVisible)
-    reserveToUpdateVisualMaxTreeLevel();
+const restVisibilityChangedTabIds = new Set();
+
+CollapseExpand.onUpdated.addListener((tab, _options) => {
+  const isFinishBatch = restVisibilityChangedTabIds.has(tab.id);
+  restVisibilityChangedTabIds.delete(tab.id);
+
+  if ((configs.indentAutoShrink &&
+       configs.indentAutoShrinkOnlyForVisible) ||
+      // On no-animation mode, we should update max indent level immediately
+      // as possible as we can without delay, to reduce visual flicking which
+      // can trigger an epileptic seizure.
+      // But we also have to reduce needless function calls for better performance.
+      // So we throttle the function call of updateVisualMaxTreeLevel() until
+      // collapsed state of all tabs notified with "kCOMMAND_NOTIFY_SUBTREE_COLLAPSED_STATE_CHANGED"
+      // are completely updated.
+      // See also: https://github.com/piroor/treestyletab/issues/3383
+      (isFinishBatch &&
+       restVisibilityChangedTabIds.size == 0))
+    tryUpdateVisualMaxTreeLevel();
 });
 
 const BUFFER_KEY_PREFIX = 'indent-';
@@ -210,7 +242,7 @@ BackgroundConnection.onMessage.addListener(async message => {
     case Constants.kCOMMAND_NOTIFY_TAB_CREATED:
     case Constants.kCOMMAND_NOTIFY_TAB_REMOVING:
       log('listen: ', message.type);
-      reserveToUpdateVisualMaxTreeLevel();
+      tryUpdateVisualMaxTreeLevel();
       break;
 
     case Constants.kCOMMAND_NOTIFY_TAB_SHOWN:
@@ -218,27 +250,35 @@ BackgroundConnection.onMessage.addListener(async message => {
     case Constants.kCOMMAND_NOTIFY_CHILDREN_CHANGED:
       log('listen: ', message.type);
       reserveToUpdateIndent();
-      reserveToUpdateVisualMaxTreeLevel();
+      tryUpdateVisualMaxTreeLevel();
       break;
 
     case Constants.kCOMMAND_NOTIFY_TAB_LEVEL_CHANGED: {
       if (BackgroundConnection.handleBufferedMessage(message, `${BUFFER_KEY_PREFIX}${message.tabId}`))
         return;
-      await Tab.waitUntilTracked(message.tabId, { element: true });
+      await Tab.waitUntilTracked(message.tabId);
       const tab = Tab.get(message.tabId);
       const lastMessage = BackgroundConnection.fetchBufferedMessage(message.type, `${BUFFER_KEY_PREFIX}${message.tabId}`);
       log('listen: ', message.type, tab, lastMessage);
       if (!tab ||
           !lastMessage)
         return;
-      if (tab.$TST.getAttribute(Constants.kLEVEL) != lastMessage.level)
+      if (tab.$TST.getAttribute(Constants.kLEVEL) != lastMessage.level) {
         tab.$TST.setAttribute(Constants.kLEVEL, lastMessage.level);
+        tryUpdateVisualMaxTreeLevel();
+      }
       reserveToUpdateIndent();
     }; break;
 
+    case Constants.kCOMMAND_NOTIFY_SUBTREE_COLLAPSED_STATE_CHANGED:
+      for (const id of message.visibilityChangedTabIds) {
+        restVisibilityChangedTabIds.add(id);
+      }
+      break;
+
     case Constants.kCOMMAND_NOTIFY_TAB_COLLAPSED_STATE_CHANGED:
-      if (!canApplyAnimation())
-        updateVisualMaxTreeLevel();
+      if (!restVisibilityChangedTabIds.has(message.tabId))
+        tryUpdateVisualMaxTreeLevel();
       break;
   }
 });
