@@ -3,26 +3,28 @@
 # License, v. 2.0. If a copy of the MPL was not distributed with this
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
+'use strict';
 
 import {
   log,
   configs
 } from '/common/common.js';
-
-import * as MetricsData from '/common/metrics-data.js';
-import * as TabsStore from '/common/tabs-store.js';
 import * as SidebarConnection from '/common/sidebar-connection.js';
-import Tab from '/common/Tab.js';
+import * as TabsStore from '/common/tabs-store.js';
+
+import MetricsData from '/common/MetricsData.js';
+import { Tab } from '/common/TreeItem.js';
 
 import * as Background from './background.js';
+import './handle-misc.js';
+import './handle-moved-tabs.js';
 import './handle-new-tabs.js';
 import './handle-removed-tabs.js';
+import './handle-tab-bunches.js';
 import './handle-tab-focus.js';
 import './handle-tab-multiselect.js';
-import './handle-moved-tabs.js';
 import './handle-tree-changes.js';
-import './handle-group-tabs.js';
-import './handle-misc.js';
+import './sync-background.js';
 
 log.context = 'BG';
 
