@@ -3,6 +3,8 @@
 # License, v. 2.0. If a copy of the MPL was not distributed with this
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
+'use strict';
+
 import '/extlib/l10n.js';
 
 import {
@@ -10,16 +12,18 @@ import {
   configs
 } from '/common/common.js';
 
-import * as MetricsData from '/common/metrics-data.js';
 import * as TabsStore from '/common/tabs-store.js';
-import Tab from '/common/Tab.js';
 
-import * as Sidebar from './sidebar.js';
+import MetricsData from '/common/MetricsData.js';
+import { Tab } from '/common/TreeItem.js';
+
 import * as BackgroundConnection from './background-connection.js';
-import './mouse-event-listener.js';
+import * as Sidebar from './sidebar.js';
 import './collapse-expand.js';
-
+import './mouse-event-listener.js';
 import './tab-context-menu.js';
+import './tab-preview-tooltip.js';
+import './tst-api-frontend.js';
 
 log.context = 'Sidebar-?';
 
