@@ -1,13 +1,12 @@
 /*
- license: The MIT License, Copyright (c) 2016-2019 YUKI "Piro" Hiroshi
+ license: The MIT License, Copyright (c) 2016-2024 YUKI "Piro" Hiroshi
  original:
    http://github.com/piroor/webextensions-lib-l10n
 */
 
 var l10n = {
   updateString(string) {
-    return string.replace(/__MSG_(.+?)__/g, matched => {
-      const key = matched.slice(6, -2);
+    return string.replace(/__MSG_([-@\.\w]+)__/g, (matched, key) => {
       return chrome.i18n.getMessage(key) || matched;
     });
   },
