@@ -5,19 +5,21 @@
 */
 'use strict';
 
+import EventListenerManager from '/extlib/EventListenerManager.js';
+import MenuUI from '/extlib/MenuUI.js';
+
 import {
   log as internalLogger,
   wait,
-  configs
+  configs,
+  shouldApplyAnimation,
+  compareAsNumber,
 } from '/common/common.js';
-import * as Constants from '/common/constants.js';
 import * as ApiTabs from '/common/api-tabs.js';
+import * as Constants from '/common/constants.js';
 import * as TabsStore from '/common/tabs-store.js';
 import * as TSTAPI from '/common/tst-api.js';
 
-import EventListenerManager from '/extlib/EventListenerManager.js';
-import MenuUI from '/extlib/MenuUI.js';
-
 import * as BackgroundConnection from './background-connection.js';
 import * as Size from './size.js';
 
@@ -30,7 +32,6 @@ export const onResized = new EventListenerManager();
 let mTargetWindow;
 let mInitialized = false;
 
-const mTabBarContainer = document.querySelector('#tabbar-container');
 const mContainer       = document.querySelector('#subpanel-container');
 const mHeader          = document.querySelector('#subpanel-header');
 const mSelector        = document.querySelector('#subpanel-selector');
@@ -60,7 +61,12 @@ export async function init() {
     browser.sessions.getWindowValue(mTargetWindow, Constants.kWINDOW_STATE_SUBPANEL_PROVIDER_ID).catch(ApiTabs.createErrorHandler()),
     browser.sessions.getWindowValue(mTargetWindow, Constants.kWINDOW_STATE_SUBPANEL_HEIGHT).catch(ApiTabs.createErrorHandler())
   ]);
-  mHeight = typeof height == 'number' ? height : Math.max(configs.lastSubPanelHeight, 0);
+  const providerSpecificHeight = providerId && await browser.sessions.getWindowValue(mTargetWindow, `${Constants.kWINDOW_STATE_SUBPANEL_HEIGHT}:${providerId}`).catch(ApiTabs.createErrorHandler());
+  mHeight = (typeof providerSpecificHeight == 'number') ?
+    providerSpecificHeight :
+    (typeof height == 'number') ?
+      height :
+      Math.max(configs.lastSubPanelHeight, 0);
 
   log('initialize ', { providerId, height: mHeight });
 
@@ -75,7 +81,7 @@ export async function init() {
     //log('onMessage: ', message, sender);
     switch (message.type) {
       case TSTAPI.kCOMMAND_BROADCAST_API_REGISTERED:
-        wait(0).then(() => { // wait until addons are updated
+        wait(0).then(async () => { // wait until addons are updated
           updateSelector();
           const provider = TSTAPI.getAddon(message.sender.id);
           if (provider &&
@@ -84,7 +90,7 @@ export async function init() {
                provider.newlyInstalled)) {
             if (mHeight == 0)
               mHeight = getDefaultHeight();
-            applyProvider(provider.id);
+            await applyProvider(provider.id);
           }
         });
         break;
@@ -108,18 +114,49 @@ TSTAPI.onInitialized.addListener(async () => {
 
   const providerId = await browser.sessions.getWindowValue(mTargetWindow, Constants.kWINDOW_STATE_SUBPANEL_PROVIDER_ID).catch(ApiTabs.createErrorHandler());
   if (providerId)
-    applyProvider(providerId);
+    await applyProvider(providerId);
   else
     restoreLastProvider();
 });
 
-function applyProvider(id) {
+function getProviderIconUrl(provider) {
+  if (!provider.icons || typeof provider.icons != 'object')
+    return null;
+
+  if ('16' in provider.icons)
+    return provider.icons['16'];
+
+  const sizes = Object.keys(provider.icons, size => parseInt(size)).sort(compareAsNumber);
+  if (sizes.length == 0)
+    return null;
+
+  // find a size most nearest to 16
+  sizes.sort((a, b) => {
+    if (a < 16) {
+      if (b >= 16)
+        return 1;
+      else
+        return b - a;
+    }
+    if (b < 16) {
+      if (a >= 16)
+        return -1;
+      else
+        return b - a;
+    }
+    return a - b;
+  });
+  return provider.icons[sizes[0]];
+}
+
+async function applyProvider(id) {
   const provider = TSTAPI.getAddon(id);
   log('applyProvider ', id, provider);
   if (provider &&
       provider.subPanel) {
     log('applyProvider: load ', id);
     configs.lastSelectedSubPanelProviderId = mProviderId = id;
+    const lastHeight = await browser.sessions.getWindowValue(mTargetWindow, `${Constants.kWINDOW_STATE_SUBPANEL_HEIGHT}:${id}`).catch(ApiTabs.createErrorHandler());
     for (const item of mSelector.querySelectorAll('.radio')) {
       item.classList.remove('checked');
     }
@@ -129,13 +166,38 @@ function applyProvider(id) {
     browser.sessions.setWindowValue(mTargetWindow, Constants.kWINDOW_STATE_SUBPANEL_PROVIDER_ID, id).catch(ApiTabs.createErrorHandler());
 
     const icon = mSelectorAnchor.querySelector('.icon > img');
-    if (provider.icons && provider.icons['16'])
-      icon.src = provider.icons['16'];
+    const iconUrl = getProviderIconUrl(provider);
+    if (iconUrl)
+      icon.src = iconUrl;
     else
       icon.removeAttribute('src');
 
     mSelectorAnchor.querySelector('.label').textContent = provider.subPanel.title || provider.name || provider.id;
 
+    if ('fixedHeight' in provider.subPanel) {
+      if (typeof provider.subPanel.fixedHeight == 'number')
+        mHeight = provider.subPanel.fixedHeight;
+      else
+        mHeight = Size.calc(provider.subPanel.fixedHeight);
+      mHeader.classList.remove('resizable');
+    }
+    else {
+      mHeader.classList.add('resizable');
+      if (typeof lastHeight == 'number') {
+        mHeight = lastHeight;
+      }
+      else if ('initialHeight' in provider.subPanel) {
+        if (typeof provider.subPanel.initialHeight == 'number')
+          mHeight = provider.subPanel.initialHeight;
+        else
+          mHeight = Size.calc(provider.subPanel.initialHeight);
+      }
+    }
+
+    const url = new URL(provider.subPanel.url);
+    url.searchParams.set('windowId', mTargetWindow);
+    provider.subPanel.url = url.href;
+
     if (mHeight > 0)
       load(provider.subPanel);
     else
@@ -146,19 +208,20 @@ function applyProvider(id) {
     const icon = mSelectorAnchor.querySelector('.icon > img');
     icon.removeAttribute('src');
     mSelectorAnchor.querySelector('.label').textContent = '';
+    mHeader.classList.add('resizable');
     load();
   }
 }
 
-function restoreLastProvider() {
+async function restoreLastProvider() {
   const lastProvider = TSTAPI.getAddon(configs.lastSelectedSubPanelProviderId);
   log('restoreLastProvider ', lastProvider);
-  if (lastProvider && lastProvider.subPanel)
-    applyProvider(lastProvider.id);
+  if (lastProvider?.subPanel)
+    await applyProvider(lastProvider.id);
   else if (mSelector.hasChildNodes())
-    applyProvider(mSelector.firstChild.dataset.value);
+    await applyProvider(mSelector.firstChild.dataset.value);
   else
-    applyProvider(mProviderId = null);
+    await applyProvider(mProviderId = null);
 }
 
 function getDefaultHeight() {
@@ -180,24 +243,22 @@ async function load(params) {
 function updateLayout() {
   if (!mProviderId && !mSelector.hasChildNodes()) {
     mContainer.classList.add('collapsed');
-    mContainer.style.visibility = mSubPanel.style.visibility = 'collapse';
-    mTabBarContainer.style.bottom = 0;
+    document.documentElement.style.setProperty('--subpanel-area-size', '0px');
   }
   else {
     mHeight = Math.max(0, mHeight);
     mContainer.classList.toggle('collapsed', mHeight == 0);
-    mContainer.style.visibility = mSubPanel.style.visibility = 'visible';
-    const headerSize = mHeader.getBoundingClientRect().height;
-    const appliedHeight = Math.min(window.innerHeight * 0.66, mHeight);
-    mContainer.style.height = `${appliedHeight + headerSize}px`;
-    mSubPanel.style.height = `${appliedHeight}px`;
-    mTabBarContainer.style.bottom = `${appliedHeight + headerSize}px`;
+    const headerSize = mHeader.offsetHeight;
+    const maxHeight = window.innerHeight * Math.max(0, Math.min(1, configs.maxSubPanelSizeRatio));
+    const appliedHeight = Math.min(maxHeight, mHeight);
+    document.documentElement.style.setProperty('--subpanel-content-size', `${appliedHeight}px`);
+    document.documentElement.style.setProperty('--subpanel-area-size', `${appliedHeight + headerSize}px`);
 
     if (mHeight > 0 &&
         (!mSubPanel.src || mSubPanel.src == 'about:blank')) {
       // delayed load
       const provider = TSTAPI.getAddon(mProviderId);
-      if (provider && provider.subPanel)
+      if (provider?.subPanel)
         mSubPanel.src = provider.subPanel.url;
     }
     else if (mHeight == 0) {
@@ -210,8 +271,10 @@ function updateLayout() {
 
   onResized.dispatch();
 
-  if (mProviderId)
+  if (mProviderId) {
     browser.sessions.setWindowValue(mTargetWindow, Constants.kWINDOW_STATE_SUBPANEL_HEIGHT, mHeight).catch(ApiTabs.createErrorHandler());
+    browser.sessions.setWindowValue(mTargetWindow, `${Constants.kWINDOW_STATE_SUBPANEL_HEIGHT}:${mProviderId}`, mHeight).catch(ApiTabs.createErrorHandler());
+  }
 }
 
 async function toggle() {
@@ -234,6 +297,11 @@ function isFiredOnClickable(event) {
   return !!target.closest('.clickable');
 }
 
+function isResizable() {
+  const provider = mProviderId && TSTAPI.getAddon(mProviderId);
+  return !provider || !provider.subPanel || !('fixedHeight' in provider.subPanel);
+}
+
 mHeader.addEventListener('mousedown', event => {
   if (isFiredOnClickable(event))
     return;
@@ -252,6 +320,8 @@ mHeader.addEventListener('mouseup', event => {
   event.stopPropagation();
   event.preventDefault();
   document.releaseCapture();
+  if (!isResizable())
+    return;
   mHeight = mDragStartHeight - (event.clientY - mDragStartY);
   updateLayout();
   saveLastHeight();
@@ -280,6 +350,8 @@ window.addEventListener('resize', _event => {
 function onMouseMove(event) {
   event.stopPropagation();
   event.preventDefault();
+  if (!isResizable())
+    return;
   mHeight = mDragStartHeight - (event.clientY - mDragStartY);
   updateLayout();
 }
@@ -302,8 +374,9 @@ function updateSelector() {
     const iconContainer = item.appendChild(document.createElement('span'));
     iconContainer.classList.add('icon');
     const icon = iconContainer.appendChild(document.createElement('img'));
-    if (addon.icons && addon.icons['16'])
-      icon.src = addon.icons['16'];
+    const url = getProviderIconUrl(addon);
+    if (url)
+      icon.src = url;
     item.appendChild(document.createTextNode(addon.subPanel.title || addon.name || id));
     items.push(item);
   }
@@ -323,12 +396,12 @@ mSelector.ui = new MenuUI({
   root:       mSelector,
   appearance: 'panel',
   onCommand:  onSelect,
-  animationDuration: configs.animation ? configs.collapseDuration : 0.001
+  animationDuration: shouldApplyAnimation() ? configs.collapseDuration : 0.001
 });
 
-function onSelect(item, _event) {
+async function onSelect(item, _event) {
   if (item.dataset.value) {
-    applyProvider(item.dataset.value);
+    await applyProvider(item.dataset.value);
     saveLastHeight();
   }
   mSelector.ui.close();
@@ -345,12 +418,12 @@ BackgroundConnection.onMessage.addListener(async message => {
       break;
 
     case Constants.kCOMMAND_INCREASE_SUBPANEL:
-      mHeight += Size.getTabHeight();
+      mHeight += Size.getRenderedTabHeight();
       updateLayout();
       break;
 
     case Constants.kCOMMAND_DECREASE_SUBPANEL:
-      mHeight -= Size.getTabHeight();
+      mHeight -= Size.getRenderedTabHeight();
       updateLayout();
       break;
   }
