@@ -11,20 +11,27 @@ import {
   log as internalLogger,
   wait,
   notify,
-  configs
+  configs,
+  shouldApplyAnimation,
+  compareAsNumber,
+  isLinux,
+  isMacOS,
 } from '/common/common.js';
-import * as Constants from '/common/constants.js';
 import * as ApiTabs from '/common/api-tabs.js';
+import * as BackgroundConnection from './background-connection.js';
+import * as Constants from '/common/constants.js';
+import * as ContextualIdentities from '/common/contextual-identities.js';
+import * as EventUtils from './event-utils.js';
+import * as Permissions from '/common/permissions.js';
 import * as TabsStore from '/common/tabs-store.js';
 import * as TSTAPI from '/common/tst-api.js';
-import * as Permissions from '/common/permissions.js';
-import * as EventUtils from './event-utils.js';
-import * as BackgroundConnection from './background-connection.js';
 
-import Tab from '/common/Tab.js';
+import { Tab, TabGroup, TreeItem } from '/common/TreeItem.js';
 
 import EventListenerManager from '/extlib/EventListenerManager.js';
 
+import * as TabGroupContextMenu from './tab-group-context-menu.js';
+
 function log(...args) {
   internalLogger('sidebar/tab-context-menu', ...args);
 }
@@ -34,6 +41,9 @@ export const onTabsClosing = new EventListenerManager();
 let mUI;
 let mMenu;
 
+let mNewTabButtonUI;
+let mNewTabButtonMenu;
+
 let mContextTab      = null;
 let mLastOpenOptions = null;
 let mIsDirty         = false;
@@ -42,17 +52,26 @@ const mExtraItems = new Map();
 
 export function init() {
   mMenu = document.querySelector('#tabContextMenu');
+  mNewTabButtonMenu = document.querySelector('#newTabButtonContextMenu');
   document.addEventListener('contextmenu', onContextMenu, { capture: true });
 
+  const commonOptions = {
+    appearance:        'menu',
+    animationDuration: shouldApplyAnimation() ? configs.collapseDuration : 0.001,
+    subMenuOpenDelay:  configs.subMenuOpenDelay,
+    subMenuCloseDelay: configs.subMenuCloseDelay,
+  };
   mUI = new MenuUI({
+    ...commonOptions,
     root: mMenu,
     onCommand,
     //onShown,
     onHidden,
-    appearance:        'menu',
-    animationDuration: configs.animation ? configs.collapseDuration : 0.001,
-    subMenuOpenDelay:  configs.subMenuOpenDelay,
-    subMenuCloseDelay: configs.subMenuCloseDelay
+  });
+  mNewTabButtonUI = new MenuUI({
+    ...commonOptions,
+    root: mNewTabButtonMenu,
+    onCommand: onNewTabButtonMenuCommand,
   });
 
   browser.runtime.onMessage.addListener(onMessage);
@@ -64,6 +83,8 @@ export function init() {
     importExtraItems(items);
     mIsDirty = true;
   }).catch(ApiTabs.createErrorSuppressor());
+
+  updateContextualIdentitiesSelector();
 }
 
 async function rebuild() {
@@ -190,7 +211,7 @@ function getAddonIcon(id) {
 function chooseIconForAddon(params) {
   const icons = params.icons || {};
   const addon = TSTAPI.getAddon(params.id) || {};
-  let sizes = Object.keys(icons).map(aSize => parseInt(aSize)).sort();
+  let sizes = Object.keys(icons).map(aSize => parseInt(aSize)).sort(compareAsNumber);
   const reducedSizes = sizes.filter(aSize => aSize < 16);
   if (reducedSizes.length > 0)
     sizes = reducedSizes;
@@ -266,7 +287,7 @@ function matchPatternToRegExp(pattern) {
 export async function open(options = {}) {
   await close();
   mLastOpenOptions = options;
-  mContextTab      = options.tab && Tab.get(options.tab.id);
+  mContextTab      = Tab.get(options.tab?.id);
   await rebuild();
   if (mIsDirty) {
     return await open(options);
@@ -328,7 +349,7 @@ async function onCommand(item, event) {
     modifiers.push('Command');
   if (event.ctrlKey) {
     modifiers.push('Ctrl');
-    if (/^Mac/i.test(navigator.platform))
+    if (isMacOS())
       modifiers.push('MacCtrl');
   }
   if (event.shiftKey)
@@ -336,7 +357,6 @@ async function onCommand(item, event) {
   const owner      = item.getAttribute('data-item-owner-id');
   const checked    = item.matches('.radio, .checkbox:not(.checked)');
   const wasChecked = item.matches('.radio.checked, .checkbox.checked');
-  const tab        = contextTab && (await (new TSTAPI.TreeItem(contextTab, { isContextTab: true })).exportFor(owner)) || null
   const message = {
     type: TSTAPI.kCONTEXT_MENU_CLICK,
     info: {
@@ -354,18 +374,27 @@ async function onCommand(item, event) {
       srcUrl:           null,
       wasChecked
     },
-    tab
+    tab: contextTab?.$TST.sanitized || contextTab,
   };
   if (owner == browser.runtime.id) {
     await browser.runtime.sendMessage(message).catch(ApiTabs.createErrorSuppressor());
   }
   else if (TSTAPI.isSafeAtIncognito(owner, { tab: contextTab, windowId: TabsStore.getCurrentWindowId() })) {
+    const cache = {};
     await Promise.all([
-      browser.runtime.sendMessage(owner, message).catch(ApiTabs.createErrorSuppressor()),
-      browser.runtime.sendMessage(owner, {
-        ...message,
-        type: TSTAPI.kFAKE_CONTEXT_MENU_CLICK
-      }).catch(ApiTabs.createErrorSuppressor())
+      TSTAPI.sendMessage(
+        owner,
+        message,
+        { tabProperties: ['tab'], cache, isContextTab: true }
+      ).catch(ApiTabs.createErrorSuppressor()),
+      TSTAPI.sendMessage(
+        owner,
+        {
+          ...message,
+          type: TSTAPI.kFAKE_CONTEXT_MENU_CLICK
+        },
+        { tabProperties: ['tab'], cache, isContextTab: true }
+      ).catch(ApiTabs.createErrorSuppressor())
     ]);
   }
 
@@ -436,20 +465,28 @@ async function onShown(contextTab) {
       viewType:         'sidebar',
       bookmarkId:       null
     },
-    tab: contextTab && new TSTAPI.TreeItem(contextTab, { isContextTab: true }) || null,
+    tab: contextTab,
     windowId: TabsStore.getCurrentWindowId()
   };
-  return Promise.all([
+  const cache = {};
+  const result = Promise.all([
     browser.runtime.sendMessage({
       ...message,
-      tab: message.tab && await message.tab.exportFor(browser.runtime.id)
+      tab: message.tab && await TSTAPI.exportTab(message.tab, browser.runtime.id, { cache })
     }).catch(ApiTabs.createErrorSuppressor()),
-    TSTAPI.sendMessage(message, { tabProperties: ['tab'] }),
-    TSTAPI.sendMessage({
-      ...message,
-      type: TSTAPI.kFAKE_CONTEXT_MENU_SHOWN
-    }, { tabProperties: ['tab'] })
+    TSTAPI.broadcastMessage(
+      message,
+      { tabProperties: ['tab'], cache, isContextTab: true }
+    ),
+    TSTAPI.broadcastMessage(
+      {
+        ...message,
+        type: TSTAPI.kFAKE_CONTEXT_MENU_SHOWN
+      },
+      { tabProperties: ['tab'], cache, isContextTab: true }
+    ),
   ]);
+  return result;
 }
 
 async function onHidden() {
@@ -459,14 +496,54 @@ async function onHidden() {
   };
   return Promise.all([
     browser.runtime.sendMessage(message).catch(ApiTabs.createErrorSuppressor()),
-    TSTAPI.sendMessage(message),
-    TSTAPI.sendMessage({
+    TSTAPI.broadcastMessage(message),
+    TSTAPI.broadcastMessage({
       ...message,
       type: TSTAPI.kFAKE_CONTEXT_MENU_HIDDEN
     })
   ]);
 }
 
+
+function updateContextualIdentitiesSelector() {
+  const disabled = document.documentElement.classList.contains('incognito') || ContextualIdentities.getCount() == 0;
+
+  const range    = document.createRange();
+  range.selectNodeContents(mNewTabButtonMenu);
+  range.deleteContents();
+
+  if (disabled)
+    return;
+
+  const fragment = ContextualIdentities.generateMenuItems({
+    hasDefault: true,
+  });
+  range.insertNode(fragment);
+  range.detach();
+}
+
+ContextualIdentities.onUpdated.addListener(() => {
+  updateContextualIdentitiesSelector();
+});
+
+async function onNewTabButtonMenuCommand(item, event) {
+  if (item.dataset.value) {
+    const action = EventUtils.isAccelAction(event) ?
+      configs.autoAttachOnNewTabButtonMiddleClick :
+      configs.autoAttachOnNewTabCommand;
+    BackgroundConnection.sendMessage({
+      type:          Constants.kCOMMAND_NEW_TAB_AS,
+      baseTabId:     Tab.getActiveTab(TabsStore.getCurrentWindowId()).id,
+      as:            action,
+      cookieStoreId: item.dataset.value,
+      inBackground:  event.shiftKey,
+    });
+  }
+
+  mNewTabButtonUI.close();
+}
+
+
 function onMessage(message, _sender) {
   log('tab-context-menu: internally called:', message);
   switch (message.type) {
@@ -502,7 +579,7 @@ function onMessageExternal(message, sender) {
       log('TSTAPI.kCONTEXT_MENU_OPEN:', message, { id: sender.id, url: sender.url });
       return (async () => {
         const tab      = message.tab ? Tab.get(message.tab) : null ;
-        const windowId = message.window || tab && tab.windowId;
+        const windowId = message.window || tab?.windowId;
         if (windowId != TabsStore.getCurrentWindowId())
           return;
         await onShown(tab);
@@ -562,50 +639,76 @@ window.addEventListener('mouseup', _event => {
 
 async function onContextMenu(event) {
   reserveToActivateSubpanel();
+  const focused = document.querySelector(':focus');
+  log('onContextMenu: start ', event, focused);
 
   const context = mReservedOverrideContext;
   mReservedOverrideContext = null;
 
-  const target         = EventUtils.getElementTarget(event);
-  const originalTarget = EventUtils.getElementOriginalTarget(event);
+  const openedByKeyboardOperation = (
+    event.button == 0 &&
+    !event.aleKey &&
+    !event.ctrlKey &&
+    !event.shiftKey &&
+    !event.metaKey
+  );
+  const keyboardOperationTarget = focused?.closest('[data-tab-id]') ||
+    Tab.getActiveTab(TabsStore.getCurrentWindowId()).$TST.element;
+  const target = openedByKeyboardOperation ?
+    keyboardOperationTarget :
+    EventUtils.getElementTarget(event);
+  const originalTarget = openedByKeyboardOperation ?
+    keyboardOperationTarget :
+    EventUtils.getElementOriginalTarget(event);
   const onInputField   = (
     target.closest('input, textarea') ||
     originalTarget.closest('input, textarea')
   );
+  log('onContextMenu: ', { target, originalTarget, onInputField, context });
 
-  if (!onInputField && context && context.context) {
+  if (!onInputField && context?.context) {
+    log('onContextMenu: override context aso something given: ', context);
     try {
       browser.menus.overrideContext(context);
     }
     catch(error) {
-      if (context.context == 'bookmark' &&
-          !(await Permissions.isGranted(Permissions.BOOKMARKS)))
-        notify({
-          title:   browser.i18n.getMessage('bookmarkContext_notification_notPermitted_title'),
-          message: browser.i18n.getMessage('bookmarkContext_notification_notPermitted_message'),
-          url:     `moz-extension://${location.host}/options/options.html#bookmarksPermissionGranted_context`
-        });
-      else
+      console.log('failed to override context: ', error);
+      try {
+        if (context.context == 'bookmark' &&
+            !(await Permissions.isGranted(Permissions.BOOKMARKS)))
+          notify({
+            title:   browser.i18n.getMessage('bookmarkContext_notification_notPermitted_title'),
+            message: browser.i18n.getMessage(`bookmarkContext_notification_notPermitted_message${isLinux() ? '_linux' : ''}`),
+            url:     `moz-extension://${location.host}/options/options.html#bookmarksPermissionGranted_context`
+          });
+        else
+          console.error(error);
+      }
+      catch(error) {
         console.error(error);
+      }
     }
     return;
   }
 
+  console.log('notify context menu is overridden');
   browser.runtime.sendMessage({
     type:    Constants.kCOMMAND_NOTIFY_CONTEXT_OVERRIDDEN,
     context: null
   });
 
-  if (onInputField)
+  if (onInputField) {
+    console.log('ignroe request on a input field');
     return;
+  }
 
-  const modifierKeyPressed = /^Mac/i.test(navigator.platform) ? event.metaKey : event.ctrlKey;
+  const modifierKeyPressed = isMacOS() ? event.metaKey : event.ctrlKey;
 
-  const originalTargetBookmarkElement = originalTarget && originalTarget.closest('[data-bookmark-id]');
-  const bookmarkId = originalTargetBookmarkElement && originalTargetBookmarkElement.dataset.bookmarkId;
+  const originalTargetBookmarkElement = originalTarget?.closest('[data-bookmark-id]');
+  const bookmarkId = originalTargetBookmarkElement?.dataset.bookmarkId;
   if (bookmarkId &&
-      !modifierKeyPressed &&
-      typeof browser.menus.overrideContext == 'function') {
+      !modifierKeyPressed) {
+    log('onContextMenu: override context as bookmark context menu');
     browser.menus.overrideContext({
       context:    'bookmark',
       bookmarkId: bookmarkId
@@ -613,13 +716,13 @@ async function onContextMenu(event) {
     return;
   }
 
-  const originalTargetTabElement = originalTarget && originalTarget.closest('[data-tab-id]');
-  const tab = originalTargetTabElement ?
-    TabsStore.ensureLivingTab(Tab.get(parseInt(originalTargetTabElement.dataset.tabId))) :
-    EventUtils.getTabFromEvent(event);
+  const originalTargetTreeItemElement = originalTarget?.closest('[data-tab-id]');
+  const tab = originalTargetTreeItemElement ?
+    TabsStore.ensureLivingItem(Tab.get(parseInt(originalTargetTreeItemElement.dataset.tabId))) :
+    EventUtils.getTreeItemFromEvent(event);
   if (tab &&
-      !modifierKeyPressed &&
-      typeof browser.menus.overrideContext == 'function') {
+      !modifierKeyPressed) {
+    log('onContextMenu: override context as tab context menu');
     browser.menus.overrideContext({
       context: 'tab',
       tabId: tab.id
@@ -627,9 +730,43 @@ async function onContextMenu(event) {
     return;
   }
 
-  if (!configs.emulateDefaultContextMenu)
+  if (EventUtils.isEventFiredOnNewTabButton(event)) {
+    log('onContextMenu: on new tab button');
+    event.stopPropagation();
+    event.preventDefault();
+    mNewTabButtonUI.open({
+      left: event.clientX,
+      top:  event.clientY,
+    });
     return;
+  }
 
+  const originalTargetNativeTabGroupElement = originalTarget?.closest(`[type="${TreeItem.TYPE_GROUP}"][data-native-tab-group-id]`);
+  const nativeTabGroup = originalTargetNativeTabGroupElement?.$TST.group;
+  if (nativeTabGroup &&
+      !modifierKeyPressed) {
+    log('onContextMenu: on native tab group');
+    event.stopPropagation();
+    event.preventDefault();
+    TabGroupContextMenu.show(nativeTabGroup);
+    return;
+  }
+
+  if (event.target == document.body) { // when the application key is pressed
+    log('onContextMenu: override context as tab context menu for blank area');
+    browser.menus.overrideContext({
+      context: 'tab',
+      tabId:   Tab.getActiveTab(TabsStore.getCurrentWindowId()).id,
+    });
+    return;
+  }
+
+  if (!configs.emulateDefaultContextMenu) {
+    log('onContextMenu: no emulation');
+    return;
+  }
+
+  log('onContextMenu: show emulated context menu');
   event.stopPropagation();
   event.preventDefault();
   await onShown(tab);
@@ -653,6 +790,19 @@ BackgroundConnection.onMessage.addListener(async message => {
     case Constants.kCOMMAND_NOTIFY_TAB_HIDDEN:
     case Constants.kCOMMAND_NOTIFY_CHILDREN_CHANGED:
       close();
+      mNewTabButtonUI.close();
       break;
+
+    case Constants.kCOMMAND_SHOW_NATIVE_TAB_GROUP_MENU_PANEL: {
+      close();
+      mNewTabButtonUI.close();
+      const group = TabGroup.get(message.groupId);
+      Promise.race([
+        group.$TST?.promisedElement,
+        wait(250),
+      ]).then(() => {
+        TabGroupContextMenu.show(group, true);
+      });
+    }; break;
   }
 });
