@@ -14,7 +14,7 @@
  * The Original Code is the Tree Style Tab.
  *
  * The Initial Developer of the Original Code is YUKI "Piro" Hiroshi.
- * Portions created by the Initial Developer are Copyright (C) 2011-2019
+ * Portions created by the Initial Developer are Copyright (C) 2011-2025
  * the Initial Developer. All Rights Reserved.
  *
  * Contributor(s): YUKI "Piro" Hiroshi <piro.outsider.reflex@gmail.com>
@@ -32,14 +32,13 @@ import {
   toLines,
   configs
 } from '/common/common.js';
-
-import * as Constants from '/common/constants.js';
 import * as ApiTabs from '/common/api-tabs.js';
-import * as TabsStore from '/common/tabs-store.js';
-import * as SidebarConnection from '/common/sidebar-connection.js';
+import * as Constants from '/common/constants.js';
 import { SequenceMatcher } from '/extlib/diff.js';
+import * as SidebarConnection from '/common/sidebar-connection.js';
+import * as TabsStore from '/common/tabs-store.js';
 
-import Tab from '/common/Tab.js';
+import { Tab, TreeItem } from '/common/TreeItem.js';
 
 function log(...args) {
   internalLogger('background/tabs-move', ...args);
@@ -55,7 +54,7 @@ function logApiTabs(...args) {
 export async function moveTabsBefore(tabs, referenceTab, options = {}) {
   log('moveTabsBefore: ', tabs, referenceTab, options);
   if (!tabs.length ||
-      !TabsStore.ensureLivingTab(referenceTab))
+      !TabsStore.ensureLivingItem(referenceTab))
     return [];
 
   if (referenceTab.$TST.isAllPlacedBeforeSelf(tabs)) {
@@ -70,12 +69,20 @@ export async function moveTabBefore(tab, referenceTab, options = {}) {
 
 async function moveTabsInternallyBefore(tabs, referenceTab, options = {}) {
   if (!tabs.length ||
-      !TabsStore.ensureLivingTab(referenceTab))
+      !TabsStore.ensureLivingItem(referenceTab))
     return [];
 
-  const window = TabsStore.windows.get(tabs[0].windowId);
+  const win = TabsStore.windows.get(tabs[0].windowId);
 
-  log('moveTabsInternallyBefore: ', tabs, referenceTab, options);
+  log('moveTabsInternallyBefore: ', tabs, `${referenceTab.id}(index=${referenceTab.index})`, options);
+  if (referenceTab.type == TreeItem.TYPE_GROUP) {
+    referenceTab = referenceTab.$TST?.firstMember;
+    if (!TabsStore.ensureLivingItem(referenceTab)) {
+      log('missing reference tab');
+      return [];
+    }
+    log(`  => reference tab: ${referenceTab.id}(index=${referenceTab.index})`);
+  }
 
   const precedingReferenceTab = referenceTab.$TST.previousTab;
   if (referenceTab.pinned) {
@@ -97,17 +104,22 @@ async function moveTabsInternallyBefore(tabs, referenceTab, options = {}) {
       the operation is asynchronous. To help synchronous operations
       following to this operation, we need to move tabs immediately.
     */
+    const tabGroups = new Set();
     for (const tab of tabs) {
       const oldPreviousTab = tab.$TST.unsafePreviousTab;
       const oldNextTab     = tab.$TST.unsafeNextTab;
-      if (oldNextTab && oldNextTab.id == referenceTab.id) // no move case
+      if (oldNextTab?.id == referenceTab.id) // no move case
         continue;
-      window.internalMovingTabs.add(tab.id);
-      window.alreadyMovedTabs.add(tab.id);
+      const fromIndex = tab.index;
       if (referenceTab.index > tab.index)
         tab.index = referenceTab.index - 1;
       else
         tab.index = referenceTab.index;
+      tabGroups.add(tab.$TST.nativeTabGroup);
+      if (SidebarConnection.isInitialized()) { // only on the background page
+        win.internalMovingTabs.set(tab.id, tab.index);
+        win.alreadyMovedTabs.set(tab.id, tab.index);
+      }
       tab.reindexedBy = `moveTabsInternallyBefore (${tab.index})`;
       Tab.track(tab);
       movedTabs.push(tab);
@@ -121,23 +133,40 @@ async function moveTabsInternallyBefore(tabs, referenceTab, options = {}) {
         type:        Constants.kCOMMAND_NOTIFY_TAB_INTERNALLY_MOVED,
         windowId:    tab.windowId,
         tabId:       tab.id,
-        newIndex:    tab.index,
-        nextTabId:   referenceTab && referenceTab.id,
+        fromIndex,
+        toIndex:     tab.index,
+        nextTabId:   referenceTab?.id,
         broadcasted: !!options.broadcasted
       });
+      if (options.doNotOptimize) {
+        win.internalMovingTabs.set(tab.id, tab.index);
+        win.alreadyMovedTabs.set(tab.id, tab.index);
+        await browser.tabs.move(tab.id, { index: tab.index });
+        win.internalMovingTabs.delete(tab.id);
+        win.alreadyMovedTabs.delete(tab.id);
+      }
+    }
+    for (const group of tabGroups) {
+      group?.$TST.reindex();
     }
     if (movedTabs.length == 0) {
       log(' => actually nothing moved');
     }
     else {
-      log('Tab nodes rearranged by moveTabsInternallyBefore:\n'+(!configs.debug ? '' :
-        () => toLines(Array.from(window.getOrderedTabs()),
-                      tab => ` - ${tab.index}: ${tab.id}${tabs.includes(tab) ? '[MOVED]' : ''}`)));
+      log(
+        'Tab nodes rearranged by moveTabsInternallyBefore:\n',
+        (!configs.debug ? '' :
+          () => toLines(Array.from(win.getOrderedTabs()),
+                        tab => ` - ${tab.index}: ${tab.id}${tabs.includes(tab) ? '[MOVED]' : ''}`))
+      );
     }
     if (SidebarConnection.isInitialized()) { // only on the background page
-      if (options.delayedMove) // Wait until opening animation is finished.
+      if (options.delayedMove) { // Wait until opening animation is finished.
         await wait(configs.newTabAnimationDuration);
-      syncToNativeTabs(tabs);
+      }
+      if (!options.doNotOptimize) {
+        syncToNativeTabs(tabs);
+      }
     }
   }
   catch(e) {
@@ -153,7 +182,7 @@ export async function moveTabInternallyBefore(tab, referenceTab, options = {}) {
 export async function moveTabsAfter(tabs, referenceTab, options = {}) {
   log('moveTabsAfter: ', tabs, referenceTab, options);
   if (!tabs.length ||
-      !TabsStore.ensureLivingTab(referenceTab))
+      !TabsStore.ensureLivingItem(referenceTab))
     return [];
 
   if (referenceTab.$TST.isAllPlacedAfterSelf(tabs)) {
@@ -168,12 +197,24 @@ export async function moveTabAfter(tab, referenceTab, options = {}) {
 
 async function moveTabsInternallyAfter(tabs, referenceTab, options = {}) {
   if (!tabs.length ||
-      !TabsStore.ensureLivingTab(referenceTab))
+      !TabsStore.ensureLivingItem(referenceTab))
     return [];
 
-  const window = TabsStore.windows.get(tabs[0].windowId);
+  const win = TabsStore.windows.get(tabs[0].windowId);
 
-  log('moveTabsInternallyAfter: ', tabs, referenceTab, options);
+  log('moveTabsInternallyAfter: ', tabs, `${referenceTab.id}(index=${referenceTab.index})`, options);
+  if (referenceTab.type == TreeItem.TYPE_GROUP) {
+    if (!referenceTab.collapsed) {
+      log('  => move before the first member tab of the reference group');
+      return moveTabsInternallyBefore(tabs, referenceTab.$TST?.firstMember, options = {});
+    }
+    referenceTab = referenceTab.$TST?.lastMember;
+    if (!TabsStore.ensureLivingItem(referenceTab)) {
+      log('missing reference tab');
+      return [];
+    }
+    log(`  => reference tab: ${referenceTab.id}(index=${referenceTab.index})`);
+  }
 
   const followingReferenceTab = referenceTab.$TST.nextTab;
   if (followingReferenceTab &&
@@ -196,16 +237,17 @@ async function moveTabsInternallyAfter(tabs, referenceTab, options = {}) {
       following to this operation, we need to move tabs immediately.
     */
     let nextTab = referenceTab.$TST.unsafeNextTab;
-    if (nextTab && tabs.find(tab => tab.id == nextTab.id))
-      nextTab = null;
+    while (nextTab && tabs.find(tab => tab.id == nextTab.id)) {
+      nextTab = nextTab.$TST.unsafeNextTab;
+    }
+    const tabGroups = new Set();
     for (const tab of tabs) {
       const oldPreviousTab = tab.$TST.unsafePreviousTab;
       const oldNextTab     = tab.$TST.unsafeNextTab;
       if ((!oldNextTab && !nextTab) ||
           (oldNextTab && nextTab && oldNextTab.id == nextTab.id)) // no move case
         continue;
-      window.internalMovingTabs.add(tab.id);
-      window.alreadyMovedTabs.add(tab.id);
+      const fromIndex = tab.index;
       if (nextTab) {
         if (nextTab.index > tab.index)
           tab.index = nextTab.index - 1;
@@ -213,7 +255,12 @@ async function moveTabsInternallyAfter(tabs, referenceTab, options = {}) {
           tab.index = nextTab.index;
       }
       else {
-        tab.index = window.tabs.size - 1
+        tab.index = win.tabs.size - 1
+      }
+      tabGroups.add(tab.$TST.nativeTabGroup);
+      if (SidebarConnection.isInitialized()) { // only on the background page
+        win.internalMovingTabs.set(tab.id, tab.index);
+        win.alreadyMovedTabs.set(tab.id, tab.index);
       }
       tab.reindexedBy = `moveTabsInternallyAfter (${tab.index})`;
       Tab.track(tab);
@@ -228,23 +275,40 @@ async function moveTabsInternallyAfter(tabs, referenceTab, options = {}) {
         type:        Constants.kCOMMAND_NOTIFY_TAB_INTERNALLY_MOVED,
         windowId:    tab.windowId,
         tabId:       tab.id,
-        newIndex:    tab.index,
-        nextTabId:   nextTab && nextTab.id,
+        fromIndex,
+        toIndex:     tab.index,
+        nextTabId:   nextTab?.id,
         broadcasted: !!options.broadcasted
       });
+      if (options.doNotOptimize) {
+        win.internalMovingTabs.set(tab.id, tab.index);
+        win.alreadyMovedTabs.set(tab.id, tab.index);
+        await browser.tabs.move(tab.id, { index: tab.index });
+        win.internalMovingTabs.delete(tab.id);
+        win.alreadyMovedTabs.delete(tab.id);
+      }
+    }
+    for (const group of tabGroups) {
+      group?.$TST.reindex();
     }
     if (movedTabs.length == 0) {
       log(' => actually nothing moved');
     }
     else {
-      log('Tab nodes rearranged by moveTabsInternallyAfter:\n'+(!configs.debug ? '' :
-        () => toLines(Array.from(window.getOrderedTabs()),
-                      tab => ` - ${tab.index}: ${tab.id}${tabs.includes(tab) ? '[MOVED]' : ''}`)));
+      log(
+        'Tab nodes rearranged by moveTabsInternallyAfter:\n',
+        (!configs.debug ? '' :
+          () => toLines(Array.from(win.getOrderedTabs()),
+                        tab => ` - ${tab.index}: ${tab.id}${tabs.includes(tab) ? '[MOVED]' : ''}`))
+      );
     }
     if (SidebarConnection.isInitialized()) { // only on the background page
-      if (options.delayedMove) // Wait until opening animation is finished.
+      if (options.delayedMove) { // Wait until opening animation is finished.
         await wait(configs.newTabAnimationDuration);
-      syncToNativeTabs(tabs);
+      }
+      if (!options.doNotOptimize) {
+        syncToNativeTabs(tabs);
+      }
     }
   }
   catch(e) {
@@ -261,20 +325,24 @@ export async function moveTabInternallyAfter(tab, referenceTab, options = {}) {
 // ========================================================
 // Synchronize order of tab elements to browser's tabs
 
-const mMovedTabs        = new Map();
 const mPreviousSync     = new Map();
 const mDelayedSync      = new Map();
 const mDelayedSyncTimer = new Map();
 
 export async function waitUntilSynchronized(windowId) {
-  return mPreviousSync.get(windowId) || mDelayedSync.get(windowId);
+  const previous = mPreviousSync.get(windowId);
+  if (previous)
+    return previous.then(() => waitUntilSynchronized(windowId));
+  return Promise.resolve(mDelayedSync.get(windowId)).then(() => {
+    const previous = mPreviousSync.get(windowId);
+    if (previous)
+      return waitUntilSynchronized(windowId);
+  });
 }
 
 function syncToNativeTabs(tabs) {
   const windowId = tabs[0].windowId;
   //log(`syncToNativeTabs(${windowId})`);
-  const movedTabs = mMovedTabs.get(windowId) || [];
-  mMovedTabs.set(windowId, movedTabs.concat(tabs));
   if (mDelayedSyncTimer.has(windowId))
     clearTimeout(mDelayedSyncTimer.get(windowId));
   const delayedSync = new Promise((resolve, _reject) => {
@@ -297,23 +365,15 @@ function syncToNativeTabs(tabs) {
 async function syncToNativeTabsInternal(windowId) {
   mDelayedSyncTimer.delete(windowId);
 
-  const oldMovedTabs = mMovedTabs.get(windowId) || [];
-  mMovedTabs.delete(windowId);
-
   if (Tab.needToWaitTracked(windowId))
     await Tab.waitUntilTrackedAll(windowId);
   if (Tab.needToWaitMoved(windowId))
     await Tab.waitUntilMovedAll(windowId);
 
-  const window = TabsStore.windows.get(windowId);
-  if (!window) // already destroyed
+  const win = TabsStore.windows.get(windowId);
+  if (!win) // already destroyed
     return;
 
-  for (const tab of oldMovedTabs) {
-    window.internalMovingTabs.delete(tab.id);
-    window.alreadyMovedTabs.delete(tab.id);
-  }
-
   // Tabs may be removed while waiting.
   const internalOrder   = TabsStore.windows.get(windowId).order;
   const nativeTabsOrder = (await browser.tabs.query({ windowId }).catch(ApiTabs.createErrorHandler())).map(tab => tab.id);
@@ -353,21 +413,35 @@ async function syncToNativeTabsInternal(windowId) {
           toIndex--;
         log(`syncToNativeTabs(${windowId}): step1, move ${moveTabIds.join(',')} before ${referenceId} / from = ${fromIndex}, to = ${toIndex}`);
         for (const movedId of moveTabIds) {
-          window.internalMovingTabs.add(movedId);
-          window.alreadyMovedTabs.add(movedId);
+          win.internalMovingTabs.set(movedId, -1);
+          win.alreadyMovedTabs.set(movedId, -1);
           movedTabs.add(movedId);
         }
         logApiTabs(`tabs-move:syncToNativeTabs(${windowId}): step1, browser.tabs.move() `, moveTabIds, {
           windowId,
           index: toIndex
         });
-        browser.tabs.move(moveTabIds, {
-          windowId,
-          index: toIndex
-        }).catch(ApiTabs.createErrorHandler(e => {
-          log(`syncToNativeTabs(${windowId}): step1, failed to move: `, String(e), e.stack);
-          throw e;
-        }));
+        let reallyMovedTabIds = new Set();
+        try {
+          const reallyMovedTabs = await browser.tabs.move(moveTabIds, {
+            windowId,
+            index: toIndex
+          }).catch(ApiTabs.createErrorHandler(e => {
+            log(`syncToNativeTabs(${windowId}): step1, failed to move: `, String(e), e.stack);
+            throw e;
+          }));
+          reallyMovedTabIds = new Set(reallyMovedTabs.map(tab => tab.id));
+        }
+        catch(error) {
+          console.error(error);
+        }
+        for (const id of moveTabIds) {
+          if (reallyMovedTabIds.has(id))
+            continue;
+          log(`syncToNativeTabs(${windowId}): failed to move tab ${id}: maybe unplacable position (regular tabs in pinned tabs/pinned tabs in regular tabs), or any other reason`);
+          win.internalMovingTabs.delete(id);
+          win.alreadyMovedTabs.delete(id);
+        }
         tabIdsForUpdatedIndices = tabIdsForUpdatedIndices.filter(id => !moveTabIds.includes(id));
         tabIdsForUpdatedIndices.splice(toIndex, 0, ...moveTabIds);
         break;
