@@ -14,7 +14,7 @@
  * The Original Code is the Tree Style Tab.
  *
  * The Initial Developer of the Original Code is YUKI "Piro" Hiroshi.
- * Portions created by the Initial Developer are Copyright (C) 2011-2019
+ * Portions created by the Initial Developer are Copyright (C) 2011-2024
  * the Initial Developer. All Rights Reserved.
  *
  * Contributor(s): YUKI "Piro" Hiroshi <piro.outsider.reflex@gmail.com>
@@ -26,20 +26,25 @@
  * ***** END LICENSE BLOCK ******/
 'use strict';
 
+import EventListenerManager from '/extlib/EventListenerManager.js';
+
 import {
   log as internalLogger,
-  configs
+  configs,
+  tryRevokeObjectURL,
 } from '/common/common.js';
-import * as Constants from '/common/constants.js';
 import * as ApiTabs from '/common/api-tabs.js';
-import * as TabsStore from '/common/tabs-store.js';
+import * as Constants from '/common/constants.js';
 import * as SidebarConnection from '/common/sidebar-connection.js';
+import * as TabsStore from '/common/tabs-store.js';
 
-import Tab from '/common/Tab.js';
+import { Tab } from '/common/TreeItem.js';
 
 import * as TabsMove from './tabs-move.js';
 import * as Tree from './tree.js';
 
+export const onForbiddenURLRequested = new EventListenerManager();
+
 function log(...args) {
   internalLogger('background/tabs-open', ...args);
 }
@@ -55,10 +60,14 @@ export async function loadURI(uri, options = {}) {
       tabId = options.tab.id;
     }
     else {
-      const tabs = await browser.tabs.query({
+      let tabs = await browser.tabs.query({
         windowId: options.windowId,
         active:   true
       }).catch(ApiTabs.createErrorHandler());
+      if (tabs.length == 0)
+        tabs = await browser.tabs.query({
+          windowId: options.windowId,
+        }).catch(ApiTabs.createErrorHandler());
       tabId = tabs[0].id;
     }
     let searchQuery = null;
@@ -78,7 +87,7 @@ export async function loadURI(uri, options = {}) {
     }
     else {
       await browser.tabs.update(tabId, {
-        url: uri
+        url: sanitizeURL(uri),
       }).catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
     }
   }
@@ -88,7 +97,9 @@ export async function loadURI(uri, options = {}) {
 }
 
 export async function openNewTab(options = {}) {
-  return openURIInTab(null, options);
+  const win = TabsStore.windows.get(options.windowId);
+  win.toBeOpenedNewTabCommandTab++;
+  return openURIInTab(options.url || options.uri || null, options);
 }
 
 export async function openURIInTab(uri, options = {}) {
@@ -99,35 +110,52 @@ export async function openURIInTab(uri, options = {}) {
 const FORBIDDEN_URL_MATCHER = /^(about|chrome|resource|file):/;
 const ALLOWED_URL_MATCHER = /^about:blank(\?|$)/;
 
-export async function openURIsInTabs(uris, options = {}) {
-  log('openURIsInTabs: ', { uris, options });
-  if (!options.windowId)
+export async function openURIsInTabs(uris, { windowId, insertBefore, insertAfter, cookieStoreId, isOrphan, active, inBackground, discarded, opener, parent, fixPositions } = {}) {
+  log('openURIsInTabs: ', { uris, windowId, insertBefore, insertAfter, cookieStoreId, isOrphan, active, inBackground, discarded, opener, parent, fixPositions });
+  if (!windowId)
     throw new Error('missing loading target window\n' + new Error().stack);
 
   const tabs = [];
   // Don't return the result of Tab.doAndGetNewTabs because their order can
   // be inverted due to browser.tabs.insertAfterCurrent=true
   const actuallyOpenedTabIds = new Set(await Tab.doAndGetNewTabs(async () => {
-    await Tab.waitUntilTrackedAll(options.windowId);
-    await TabsMove.waitUntilSynchronized(options.windowId);
-    const startIndex = Tab.calculateNewTabIndex(options);
+    await Tab.waitUntilTrackedAll(windowId);
+    await TabsMove.waitUntilSynchronized(windowId);
+    const startIndex = Tab.calculateNewTabIndex({ insertAfter, insertBefore });
     log('startIndex: ', startIndex);
-    const window = TabsStore.windows.get(options.windowId);
-    window.toBeOpenedTabsWithPositions += uris.length;
-    if (options.isOrphan)
-      window.toBeOpenedOrphanTabs += uris.length;
+    const win = TabsStore.windows.get(windowId);
+    if (insertBefore ||
+        insertAfter ||
+        uris.some(uri => uri && typeof uri == 'object' && 'index' in uri))
+      win.toBeOpenedTabsWithPositions += uris.length;
+    if (cookieStoreId)
+      win.toBeOpenedTabsWithCookieStoreId += uris.length;
+    if (isOrphan)
+      win.toBeOpenedOrphanTabs += uris.length;
     return Promise.all(uris.map(async (uri, index) => {
       const params = {
-        windowId: options.windowId,
-        active:   index == 0 && !options.inBackground
+        windowId,
+        active: index == 0 && (active || (inBackground === false)),
       };
-      if (uri && typeof uri == 'object') {
-        if ('title' in uri)
-          params.title = uri.title;
-        if ('discarded' in uri)
-          params.discarded = uri.discarded;
+      if (uri && typeof uri == 'object') { // tabs.create() compatible
+        if ('active' in uri)
+          params.active = uri.active;
         if ('cookieStoreId' in uri)
           params.cookieStoreId = uri.cookieStoreId;
+        if ('discarded' in uri)
+          params.discarded = uri.discarded;
+        if ('index' in uri)
+          params.index = uri.index;
+        if ('openerTabId' in uri)
+          params.openerTabId = uri.openerTabId;
+        if ('openInReaderMode' in uri)
+          params.openInReaderMode = uri.openInReaderMode;
+        if ('pinned' in uri)
+          params.pinned = uri.pinned;
+        if ('selected' in uri)
+          params.active = uri.selected;
+        if ('title' in uri)
+          params.title = uri.title;
         uri = uri.uri || uri.url;
       }
       let searchQuery = null;
@@ -144,24 +172,25 @@ export async function openURIsInTabs(uris, options = {}) {
           params.url = uri;
         }
       }
-      if (options.discarded &&
+      if (discarded &&
           !params.active &&
           !('discarded' in params))
         params.discarded = true;
-      if (params.url &&
-          FORBIDDEN_URL_MATCHER.test(params.url) &&
-          !ALLOWED_URL_MATCHER.test(params.url))
-        params.url = `about:blank?${params.url}`;
-      if (/^about:/.test(params.url))
+      if (params.url == 'about:newtab')
+        delete params.url
+      if (params.url)
+        params.url = sanitizeURL(params.url);
+      if (!('url' in params /* about:newtab case */) ||
+          /^about:/.test(params.url))
         params.discarded = false; // discarded tab cannot be opened with any about: URL
       if (!params.discarded) // title cannot be set for non-discarded tabs
         params.title = null;
-      if (options.opener)
-        params.openerTabId = options.opener.id;
-      if (startIndex > -1)
+      if (opener && !params.openerTabId)
+        params.openerTabId = opener.id;
+      if (startIndex > -1 && !('index' in params))
         params.index = startIndex + index;
-      if (options.cookieStoreId && !params.cookieStoreId)
-        params.cookieStoreId = options.cookieStoreId;
+      if (cookieStoreId && !params.cookieStoreId)
+        params.cookieStoreId = cookieStoreId;
         // Tabs opened with different container can take time to be tracked,
         // then TabsStore.waitUntilTabsAreCreated() may be resolved before it is
         // tracked like as "the tab is already closed". So we wait until the
@@ -187,32 +216,33 @@ export async function openURIsInTabs(uris, options = {}) {
       log('created tab: ', tab);
       if (!tab)
         throw new Error('tab is already closed');
-      if (!options.opener &&
-          options.parent &&
-          !options.isOrphan)
-        await Tree.attachTabTo(tab, options.parent, {
-          insertBefore: options.insertBefore,
-          insertAfter:  options.insertAfter,
+      if (!opener &&
+          parent &&
+          !isOrphan)
+        await Tree.attachTabTo(tab, parent, {
+          insertBefore: insertBefore,
+          insertAfter:  insertAfter,
           forceExpand:  params.active,
           broadcast:    true
         });
-      else if (options.insertBefore)
-        await TabsMove.moveTabInternallyBefore(tab, options.insertBefore, {
+      else if (insertBefore)
+        await TabsMove.moveTabInternallyBefore(tab, insertBefore, {
           broadcast: true
         });
-      else if (options.insertAfter)
-        await TabsMove.moveTabInternallyAfter(tab, options.insertAfter, {
+      else if (insertAfter)
+        await TabsMove.moveTabInternallyAfter(tab, insertAfter, {
           broadcast: true
         });
       log('tab is opened.');
       await tab.$TST.opened;
       tabs.push(tab);
+      tryRevokeObjectURL(tab.url);
       return tab;
     }));
-  }, options.windowId));
+  }, windowId));
   const openedTabs = tabs.filter(tab => actuallyOpenedTabIds.has(tab));
 
-  if (options.fixPositions &&
+  if (fixPositions &&
       openedTabs.every((tab, index) => (index == 0) || (openedTabs[index-1].index - tab.index) == 1)) {
     // tabs are opened with reversed order due to browser.tabs.insertAfterCurrent=true
     let lastTab;
@@ -221,11 +251,31 @@ export async function openURIsInTabs(uris, options = {}) {
         TabsMove.moveTabInternallyBefore(tab, lastTab);
       lastTab = tab;
     }
-    await TabsMove.waitUntilSynchronized(options.windowId);
+    await TabsMove.waitUntilSynchronized(windowId);
   }
   return openedTabs;
 }
 
+function sanitizeURL(url) {
+  if (ALLOWED_URL_MATCHER.test(url))
+    return url;
+
+  // tabs.create() doesn't accept about:reader URLs so we fallback them to regular URLs.
+  if (/^about:reader\?/.test(url))
+    return (new URL(url)).searchParams.get('url') || 'about:blank';
+
+  if (FORBIDDEN_URL_MATCHER.test(url)) {
+    onForbiddenURLRequested.dispatch(url);
+    return `about:blank?forbidden-url=${url}`;
+  }
+
+  return url;
+}
+
+export function isOpenable(url) {
+  return !url || url == sanitizeURL(url);
+}
+
 
 function onMessage(message, openerTab) {
   switch (message.type) {
@@ -249,7 +299,9 @@ function onMessage(message, openerTab) {
           windowId:     message.windowId,
           parent:       Tab.get(message.parentId),
           insertBefore: Tab.get(message.insertBeforeId),
-          insertAfter:  Tab.get(message.insertAfterId)
+          insertAfter:  Tab.get(message.insertAfterId),
+          active:       !!message.active,
+          discarded:    message.discarded,
         });
       });
       break;
@@ -267,7 +319,9 @@ function onMessage(message, openerTab) {
           opener:       Tab.get(message.openerId),
           parent:       Tab.get(message.parentId),
           insertBefore: Tab.get(message.insertBeforeId),
-          insertAfter:  Tab.get(message.insertAfterId)
+          insertAfter:  Tab.get(message.insertAfterId),
+          active:       message.active,
+          discarded:    message.discarded,
         });
       });
       break;
