@@ -5,11 +5,11 @@
 */
 'use strict';
 
-import * as Constants from './constants.js';
 import {
   log as internalLogger,
   configs
 } from './common.js';
+import * as Constants from './constants.js';
 
 // eslint-disable-next-line no-unused-vars
 function log(...args) {
@@ -34,6 +34,7 @@ export function getCurrentWindowId() {
 
 export const windows        = new Map();
 export const tabs           = new Map();
+export const tabGroups      = new Map();
 export const tabsByUniqueId = new Map();
 
 export function clear() {
@@ -71,17 +72,17 @@ export function queryAll(query) {
   if (configs.loggingQueries) {
     queryLogs.push(query);
     queryLogs.splice(0, Math.max(0, queryLogs.length - MAX_LOGS));
-    if (query.tabs && query.tabs.name)
+    if (query.tabs?.name)
       query.indexedTabs = query.tabs.name;
   }
   fixupQuery(query);
   const startAt = Date.now();
   if (query.windowId || query.ordered) {
     let tabs = [];
-    for (const window of windows.values()) {
-      if (query.windowId && !matched(window.id, query.windowId))
+    for (const win of windows.values()) {
+      if (query.windowId && !matched(win.id, query.windowId))
         continue;
-      const [sourceTabs, offset] = sourceTabsForQuery(query, window);
+      const [sourceTabs, offset] = sourceTabsForQuery(query, win);
       tabs = tabs.concat(query.iterator ? getMatchedTabsIterator(sourceTabs, query, offset) : extractMatchedTabs(sourceTabs, query, offset));
     }
     query.elapsed = Date.now() - startAt;
@@ -106,34 +107,40 @@ export function queryAll(query) {
   }
 }
 
-function sourceTabsForQuery(query, window) {
+function sourceTabsForQuery(query, win) {
   let offset = 0;
   if (!query.ordered)
-    return [query.tabs && query.tabs.values() || window.tabs.values(), offset];
+    return [query.tabs?.values() || win.tabs.values(), offset];
   let fromId;
+  let toId = query.toId;
   if (typeof query.index == 'number') {
-    fromId = window.order[query.index];
+    fromId = win.order[query.index];
     offset = query.index;
   }
   if (typeof query.fromIndex == 'number') {
-    fromId = window.order[query.fromIndex];
+    fromId = win.order[query.fromIndex];
     offset = query.fromIndex;
   }
+  if (typeof query.toIndex == 'number') {
+    toId = win.order[query.toIndex];
+  }
   if (typeof fromId != 'number') {
     fromId = query.fromId;
-    offset = window.order.indexOf(query.fromId);
+    offset = win.order.indexOf(query.fromId);
   }
   if (query.last || query.reversed)
-    return [window.getReversedOrderedTabs(fromId, query.toId, query.tabs), offset];
-  return [window.getOrderedTabs(fromId, query.toId, query.tabs), offset];
+    return [win.getReversedOrderedTabs(fromId, toId, query.tabs), offset];
+  return [win.getOrderedTabs(fromId, toId, query.tabs), offset];
 }
 
-function extractMatchedTabs(tabs, query, offset) {
+function extractMatchedTabs(tabsInStore, query, offset) {
   const matchedTabs = [];
   let firstTime     = true;
   let logicalIndex  = offset || 0;
-  for (const tab of tabs) {
-    if (!matchedWithQuery(tab, query))
+  for (const tab of tabsInStore) {
+    if (!tabs.has(tab.id) ||
+        (!query.skipMatching &&
+         !matchedWithQuery(tab, query)))
       continue;
 
     if (!firstTime)
@@ -150,11 +157,13 @@ function extractMatchedTabs(tabs, query, offset) {
   return matchedTabs;
 }
 
-function* getMatchedTabsIterator(tabs, query, offset) {
+function* getMatchedTabsIterator(tabsInStore, query, offset) {
   let firstTime     = true;
   let logicalIndex  = offset || 0;
-  for (const tab of tabs) {
-    if (!matchedWithQuery(tab, query))
+  for (const tab of tabsInStore) {
+    if (!tabs.has(tab.id) ||
+        (!query.skipMatching &&
+         !matchedWithQuery(tab, query)))
       continue;
 
     if (!firstTime)
@@ -206,19 +215,27 @@ function matchedWithQuery(tab, query) {
   }
 
   if (query.living &&
-      !ensureLivingTab(tab))
+      !ensureLivingItem(tab))
     return false;
   if (query.normal &&
       (tab.hidden ||
        tabStates.has(Constants.kTAB_STATE_SHOWING) ||
        tab.pinned))
     return false;
+  if (query.pinned &&
+      (tab.hidden ||
+       tabStates.has(Constants.kTAB_STATE_SHOWING) ||
+       !tab.pinned))
+    return false;
   if (query.visible &&
       ((tabStates.has(Constants.kTAB_STATE_COLLAPSED) &&
         !tabStates.has(Constants.kTAB_STATE_EXPANDING)) ||
        tab.hidden ||
        tabStates.has(Constants.kTAB_STATE_SHOWING)))
     return false;
+  if (query.hidden &&
+      !tab.hidden)
+    return false;
   if (query.controllable &&
       (tab.hidden ||
        tabStates.has(Constants.kTAB_STATE_SHOWING)))
@@ -235,6 +252,9 @@ function matchedWithQuery(tab, query) {
   if ('descendantOf' in query &&
       !tab.$TST.ancestorIds.includes(query.descendantOf))
     return false;
+  if (query.groupId &&
+      tab.groupId != query.groupId)
+    return false;
 
   return true;
 }
@@ -268,7 +288,7 @@ export function query(query) {
   if (configs.loggingQueries) {
     queryLogs.push(query);
     queryLogs.splice(0, Math.max(0, queryLogs.length - MAX_LOGS));
-    if (query.tabs && query.tabs.name)
+    if (query.tabs?.name)
       query.indexedTabs = query.tabs.name;
   }
   fixupQuery(query);
@@ -279,10 +299,10 @@ export function query(query) {
   const startAt = Date.now();
   let tabs = [];
   if (query.windowId || query.ordered) {
-    for (const window of windows.values()) {
-      if (query.windowId && !matched(window.id, query.windowId))
+    for (const win of windows.values()) {
+      if (query.windowId && !matched(win.id, query.windowId))
         continue;
-      const [sourceTabs, offset] = sourceTabsForQuery(query, window);
+      const [sourceTabs, offset] = sourceTabsForQuery(query, win);
       tabs = tabs.concat(extractMatchedTabs(sourceTabs, query, offset));
       if (tabs.length > 0)
         break;
@@ -330,15 +350,17 @@ export const pinnedTabsInWindow      = new Map();
 export const unpinnedTabsInWindow    = new Map();
 export const rootTabsInWindow        = new Map();
 export const groupTabsInWindow       = new Map();
-export const collapsingTabsInWindow  = new Map();
-export const expandingTabsInWindow   = new Map();
 export const toBeExpandedTabsInWindow = new Map();
 export const subtreeCollapsableTabsInWindow = new Map();
 export const draggingTabsInWindow    = new Map();
 export const duplicatingTabsInWindow = new Map();
 export const toBeGroupedTabsInWindow = new Map();
+export const nativelyGroupedTabsInWindow = new Map();
 export const loadingTabsInWindow     = new Map();
 export const unsynchronizedTabsInWindow = new Map();
+export const virtualScrollRenderableTabsInWindow  = new Map();
+export const scrollPositionCalculationTargetTabsInWindow = new Map();
+export const canBecomeStickyTabsInWindow = new Map();
 
 function createMapWithName(name) {
   const map = new Map();
@@ -361,15 +383,17 @@ export function prepareIndexesForWindow(windowId) {
   unpinnedTabsInWindow.set(windowId, createMapWithName(`unpinned tabs in window ${windowId}`));
   rootTabsInWindow.set(windowId, createMapWithName(`root tabs in window ${windowId}`));
   groupTabsInWindow.set(windowId, createMapWithName(`group tabs in window ${windowId}`));
-  collapsingTabsInWindow.set(windowId, createMapWithName(`collapsing tabs in window ${windowId}`));
-  expandingTabsInWindow.set(windowId, createMapWithName(`expanding tabs in window ${windowId}`));
   toBeExpandedTabsInWindow.set(windowId, createMapWithName(`to-be-expanded tabs in window ${windowId}`));
   subtreeCollapsableTabsInWindow.set(windowId, createMapWithName(`collapsable parent tabs in window ${windowId}`));
   draggingTabsInWindow.set(windowId, createMapWithName(`dragging tabs in window ${windowId}`));
   duplicatingTabsInWindow.set(windowId, createMapWithName(`duplicating tabs in window ${windowId}`));
   toBeGroupedTabsInWindow.set(windowId, createMapWithName(`to-be-grouped tabs in window ${windowId}`));
+  nativelyGroupedTabsInWindow.set(windowId, createMapWithName(`natively grouped tabs in window ${windowId}`));
   loadingTabsInWindow.set(windowId, createMapWithName(`loading tabs in window ${windowId}`));
   unsynchronizedTabsInWindow.set(windowId, createMapWithName(`unsynchronized tabs in window ${windowId}`));
+  virtualScrollRenderableTabsInWindow.set(windowId, createMapWithName(`virtual scroll renderable tabs in window ${windowId}`));
+  scrollPositionCalculationTargetTabsInWindow.set(windowId, createMapWithName(`scroll position calculation target tabs in window ${windowId}`));
+  canBecomeStickyTabsInWindow.set(windowId, createMapWithName(`can become sticky tabs in window ${windowId}`));
 }
 
 export function unprepareIndexesForWindow(windowId) {
@@ -388,13 +412,21 @@ export function unprepareIndexesForWindow(windowId) {
   unpinnedTabsInWindow.delete(windowId);
   rootTabsInWindow.delete(windowId);
   groupTabsInWindow.delete(windowId);
-  collapsingTabsInWindow.delete(windowId);
-  expandingTabsInWindow.delete(windowId);
   toBeExpandedTabsInWindow.delete(windowId);
   subtreeCollapsableTabsInWindow.delete(windowId);
   toBeGroupedTabsInWindow.delete(windowId);
+  nativelyGroupedTabsInWindow.delete(windowId);
   loadingTabsInWindow.delete(windowId);
   unsynchronizedTabsInWindow.delete(windowId);
+  virtualScrollRenderableTabsInWindow.delete(windowId);
+  scrollPositionCalculationTargetTabsInWindow.delete(windowId);
+  canBecomeStickyTabsInWindow.delete(windowId);
+}
+
+export function getTabsMap(tabsStore, windowId = null) {
+  return windowId ?
+    tabsStore.get(windowId) :
+    new Map([...tabsStore.values()].map(tabs => [...tabs.entries()]).flat());
 }
 
 export function updateIndexesForTab(tab) {
@@ -471,6 +503,37 @@ export function updateIndexesForTab(tab) {
     addBundledActiveTab(tab);
   else
     removeBundledActiveTab(tab);
+
+  if (tab.groupId && tab.groupId != -1)
+    addNativelyGroupedTab(tab);
+  else
+    removeNativelyGroupedTab(tab);
+
+  updateVirtualScrollRenderabilityIndexForTab(tab);
+
+  if (tab.$TST.states.has(Constants.kTAB_STATE_COLLAPSED_DONE))
+    removeScrollPositionCalculationTargetTab(tab);
+  else
+    addScrollPositionCalculationTargetTab(tab);
+
+  if (tab.$TST.canBecomeSticky)
+    addCanBecomeStickyTab(tab);
+  else
+    removeCanBecomeStickyTab(tab);
+}
+
+export function updateVirtualScrollRenderabilityIndexForTab(tab) {
+  if (tab.pinned ||
+      (tab.hidden &&
+       ((tab.url == 'about:firefoxview' &&
+         tab.cookieStoreId == 'firefox-default') ||
+        !configs.renderHiddenTabs)) ||
+      tab.$TST.states.has(Constants.kTAB_STATE_COLLAPSED_DONE) ||
+      (tabGroups.get(tab.groupId)?.$TST.states.has(Constants.kTAB_STATE_COLLAPSED_DONE) &&
+       !tab.active))
+    removeVirtualScrollRenderableTab(tab);
+  else
+    addVirtualScrollRenderableTab(tab);
 }
 
 export function removeTabFromIndexes(tab) {
@@ -487,29 +550,31 @@ export function removeTabFromIndexes(tab) {
   removeUnpinnedTab(tab);
   removeRootTab(tab);
   removeGroupTab(tab);
-  removeCollapsingTab(tab);
-  removeExpandingTab(tab);
   removeToBeExpandedTab(tab);
   removeSubtreeCollapsableTab(tab);
   removeDuplicatingTab(tab);
   removeDraggingTab(tab);
   removeToBeGroupedTab(tab);
+  removeNativelyGroupedTab(tab);
   removeLoadingTab(tab);
   removeUnsynchronizedTab(tab);
+  //removeVirtualScrollRenderableTab(tab);
+  removeScrollPositionCalculationTargetTab(tab);
+  removeCanBecomeStickyTab(tab);
 }
 
-function addTabToIndex(tab, indexes) {
+function addTabToIndex(tab, indexes, windowId = null) {
   if (!tab)
     throw new Error(`TabsStore.addTabToIndex gets non-tab parameter!: ${JSON.stringify(tab)} : ${new Error().stack}`);
-  const tabs = indexes.get(tab.windowId);
+  const tabs = indexes.get(windowId || tab.windowId);
   if (tabs)
     tabs.set(tab.id, tab);
 }
 
-function removeTabFromIndex(tab, indexes) {
+function removeTabFromIndex(tab, indexes, windowId = null) {
   if (!tab)
     throw new Error(`TabsStore.removeTabFromIndex gets non-tab parameter!: ${JSON.stringify(tab)} : ${new Error().stack}`);
-  const tabs = indexes.get(tab.windowId);
+  const tabs = indexes.get(windowId || tab.windowId);
   if (tabs)
     tabs.delete(tab.id);
 }
@@ -603,20 +668,6 @@ export function removeGroupTab(tab) {
   removeTabFromIndex(tab, groupTabsInWindow);
 }
 
-export function addCollapsingTab(tab) {
-  addTabToIndex(tab, collapsingTabsInWindow);
-}
-export function removeCollapsingTab(tab) {
-  removeTabFromIndex(tab, collapsingTabsInWindow);
-}
-
-export function addExpandingTab(tab) {
-  addTabToIndex(tab, expandingTabsInWindow);
-}
-export function removeExpandingTab(tab) {
-  removeTabFromIndex(tab, expandingTabsInWindow);
-}
-
 export function addToBeExpandedTab(tab) {
   addTabToIndex(tab, toBeExpandedTabsInWindow);
 }
@@ -652,6 +703,13 @@ export function removeToBeGroupedTab(tab) {
   removeTabFromIndex(tab, toBeGroupedTabsInWindow);
 }
 
+export function addNativelyGroupedTab(tab, windowId = null) {
+  addTabToIndex(tab, nativelyGroupedTabsInWindow, windowId);
+}
+export function removeNativelyGroupedTab(tab, windowId = null) {
+  removeTabFromIndex(tab, nativelyGroupedTabsInWindow, windowId);
+}
+
 export function addLoadingTab(tab) {
   addTabToIndex(tab, loadingTabsInWindow);
 }
@@ -673,6 +731,27 @@ export function removeBundledActiveTab(tab) {
   removeTabFromIndex(tab, bundledActiveTabsInWindow);
 }
 
+export function addVirtualScrollRenderableTab(tab) {
+  addTabToIndex(tab, virtualScrollRenderableTabsInWindow);
+}
+export function removeVirtualScrollRenderableTab(tab) {
+  removeTabFromIndex(tab, virtualScrollRenderableTabsInWindow);
+}
+
+export function addScrollPositionCalculationTargetTab(tab) {
+  addTabToIndex(tab, scrollPositionCalculationTargetTabsInWindow);
+}
+export function removeScrollPositionCalculationTargetTab(tab) {
+  removeTabFromIndex(tab, scrollPositionCalculationTargetTabsInWindow);
+}
+
+export function addCanBecomeStickyTab(tab) {
+  addTabToIndex(tab, canBecomeStickyTabsInWindow);
+}
+export function removeCanBecomeStickyTab(tab) {
+  removeTabFromIndex(tab, canBecomeStickyTabsInWindow);
+}
+
 
 
 //===================================================================
@@ -680,21 +759,22 @@ export function removeBundledActiveTab(tab) {
 //===================================================================
 
 export function assertValidTab(tab) {
-  if (tab && tab.$TST)
+  if (tab?.$TST)
     return;
   const error = new Error('FATAL ERROR: invalid tab is given');
   console.log(error.message, tab, error.stack);
   throw error;
 }
 
-export function ensureLivingTab(tab) {
+export function ensureLivingItem(tab) {
+  const isNativeTabGroup = tab?.type == 'group';
   if (!tab ||
       !tab.id ||
       !tab.$TST ||
-      (tab.$TST.element &&
-       !tab.$TST.element.parentNode) ||
-      !tabs.has(tab.id) ||
-      tab.$TST.states.has(Constants.kTAB_STATE_REMOVING))
+      (!isNativeTabGroup && !tabs.has(tab.id)) ||
+      tab.$TST.removing ||
+      !windows.get(tab.windowId) ||
+      (isNativeTabGroup && !windows.get(tab.windowId).tabGroups.has(tab.id)))
     return null;
   return tab;
 }
