@@ -12,6 +12,7 @@ import {
 
 import * as ApiTabsListener from '/background/api-tabs-listener.js';
 import { Diff } from '/common/diff.js';
+import * as Utils from './utils.js';
 
 import * as TestDOMUpdater from './test-dom-updater.js';
 import * as TestContextMenu from './test-context-menu.js';
@@ -19,16 +20,44 @@ import * as TestHidden from './test-hidden.js';
 import * as TestNewTab from './test-new-tab.js';
 import * as TestSuccessor from './test-successor.js';
 import * as TestTree from './test-tree.js';
+import * as TestCloseParentBehavior from './test-close-parent-behavior.js';
+import * as TestSidebar from './test-sidebar.js';
 
+
+const NO_RESET_CONFIG_KEYS = new Set([
+  'debug',
+  'logFor',
+  'style', // prevent reloading of the background page
+  'useCachedTree' // prevent reloading of the background page
+]);
+
+const mTestMatchers = [];
 let mResults;
 let mLogs;
+let mStopped = false;
 
 async function run() {
   await configs.$loaded;
   ApiTabsListener.init();
   ApiTabsListener.start();
+
+  // you can run specific tests with URL parameter like:
+  // ...runner.html?testMoveAttachedTabBeforeHiddenTab,/^testMove.+/,...
+  for (const part of (location.search || '').replace(/^\?/, '').split(/,/)) {
+    const trimmedPart = part.trim();
+    if (!trimmedPart.trim())
+      continue;
+    if (/^\/([^\/]+)\/(i)?$/.test(trimmedPart))
+      mTestMatchers.push(new RegExp(RegExp.$1, RegExp.$2 || ''));
+    else
+      mTestMatchers.push(new RegExp(`^${trimmedPart}$`, 'i'));
+  }
+
   mResults = document.getElementById('results');
   mLogs = document.getElementById('logs');
+  mStopped = false;
+  document.getElementById('stop').addEventListener('click', () => mStopped = true);
+  document.getElementById('stop').addEventListener('keypress', () => mStopped = true);
   const configValues = backupConfigs();
   await runAll();
   ApiTabsListener.destroy();
@@ -44,9 +73,13 @@ function backupConfigs() {
 }
 
 async function restoreConfigs(values) {
+  const resetConfigs = {};
   for (const key of Object.keys(values)) {
-    configs[key] = values[key];
+    if (NO_RESET_CONFIG_KEYS.has(key))
+      continue;
+    resetConfigs[key] = values[key];
   }
+  await Utils.setConfigs(resetConfigs);
   // wait until updated configs are delivered to other namespaces
   await nextFrame();
 }
@@ -62,7 +95,9 @@ async function runAll() {
       TestHidden,
       TestNewTab,
       TestSuccessor,
-      TestTree
+      TestTree,
+      TestCloseParentBehavior,
+      TestSidebar
     );
   }
   let runOnlyRunnable = false;
@@ -78,16 +113,24 @@ async function runAll() {
     }
   }
   for (const tests of testCases) {
+    if (mStopped)
+      break;
     const setup    = tests.setUp || tests.setup;
     const teardown = tests.tearDown || tests.teardown;
     for (const name of Object.keys(tests)) {
+      if (mStopped)
+        break;
       if (!name.startsWith('test'))
         continue;
       if (runOnlyRunnable && !tests[name].runnable)
         continue;
+      if (mTestMatchers.length > 0 &&
+          mTestMatchers.every(matcher => !matcher.test(name)))
+        continue;
       await restoreConfigs(configs.$default);
       let shouldTearDown = true;
       const result = mResults.appendChild(document.createElement('span'));
+      console.log('running test: ', name);
       try {
         if (typeof setup == 'function')
           await setup();
@@ -165,25 +208,21 @@ function logFailure(name, error) {
     expectedBlock.appendChild(document.createElement('legend')).textContent = 'Expected';
     const expected = expectedBlock.appendChild(document.createElement('pre'));
     expected.classList.add('expected');
-    expected.textContent = error.expected.trim();
+    expected.textContent = (error.expected || '').trim();
   }
 
   const actualBlock = item.appendChild(document.createElement('fieldset'));
   actualBlock.appendChild(document.createElement('legend')).textContent = 'Actual';
   const actual = actualBlock.appendChild(document.createElement('pre'));
   actual.classList.add('actual');
-  actual.textContent = error.actual.trim();
+  actual.textContent = (error.actual || '').trim();
 
   if ('expected' in error) {
     const diffBlock = item.appendChild(document.createElement('fieldset'));
     diffBlock.appendChild(document.createElement('legend')).textContent = 'Difference';
     const diff = diffBlock.appendChild(document.createElement('pre'));
     diff.classList.add('diff');
-    const range = document.createRange();
-    range.selectNodeContents(diff);
-    range.collapse(false);
-    diff.appendChild(range.createContextualFragment(Diff.readable(error.expected, error.actual, true)));
-    range.detach();
+    diff.insertAdjacentHTML('beforeend', Diff.readable(error.expected, error.actual, true));
   }
 }
 
