@@ -10,7 +10,7 @@ import {
 } from '/common/common.js';
 import { is /*, ok, ng*/ } from '/tests/assert.js';
 import * as TSTAPI from '/common/tst-api.js';
-//import Tab from '/common/Tab.js';
+//import { Tab } from '/common/TreeItem.js';
 
 import * as Constants from '/common/constants.js';
 import * as Utils from './utils.js';
@@ -30,12 +30,10 @@ export async function teardown() {
 export async function testAutoFixupForHiddenTabs() {
   await Utils.setConfigs({
     fixupTreeOnTabVisibilityChanged: true,
-    inheritContextualIdentityToNewChildTab: false,
-    inheritContextualIdentityToSameSiteOrphan: false
+    inheritContextualIdentityToChildTabMode: Constants.kCONTEXTUAL_IDENTITY_DEFAULT,
+    inheritContextualIdentityToSameSiteOrphanMode: Constants.kCONTEXTUAL_IDENTITY_DEFAULT
   });
 
-  let tabs;
-
   /*
   - A
     - B (with the "Personal" container)
@@ -46,36 +44,30 @@ export async function testAutoFixupForHiddenTabs() {
       - G (with the "Personal" container)
     - H
   */
-  tabs = await Utils.createTabs({
-    A: { index: 1, cookieStoreId: 'firefox-default' },
-    B: { index: 2, cookieStoreId: 'firefox-container-1', openerTabId: 'A' },
-    C: { index: 3, cookieStoreId: 'firefox-container-1', openerTabId: 'B' },
-    D: { index: 4, cookieStoreId: 'firefox-default', openerTabId: 'C' },
-    E: { index: 5, cookieStoreId: 'firefox-default', openerTabId: 'D' },
-    F: { index: 6, cookieStoreId: 'firefox-container-1', openerTabId: 'A' },
-    G: { index: 7, cookieStoreId: 'firefox-container-1', openerTabId: 'F' },
-    H: { index: 8, cookieStoreId: 'firefox-default', openerTabId: 'A' }
-  }, { windowId: win.id });
+  let tabs = await Utils.prepareTabsInWindow(
+    { A: { index: 1, cookieStoreId: 'firefox-default' },
+      B: { index: 2, cookieStoreId: 'firefox-container-1', openerTabId: 'A' },
+      C: { index: 3, cookieStoreId: 'firefox-container-1', openerTabId: 'B' },
+      D: { index: 4, cookieStoreId: 'firefox-default', openerTabId: 'C' },
+      E: { index: 5, cookieStoreId: 'firefox-default', openerTabId: 'D' },
+      F: { index: 6, cookieStoreId: 'firefox-container-1', openerTabId: 'A' },
+      G: { index: 7, cookieStoreId: 'firefox-container-1', openerTabId: 'F' },
+      H: { index: 8, cookieStoreId: 'firefox-default', openerTabId: 'A' } },
+    win.id,
+    [ 'A',
+      'A => B',
+      'A => B => C',
+      'A => B => C => D',
+      'A => B => C => D => E',
+      'A => F',
+      'A => F => G',
+      'A => H' ]
+  );
   await wait(1000);
 
   tabs = await Utils.refreshTabs(tabs);
   {
-    const { A, B, C, D, E, F, G, H } = tabs;
-    is([
-      `${A.id}`,
-      `${A.id} => ${B.id}`,
-      `${A.id} => ${B.id} => ${C.id}`,
-      `${A.id} => ${B.id} => ${C.id} => ${D.id}`,
-      `${A.id} => ${B.id} => ${C.id} => ${D.id} => ${E.id}`,
-      `${A.id} => ${F.id}`,
-      `${A.id} => ${F.id} => ${G.id}`,
-      `${A.id} => ${H.id}`
-    ], Utils.treeStructure(Object.values(tabs)),
-       'tabs must be initialized with specified structure');
-    is([A.id, B.id, C.id, D.id, E.id, F.id, G.id, H.id],
-       await Utils.tabsOrder([A, B, C, D, E, F, G, H]),
-       'tabs must be initialized with specified order');
-
+    const { /* A, */ B, C, /* D, E, */ F, G /*, H */ } = tabs;
     await new Promise(resolve => {
       // wait until tabs are updated by TST
       let count = 0;
@@ -139,6 +131,10 @@ export async function testAutoFixupForHiddenTabs() {
 }
 
 export async function testCalculateNewTabPositionWithHiddenTabs() {
+  await Utils.setConfigs({
+    insertNewChildAt:            Constants.kINSERT_END,
+    insertNewTabFromPinnedTabAt: Constants.kINSERT_END
+  });
   let tabs = await Utils.createTabs({
     A: { index: 1 },
     B: { index: 2, active: true },
@@ -159,13 +155,14 @@ export async function testCalculateNewTabPositionWithHiddenTabs() {
       `${C.id}`,
     ], Utils.treeStructure([B, D, C]),
        'new tab must be a child of the opener');
-    is([B.id, D.id, C.id],
-       await Utils.tabsOrder([B, C, D]),
-       'new tab must be placed next to the opener');
   }
 }
 
 export async function testNewTabBeforeHiddenTab() {
+  await Utils.setConfigs({
+    insertNewChildAt:            Constants.kINSERT_END,
+    insertNewTabFromPinnedTabAt: Constants.kINSERT_END
+  });
   let tabs = await Utils.createTabs({
     A: { index: 1, active: true },
     B: { index: 2 }, // => hidden
@@ -209,8 +206,7 @@ export async function testNewTabBeforeHiddenTab() {
     is(expectedIndex, newTabs[0].index, 'a new tab must be placed before hidden tab');
 
     await browser.tabs.update(baseTab.id, { active: true });
-    await browser.tabs.remove(newTabs[0].id);
-    await wait(1000);
+    await Utils.waitUntilAllTabChangesFinished(() => browser.tabs.remove(newTabs[0].id));
   }
   await assertNewTabOpenedAs(tabs.A.index + 1, tabs.A, Constants.kNEWTAB_OPEN_AS_CHILD);
   await assertNewTabOpenedAs(tabs.A.index + 1, tabs.A, Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING);
