@@ -10,7 +10,7 @@ import {
   configs
 } from '/common/common.js';
 import { is, isNot /*, ok, ng*/ } from '/tests/assert.js';
-//import Tab from '/common/Tab.js';
+//import { Tab } from '/common/TreeItem.js';
 
 import * as Constants from '/common/constants.js';
 import * as Utils from './utils.js';
@@ -20,7 +20,7 @@ let win;
 export async function setup() {
   win = await browser.windows.create();
   // wait until timeout the special timer which prevents grouping of new tabs
-  await wait(configs.autoGroupNewTabsDelayOnNewWindow);
+  await wait(configs.tabBunchesDetectionDelayOnNewWindow);
 }
 
 export async function teardown() {
@@ -31,7 +31,7 @@ export async function teardown() {
 
 export async function testInheritContainerFromAutoAttachedParent() {
   await Utils.setConfigs({
-    inheritContextualIdentityToNewChildTab: true,
+    inheritContextualIdentityToChildTabMode: Constants.kCONTEXTUAL_IDENTITY_FROM_PARENT,
     autoAttachOnNewTabCommand: Constants.kNEWTAB_OPEN_AS_CHILD
   });
 
@@ -62,7 +62,7 @@ export async function testInheritContainerFromAutoAttachedParent() {
 
 export async function testDoNotInheritContainerFromExplicitParent() {
   await Utils.setConfigs({
-    inheritContextualIdentityToNewChildTab: true
+    inheritContextualIdentityToChildTabMode: Constants.kCONTEXTUAL_IDENTITY_FROM_PARENT
   });
 
   const parent = await browser.tabs.create({
@@ -113,6 +113,11 @@ export async function testTryToOpenUnpinnedTabBeforePinnedTab() {
 }
 
 export async function testReopenedWithPositionByAnotherAddonImmediatelyWhileCreating() {
+  await Utils.setConfigs({
+    insertNewChildAt:            Constants.kINSERT_END,
+    insertNewTabFromPinnedTabAt: Constants.kINSERT_END
+  });
+  await wait(1000); // this is required to avoid affections from other tests
   let tabs = await Utils.createTabs({
     A: { index: 1, active: true },
     B: { index: 2 }
@@ -159,9 +164,15 @@ export async function testReopenedWithPositionByAnotherAddonImmediatelyWhileCrea
           url:   `${location.origin}/resources/ui-color.css`,
           index: 2
         });
-        await browser.tabs.executeScript(reopenedTab.id, {
-          code: 'location.href'
-        });
+        if (browser.scripting)
+          await browser.scripting.executeScript({ // Manifest V3
+            target: { tabId: reopenedTab.id },
+            func: function() { return location.href; },
+          });
+        else
+          await browser.tabs.executeScript(reopenedTab.id, {
+            code: 'location.href',
+          });
         browser.tabs.remove(tab.id);
         resolve(reopenedTab);
       }
