@@ -9,10 +9,29 @@ import {
   wait,
   configs
 } from '/common/common.js';
+import { is /*, ok, ng*/ } from './assert.js';
 
 import * as Constants from '/common/constants.js';
 //import * as Tree from '/background/tree.js';
-import Tab from '/common/Tab.js';
+import { Tab, TreeItem } from '/common/TreeItem.js';
+
+export async function prepareTabsInWindow(definition, windowId, expectedStructure) {
+  let tabs = await createTabs(definition, { windowId });
+  tabs = await refreshTabs(tabs);
+  const actualTabs = new Set();
+  const expected = expectedStructure.map(line => {
+    return line.replace(/\b[a-z]+\b/gi, matched => {
+      const tab = tabs[matched];
+      if (!actualTabs.has(tab))
+        actualTabs.add(tab);
+      return tab.id;
+    });
+  });
+  is(expected,
+     treeStructure(Array.from(actualTabs)),
+     'tabs must be initialized with specified structure');
+  return tabs;
+}
 
 export async function createTab(params = {}) {
   return browser.tabs.create(params);
@@ -101,6 +120,7 @@ export async function refreshTabs(tabs) {
     for (const tab of refreshedTabsArray) {
       refreshedTabs[idToName[tab.id]] = Tab.import(tab);
     }
+    console.log('refreshedTabs: ', Object.entries(refreshedTabs).map(([name, tab]) => `${name}(${tab.id})`));
     return refreshedTabs;
   }
 
@@ -121,7 +141,7 @@ export function treeStructure(tabs) {
       return `${outputNestedRelation(tabsById[tab.$TST.parentId])} => ${tab.id}`;
     return `${tab.id}`;
   };
-  return tabs.map(outputNestedRelation);
+  return tabs.slice(0).sort((a, b) => a.index - b.index).map(outputNestedRelation);
 }
 
 export async function tabsOrder(tabs) {
@@ -130,7 +150,7 @@ export async function tabsOrder(tabs) {
       type:   Constants.kCOMMAND_PULL_TABS,
       tabIds: tabs.map(tab => tab.id || tab)
     });
-    return Tab.sort(tabs).map(tab => tab.id);
+    return TreeItem.sort(tabs).map(tab => tab.id);
   }
 
   if (typeof tabs == 'object') {
@@ -138,7 +158,7 @@ export async function tabsOrder(tabs) {
       type:   Constants.kCOMMAND_PULL_TABS,
       tabIds: Object.values(tabs).map(tab => tab.id)
     });
-    return Tab.sort(refreshedTabsArray).map(tab => tab.id);
+    return TreeItem.sort(refreshedTabsArray).map(tab => tab.id);
   }
 
   throw new Error('Invalid tab collection: ', tabs);
@@ -184,3 +204,38 @@ export async function callAPI(message) {
     type: `treestyletab:api:${message.type}`
   });
 }
+
+export async function waitUntilAllTabChangesFinished(operation) {
+  return new Promise(async (resolve, reject) => {
+    let changeCount = 0;
+    let operationFinished = false;
+    const onChanged = () => {
+      changeCount++;
+      setTimeout(() => {
+        changeCount--;
+        if (changeCount > 0)
+          return;
+        browser.tabs.onCreated.removeListener(onChanged);
+        browser.tabs.onRemoved.removeListener(onChanged);
+        browser.tabs.onMoved.removeListener(onChanged);
+        if (operationFinished)
+          resolve();
+      }, 500);
+    };
+    browser.tabs.onCreated.addListener(onChanged);
+    browser.tabs.onRemoved.addListener(onChanged);
+    browser.tabs.onMoved.addListener(onChanged);
+    if (typeof operation == 'function') {
+      try {
+        await operation();
+      }
+      catch(error) {
+        operationFinished = true;
+        return reject(error);
+      }
+    }
+    operationFinished = true;
+    if (changeCount == 0)
+      resolve();
+  });
+}
