@@ -5,31 +5,30 @@
 */
 'use strict';
 
+import EventListenerManager from '/extlib/EventListenerManager.js';
+
 import {
   log as internalLogger,
   dumpTab,
   toLines,
   configs,
-  wait
+  wait,
 } from '/common/common.js';
-
-import * as Constants from '/common/constants.js';
 import * as ApiTabs from '/common/api-tabs.js';
-import * as TabsStore from '/common/tabs-store.js';
+import * as Constants from '/common/constants.js';
 import * as SidebarConnection from '/common/sidebar-connection.js';
-import * as MetricsData from '/common/metrics-data.js';
-import * as UserOperationBlocker from '/common/user-operation-blocker.js';
-import * as TreeBehavior from '/common/tree-behavior.js';
 import * as TabsInternalOperation from '/common/tabs-internal-operation.js';
+import * as TabsStore from '/common/tabs-store.js';
+import * as TreeBehavior from '/common/tree-behavior.js';
+import * as UserOperationBlocker from '/common/user-operation-blocker.js';
 
-import Tab from '/common/Tab.js';
+import MetricsData from '/common/MetricsData.js';
+import { Tab } from '/common/TreeItem.js';
 
-import * as Tree from './tree.js';
-import * as TabsOpen from './tabs-open.js';
-import * as TabsMove from './tabs-move.js';
 import * as Commands from './commands.js';
-
-import EventListenerManager from '/extlib/EventListenerManager.js';
+import * as TabsMove from './tabs-move.js';
+import * as TabsOpen from './tabs-open.js';
+import * as Tree from './tree.js';
 
 function log(...args) {
   internalLogger('background/tree-structure', ...args);
@@ -57,19 +56,19 @@ export function startTracking() {
 }
 
 export function reserveToSaveTreeStructure(windowId) {
-  const window = TabsStore.windows.get(windowId);
-  if (!window)
+  const win = TabsStore.windows.get(windowId);
+  if (!win)
     return;
 
-  if (window.waitingToSaveTreeStructure)
-    clearTimeout(window.waitingToSaveTreeStructure);
-  window.waitingToSaveTreeStructure = setTimeout(() => {
+  if (win.waitingToSaveTreeStructure)
+    clearTimeout(win.waitingToSaveTreeStructure);
+  win.waitingToSaveTreeStructure = setTimeout(() => {
     saveTreeStructure(windowId);
   }, 150);
 }
 async function saveTreeStructure(windowId) {
-  const window = TabsStore.windows.get(windowId);
-  if (!window)
+  const win = TabsStore.windows.get(windowId);
+  if (!win)
     return;
 
   const structure = TreeBehavior.getTreeStructureFromTabs(Tab.getAllTabs(windowId));
@@ -83,17 +82,17 @@ async function saveTreeStructure(windowId) {
 export async function loadTreeStructure(windows, restoredFromCacheResults) {
   log('loadTreeStructure');
   MetricsData.add('loadTreeStructure: start');
-  return MetricsData.addAsync('loadTreeStructure: restoration for windows', Promise.all(windows.map(async window => {
+  return MetricsData.addAsync('loadTreeStructure: restoration for windows', Promise.all(windows.map(async win => {
     if (restoredFromCacheResults &&
-        restoredFromCacheResults.get(window.id)) {
-      log(`skip tree structure restoration for window ${window.id} (restored from cache)`);
+        restoredFromCacheResults.get(win.id)) {
+      log(`skip tree structure restoration for window ${win.id} (restored from cache)`);
       return;
     }
-    const tabs = Tab.getAllTabs(window.id);
+    const tabs = Tab.getAllTabs(win.id);
     let windowStateCompletelyApplied = false;
     try {
-      const structure = await browser.sessions.getWindowValue(window.id, Constants.kWINDOW_STATE_TREE_STRUCTURE).catch(ApiTabs.createErrorHandler());
-      let uniqueIds = tabs.map(tab => tab.$TST.uniqueId && tab.$TST.uniqueId || '?');
+      const structure = await browser.sessions.getWindowValue(win.id, Constants.kWINDOW_STATE_TREE_STRUCTURE).catch(ApiTabs.createErrorHandler());
+      let uniqueIds = tabs.map(tab => tab.$TST.uniqueId || '?');
       MetricsData.add('loadTreeStructure: read stored data');
       if (structure &&
           structure.length > 0 &&
@@ -110,7 +109,11 @@ export async function loadTreeStructure(windows, restoredFromCacheResults) {
           windowStateCompletelyApplied = structure.length == tabs.length;
         }
         if (tabsOffset > -1) {
-          await Tree.applyTreeStructureToTabs(tabs.slice(tabsOffset), structure);
+          const structureRestoreTabs = tabs.slice(tabsOffset);
+          await Tree.applyTreeStructureToTabs(structureRestoreTabs, structure);
+          for (const tab of structureRestoreTabs) {
+            tab.$TST.temporaryMetadata.set('treeStructureAlreadyRestoredFromSessionData', true);
+          }
           MetricsData.add('loadTreeStructure: Tree.applyTreeStructureToTabs');
         }
         else {
@@ -126,16 +129,42 @@ export async function loadTreeStructure(windows, restoredFromCacheResults) {
       MetricsData.add('loadTreeStructure: failed to apply tree structure');
     }
     if (!windowStateCompletelyApplied) {
-      log(`Tree information for the window ${window.id} is not same to actual state. Fallback to restoration from tab relations.`);
+      log(`Tree information for the window ${win.id} is not same to actual state. Fallback to restoration from tab relations.`);
       MetricsData.add('loadTreeStructure: fallback to reserveToAttachTabFromRestoredInfo');
+      const unattachedTabs = new Set(tabs);
       for (const tab of tabs) {
         reserveToAttachTabFromRestoredInfo(tab, {
           keepCurrentTree: true,
           canCollapse:     true
+        }).then(attached => {
+          if (attached ||
+              tab.$TST.parent ||
+              tab.$TST.hasChild)
+            unattachedTabs.delete(tab);
         });
       }
       await reserveToAttachTabFromRestoredInfo.promisedDone;
       MetricsData.add('loadTreeStructure: attachTabFromRestoredInfo finish');
+
+      // unknown tabs may appear inside tree, so we need to fixup tree based on their position.
+      for (const tab of unattachedTabs) {
+        const action = Tree.detectTabActionFromNewPosition(tab, {
+          fromIndex: tabs.length - 1,
+          toIndex:   tab.index,
+          isTabCreating: true,
+        });
+        switch (action.action) {
+          default:
+            break;
+
+          case 'attach':
+          case 'detach':
+            log('loadTreeStructure: apply action for unattached tab: ', tab, action);
+            await action.applyIfNeeded();
+            break;
+        }
+      }
+      MetricsData.add('loadTreeStructure: finish to fixup tree structure');
     }
     Tab.dumpAll();
   })));
@@ -156,7 +185,7 @@ async function reserveToAttachTabFromRestoredInfo(tab, options = {}) {
     reserveToAttachTabFromRestoredInfo.tasks = [];
     const uniqueIds = tasks.map(task => task.tab.$TST.uniqueId);
     const bulk = tasks.length > 1;
-    await Promise.all(uniqueIds.map((uniqueId, index) => {
+    const attachedResults = await Promise.all(uniqueIds.map((uniqueId, index) => {
       const task = tasks[index];
       return attachTabFromRestoredInfo(task.tab, {
         ...task.options,
@@ -164,10 +193,11 @@ async function reserveToAttachTabFromRestoredInfo(tab, options = {}) {
         bulk
       }).catch(error => {
         console.log(`TreeStructure.reserveToAttachTabFromRestoredInfo: Fatal error on processing task ${index}, ${error}`, error.stack);
+        return false;
       });
     }));
     if (typeof reserveToAttachTabFromRestoredInfo.onDone == 'function')
-      reserveToAttachTabFromRestoredInfo.onDone();
+      reserveToAttachTabFromRestoredInfo.onDone(attachedResults.every(attached => !!attached));
     delete reserveToAttachTabFromRestoredInfo.onDone;
     delete reserveToAttachTabFromRestoredInfo.promisedDone;
     Tab.dumpAll();
@@ -181,12 +211,19 @@ reserveToAttachTabFromRestoredInfo.promisedDone = null;
 
 async function attachTabFromRestoredInfo(tab, options = {}) {
   log('attachTabFromRestoredInfo ', tab, options);
-  let uniqueId, insertBefore, insertAfter, ancestors, children, states, collapsed /* for backward compatibility */;
+  if (tab.$TST.temporaryMetadata.has('treeStructureAlreadyRestoredFromSessionData')) {
+    log(' => already restored ', tab.id);
+    return;
+  }
+
+  let uniqueId, insertBefore, insertAfter, insertAfterLegacy, ancestors, children, states, collapsed /* for backward compatibility */;
   // eslint-disable-next-line prefer-const
-  [uniqueId, insertBefore, insertAfter, ancestors, children, states, collapsed] = await Promise.all([
+  [uniqueId, insertBefore, insertAfter, insertAfterLegacy, ancestors, children, states, collapsed] = await Promise.all([
     options.uniqueId || tab.$TST.uniqueId || tab.$TST.promisedUniqueId,
     browser.sessions.getTabValue(tab.id, Constants.kPERSISTENT_INSERT_BEFORE).catch(ApiTabs.createErrorHandler()),
     browser.sessions.getTabValue(tab.id, Constants.kPERSISTENT_INSERT_AFTER).catch(ApiTabs.createErrorHandler()),
+    // This legacy should be removed after legacy data are cleared enough, maybe after Firefox 128 is released.
+    browser.sessions.getTabValue(tab.id, Constants.kPERSISTENT_INSERT_AFTER_LEGACY).catch(ApiTabs.createErrorHandler()),
     browser.sessions.getTabValue(tab.id, Constants.kPERSISTENT_ANCESTORS).catch(ApiTabs.createErrorHandler()),
     browser.sessions.getTabValue(tab.id, Constants.kPERSISTENT_CHILDREN).catch(ApiTabs.createErrorHandler()),
     tab.$TST.getPermanentStates(),
@@ -196,6 +233,7 @@ async function attachTabFromRestoredInfo(tab, options = {}) {
   children  = children  || [];
   log(`persistent references for ${dumpTab(tab)} (${uniqueId.id}): `, {
     insertBefore, insertAfter,
+    insertAfterLegacy,
     ancestors: ancestors.join(', '),
     children:  children.join(', '),
     states,
@@ -207,10 +245,10 @@ async function attachTabFromRestoredInfo(tab, options = {}) {
     browser.sessions.removeTabValue(tab.id, Constants.kPERSISTENT_SUBTREE_COLLAPSED).catch(ApiTabs.createErrorSuppressor());
   }
   insertBefore = Tab.getByUniqueId(insertBefore);
-  insertAfter  = Tab.getByUniqueId(insertAfter);
+  insertAfter  = Tab.getByUniqueId(insertAfter || insertAfterLegacy);
   ancestors    = ancestors.map(Tab.getByUniqueId);
   children     = children.map(Tab.getByUniqueId);
-  log(' => references: ', () => ({
+  log(' => references: ', tab.id, () => ({
     insertBefore: dumpTab(insertBefore),
     insertAfter:  dumpTab(insertAfter),
     ancestors:    ancestors.map(dumpTab).join(', '),
@@ -219,7 +257,7 @@ async function attachTabFromRestoredInfo(tab, options = {}) {
   if (configs.fixupTreeOnTabVisibilityChanged) {
     ancestors = ancestors.filter(ancestor => ancestor && (ancestor.hidden == tab.hidden));
     children = children.filter(child => child && (child.hidden == tab.hidden));
-    log(' ==> references: ', () => ({
+    log(' ==> references: ', tab.id, () => ({
       ancestors: ancestors.map(dumpTab).join(', '),
       children:  children.map(dumpTab).join(', ')
     }));
@@ -227,9 +265,9 @@ async function attachTabFromRestoredInfo(tab, options = {}) {
 
   // clear wrong positioning information
   if (tab.pinned ||
-      (insertBefore && insertBefore.pinned))
+      insertBefore?.pinned)
     insertBefore = null;
-  const nextOfInsertAfter = insertAfter && insertAfter.$TST.nextTab;
+  const nextOfInsertAfter = insertAfter?.$TST.nextTab;
   if (nextOfInsertAfter &&
       nextOfInsertAfter.pinned)
     insertAfter = null;
@@ -240,7 +278,7 @@ async function attachTabFromRestoredInfo(tab, options = {}) {
   for (const ancestor of ancestors) {
     if (!ancestor)
       continue;
-    log(' attach to old ancestor: ', { child: tab, parent: ancestor });
+    log(' attach to old ancestor: ', tab.id, { child: tab, parent: ancestor });
     const promisedDone = Tree.attachTabTo(tab, ancestor, {
       insertBefore,
       insertAfter,
@@ -259,7 +297,7 @@ async function attachTabFromRestoredInfo(tab, options = {}) {
     const opener = tab.$TST.openerTab;
     if (opener &&
         configs.syncParentTabAndOpenerTab) {
-      log(' attach to opener: ', { child: tab, parent: opener });
+      log(' attach to opener: ', tab.id, { child: tab, parent: opener });
       const promisedDone = Tree.attachTabTo(tab, opener, {
         dontExpand:  !active,
         forceExpand: active,
@@ -274,7 +312,7 @@ async function attachTabFromRestoredInfo(tab, options = {}) {
     else if (!options.bulk &&
              (tab.$TST.nearestCompletelyOpenedNormalFollowingTab ||
               tab.$TST.nearestCompletelyOpenedNormalPrecedingTab)) {
-      log(' attach from position');
+      log(' attach from position: ', tab.id);
       onTabAttachedFromRestoredInfo.dispatch(tab, {
         toIndex:   tab.index,
         fromIndex: Tab.getLastTab(tab.windowId).index
@@ -313,6 +351,7 @@ async function attachTabFromRestoredInfo(tab, options = {}) {
   }
 
   const subtreeCollapsed = states.includes(Constants.kTAB_STATE_SUBTREE_COLLAPSED);
+  log('restore subtree collapsed state: ', tab.id, { current: tab.$TST.subtreeCollapsed, expected: subtreeCollapsed, ...options });
   if ((options.canCollapse || options.bulk) &&
       tab.$TST.subtreeCollapsed != subtreeCollapsed) {
     const promisedDone = Tree.collapseExpandSubtree(tab, {
@@ -323,12 +362,32 @@ async function attachTabFromRestoredInfo(tab, options = {}) {
     promises.push(promisedDone);
   }
 
+  const updateCollapsedState = () => {
+    const shouldBeCollapsed = tab.$TST.ancestors.some(ancestor => ancestor.$TST.collapsed || ancestor.$TST.subtreeCollapsed);
+    log('update collapsed state: ', tab.id, { current: tab.$TST.collapsed, expected: shouldBeCollapsed });
+    if ((options.canCollapse || options.bulk) &&
+        tab.$TST.collapsed != shouldBeCollapsed) {
+      Tree.collapseExpandTabAndSubtree(tab, {
+        broadcast: true,
+        collapsed: !tab.$TST.collapsed,
+        justNow:   true
+      });
+    }
+  };
+
+  tab.$TST.temporaryMetadata.set('treeStructureAlreadyRestoredFromSessionData', true);
+
   if (options.bulk)
-    return Promise.all(promises);
+    await Promise.all(promises).then(updateCollapsedState);
+  else
+    updateCollapsedState();
+
+  return attached;
 }
 
 const mRestoringTabs = new Map();
 const mMaxRestoringTabs = new Map();
+const mRestoredTabIds = new Set();
 const mProcessingTabRestorations = [];
 
 Tab.onRestored.addListener(tab => {
@@ -349,17 +408,18 @@ Tab.onRestored.addListener(tab => {
       const maxCount = mMaxRestoringTabs.get(tab.windowId) || 0;
       mMaxRestoringTabs.set(tab.windowId, Math.max(count, maxCount));
 
+      const uniqueId = await tab.$TST.promisedUniqueId;
+      mRestoredTabIds.add(uniqueId.id);
+
       if (count == 0) {
         // Force restore recycled active tab.
         // See also: https://github.com/piroor/treestyletab/issues/2191#issuecomment-489271889
         const activeTab = Tab.getActiveTab(tab.windowId);
-        let uniqueId, restoredUniqueId;
-        // eslint-disable-next-line prefer-const
-        [uniqueId, restoredUniqueId] = await Promise.all([
+        const [uniqueId, restoredUniqueId] = await Promise.all([
           activeTab.$TST.promisedUniqueId,
           browser.sessions.getTabValue(activeTab.id, Constants.kPERSISTENT_ID).catch(ApiTabs.createErrorHandler())
         ]);
-        if (restoredUniqueId && restoredUniqueId.id != uniqueId.id) {
+        if (restoredUniqueId?.id != uniqueId.id) {
           activeTab.$TST.updateUniqueId({ id: restoredUniqueId.id });
           reserveToAttachTabFromRestoredInfo(activeTab, {
             children: true
@@ -387,11 +447,16 @@ Tab.onRestored.addListener(tab => {
         if (count == 0) {
           mRestoringTabs.delete(tab.windowId);
           mMaxRestoringTabs.delete(tab.windowId);
-          setTimeout(() => { // unblock in the next event loop, after other asynchronous operations are finished
+          setTimeout(() => { // because window.requestAnimationFrame is decelerate for an invisible document.
+            // unblock in the next event loop, after other asynchronous operations are finished
             UserOperationBlocker.unblockIn(tab.windowId, { throbber: true });
           }, 0);
 
-          tryRestoreClosedSetFor(tab);
+          const countToBeRestored = mRecentlyClosedTabs.filter(tab => !mRestoredTabIds.has(tab.uniqueId));
+          log('countToBeRestored: ', countToBeRestored);
+          if (countToBeRestored > 0)
+            tryRestoreClosedSetFor(tab, countToBeRestored);
+          mRestoredTabIds.clear();
         }
         else {
           mRestoringTabs.set(tab.windowId, count);
@@ -430,7 +495,9 @@ Tab.onRemoved.addListener((_tab, _info) => {
     Tab.onChangeMultipleTabsRestorability.dispatch(newlyRestorable);
 });
 
-Tab.onMultipleTabsRemoving.addListener(tabs => {
+Tab.onMultipleTabsRemoving.addListener((tabs, { triggerTab, originalStructure } = {}) => {
+  if (triggerTab)
+    tabs = [triggerTab, ...tabs];
   mPendingRecentlyClosedTabsInfo.tabs = tabs.map(tab => ({
     originalId:    tab.id,
     uniqueId:      tab.$TST.uniqueId.id,
@@ -439,16 +506,20 @@ Tab.onMultipleTabsRemoving.addListener(tabs => {
     url:           tab.url,
     cookieStoreId: tab.cookieStoreId
   }));
-  mPendingRecentlyClosedTabsInfo.structure = TreeBehavior.getTreeStructureFromTabs(tabs, {
+  mPendingRecentlyClosedTabsInfo.structure = originalStructure || TreeBehavior.getTreeStructureFromTabs(tabs, {
     full:                 true,
     keepParentOfRootTabs: true
   });
+  log('mPendingRecentlyClosedTabsInfo.tabs = ', mPendingRecentlyClosedTabsInfo.tabs);
+  log('mPendingRecentlyClosedTabsInfo.structure = ', mPendingRecentlyClosedTabsInfo.structure);
 });
 
-Tab.onMultipleTabsRemoved.addListener(tabs => {
+Tab.onMultipleTabsRemoved.addListener((tabs, { triggerTab } = {}) => {
   log('multiple tabs are removed');
   const currentlyRestorable = mRecentlyClosedTabs.length > 1;
 
+  if (triggerTab)
+    tabs = [triggerTab, ...tabs];
   const tabIds = new Set(tabs.map(tab => tab.id));
   mRecentlyClosedTabs = mPendingRecentlyClosedTabsInfo.tabs.filter(info => tabIds.has(info.originalId));
   mRecentlyClosedTabsTreeStructure = mPendingRecentlyClosedTabsInfo.structure.filter(structure => tabIds.has(structure.originalId));
@@ -462,7 +533,19 @@ Tab.onMultipleTabsRemoved.addListener(tabs => {
   mPendingRecentlyClosedTabsInfo.structure = [];
 });
 
-async function tryRestoreClosedSetFor(tab) {
+let mToBeActivatedRestoredTabId = null;
+
+function onRestoredTabActivated(activeInfo) {
+  if (mToBeActivatedRestoredTabId &&
+      activeInfo.id != mToBeActivatedRestoredTabId) {
+    TabsInternalOperation.activateTab(mToBeActivatedRestoredTabId);
+  }
+  mToBeActivatedRestoredTabId = null;
+}
+
+browser.tabs.onActivated.addListener(onRestoredTabActivated);
+
+async function tryRestoreClosedSetFor(tab, countToBeRestored) {
   const lastRecentlyClosedTabs = mRecentlyClosedTabs;
   const lastRecentlyClosedTabsTreeStructure = mRecentlyClosedTabsTreeStructure;
   mRecentlyClosedTabs = [];
@@ -480,7 +563,10 @@ async function tryRestoreClosedSetFor(tab) {
     return;
   }
 
-  const toBeRestoredTabsCount = lastRecentlyClosedTabs.length - 1;
+  const toBeRestoredTabsCount = Math.min(
+    typeof countToBeRestored == 'number' ? countToBeRestored : Number.MAX_SAFE_INTEGER,
+    lastRecentlyClosedTabs.filter(tabInfo => tabInfo.uniqueId != tab.$TST.uniqueId.id).length
+  );
   if (toBeRestoredTabsCount == 0) {
     log(' => no more tab to be restored.');
     return;
@@ -490,12 +576,45 @@ async function tryRestoreClosedSetFor(tab) {
     maxResults: browser.sessions.MAX_SESSION_RESULTS
   }).catch(ApiTabs.createErrorHandler())).filter(session => session.tab);
 
+  const canRestoreWithSession = toBeRestoredTabsCount <= sessions.length;
+
   let restoredTabs;
-  if (toBeRestoredTabsCount <= sessions.length) {
-    log('tryRestoreClosedSetFor: restore tabs with the sessions API');
-    restoredTabs = await Commands.restoreTabs(toBeRestoredTabsCount);
-    restoredTabs.push(tab);
-    Tab.sort(restoredTabs)
+  if (canRestoreWithSession) {
+    log(`tryRestoreClosedSetFor: restore ${toBeRestoredTabsCount} tabs with the sessions API`);
+    const unsortedRestoredTabs = await Commands.restoreTabs(toBeRestoredTabsCount);
+    unsortedRestoredTabs.push(tab);
+    // tabs can be restored in different order, then we need to rearrange them manually
+    const tabsByUniqueId = new Map();
+    for (const tab of unsortedRestoredTabs) {
+      tabsByUniqueId.set(tab.$TST.uniqueId.id, tab);
+    }
+    restoredTabs = await Promise.all(lastRecentlyClosedTabsTreeStructure.map(tabInfo => {
+      const restoredTab = tabsByUniqueId.get(tabInfo.id);
+      if (restoredTab)
+        return restoredTab;
+      log('tryRestoreClosedSetFor: recreate tab for ', tabInfo);
+      return TabsOpen.openURIInTab({
+        title:         tabInfo.title,
+        url:           tabInfo.url,
+        cookieStoreId: tabInfo.cookieStoreId
+      }, {
+        windowId:      tab.windowId,
+        isOrphan:      true,
+        inBackground:  true,
+        discarded:     true,
+        fixPositions:  true
+      });
+    }));
+    let lastTab;
+    for (const tab of restoredTabs) {
+      if (lastTab && tab.$TST.previousTab != lastTab)
+        await TabsMove.moveTabAfter(
+          tab,
+          lastTab,
+          { broadcast: true }
+        );
+      lastTab = tab;
+    }
   }
   else {
     log('tryRestoreClosedSetFor: recreate tabs');
@@ -543,8 +662,9 @@ async function tryRestoreClosedSetFor(tab) {
   }
 
   const rootTabs = restoredTabs.filter((tab, index) => lastRecentlyClosedTabsTreeStructure[index].parent == TreeBehavior.STRUCTURE_KEEP_PARENT || lastRecentlyClosedTabsTreeStructure[index].parent == TreeBehavior.STRUCTURE_NO_PARENT);
+  log(`tryRestoreClosedSetFor: rootTabs, restoredTabs = `, rootTabs, restoredTabs);
   for (const rootTab of rootTabs) {
-    const referenceTabs = TreeBehavior.calculateReferenceTabsFromInsertionPosition(rootTab, {
+    const referenceTabs = TreeBehavior.calculateReferenceItemsFromInsertionPosition(rootTab, {
       context:      Constants.kINSERTION_CONTEXT_MOVED,
       insertAfter:  rootTab.$TST.previousTab,
       insertBefore: restoredTabs[restoredTabs.length - 1].$TST.nextTab
@@ -566,11 +686,6 @@ async function tryRestoreClosedSetFor(tab) {
 
   // Firefox itself activates the initially restored tab with delay,
   // so we need to activate the first tab of the restored tabs again.
-  const onActivated = activeInfo => {
-    browser.tabs.onActivated.removeListener(onActivated);
-    if (activeInfo.id != restoredTabs[0].id)
-      TabsInternalOperation.activateTab(restoredTabs[0]);
-  };
-  browser.tabs.onActivated.addListener(onActivated);
-  wait(100).then(() => onActivated({ id: -1 })); // failsafe
+  mToBeActivatedRestoredTabId = restoredTabs[0].id;
+  wait(100).then(() => onRestoredTabActivated({ id: -1 })); // failsafe
 }
