@@ -6,11 +6,13 @@
 'use strict';
 
 import {
-  log as internalLogger
+  log as internalLogger,
+  configs,
+  wait,
 } from './common.js';
+import * as ApiTabs from './api-tabs.js';
 import * as Constants from './constants.js';
 import * as TabsStore from './tabs-store.js';
-import * as ApiTabs from './api-tabs.js';
 
 // eslint-disable-next-line no-unused-vars
 function log(...args) {
@@ -80,6 +82,12 @@ Yak
 Zapus
 `.toLowerCase().trim().split(/\s+/);
 
+let mReadyToDetectDuplicatedTab = false;
+
+export function readyToDetectDuplicatedTab() {
+  mReadyToDetectDuplicatedTab = true;
+}
+
 export async function request(tabOrId, options = {}) {
   if (typeof options != 'object')
     options = {};
@@ -99,6 +107,13 @@ export async function request(tabOrId, options = {}) {
   let originalTabId = null;
   let duplicated    = false;
   if (!options.forceNew) {
+    // https://github.com/piroor/treestyletab/issues/2845
+    // This delay may break initial restoration of tabs, so we should
+    // ignore it until all restoration processes are finished.
+    if (mReadyToDetectDuplicatedTab &&
+        configs.delayForDuplicatedTabDetection > 0)
+      await wait(configs.delayForDuplicatedTabDetection);
+
     let oldId = await browser.sessions.getTabValue(tab.id, Constants.kPERSISTENT_ID).catch(ApiTabs.createErrorHandler());
     if (oldId && !oldId.tabId) // ignore broken information!
       oldId = null;
@@ -136,17 +151,31 @@ export async function request(tabOrId, options = {}) {
     }
   }
 
-  const adjective   = kID_ADJECTIVES[Math.floor(Math.random() * kID_ADJECTIVES.length)];
-  const noun        = kID_NOUNS[Math.floor(Math.random() * kID_NOUNS.length)];
-  const randomValue = Math.floor(Math.random() * 1000);
-  const id          = `tab-${adjective}-${noun}-${Date.now()}-${randomValue}`;
+  const id = `tab-${generate()}`;
   // tabId is for detecttion of duplicated tabs
   await browser.sessions.setTabValue(tab.id, Constants.kPERSISTENT_ID, { id, tabId: tab.id }).catch(ApiTabs.createErrorSuppressor());
   return { id, originalId, originalTabId, duplicated };
 }
 
+function generate() {
+  const adjective   = kID_ADJECTIVES[Math.floor(Math.random() * kID_ADJECTIVES.length)];
+  const noun        = kID_NOUNS[Math.floor(Math.random() * kID_NOUNS.length)];
+  const randomValue = Math.floor(Math.random() * 1000);
+  return `${adjective}-${noun}-${Date.now()}-${randomValue}`;
+}
+
 export async function getFromTabs(tabs) {
   return Promise.all(tabs.map(tab =>
     browser.sessions.getTabValue(tab.id, Constants.kPERSISTENT_ID).catch(ApiTabs.createErrorHandler())
   ));
 }
+
+export async function ensureWindowId(windowId) {
+  const storedUniqueId = await browser.sessions.getWindowValue(windowId, 'uniqueId').catch(_ => null);
+  if (storedUniqueId)
+    return storedUniqueId;
+
+  const uniqueId = `window-${generate()}`;
+  await browser.sessions.setWindowValue(windowId, 'uniqueId', uniqueId).catch(_ => null);
+  return uniqueId;
+}
